
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>


#include <glib.h>

#include "gwp_plugin_api.h"
#include "wp_plugin.h"
#include "wp_mdi_document.h"

#include "wp.h"
#include "wp_view.h"
#include "wp_document.h"

#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"


/* XXX make this code draw distinction between WPDocument and WPFrame */


GList *go_plugins = NULL;


/* XXX why do i have to call this all the time? it gtk resetting it? */
static void setup_signals ()
{
  if (signal (SIGPIPE, SIG_IGN) == SIG_ERR)
    perror ("can't trap SIGPIPE:");
}


void gwp_setup_plugin_callbacks (plugin_callback_struct *callbacks)
{
  setup_signals ();

  callbacks->document.create = gwp_plugin_document_create;
  callbacks->document.open = gwp_plugin_document_open;
  callbacks->document.close = gwp_plugin_document_close;
  callbacks->document.show = gwp_plugin_document_show;
  callbacks->document.current = gwp_plugin_document_current;
  callbacks->document.filename = gwp_plugin_document_filename;
  callbacks->document.get_position = gwp_plugin_document_get_position;
  callbacks->document.get_selection = gwp_plugin_document_get_selection_range;
  callbacks->document.set_selection = gwp_plugin_document_set_selection_range;

  callbacks->text.append = gwp_plugin_text_append;
  callbacks->text.insert = gwp_plugin_text_insert;
  callbacks->text.get = gwp_plugin_text_get;
  callbacks->text.get_selected_text = gwp_plugin_text_get_selected_text;
  callbacks->text.set_selected_text = gwp_plugin_text_set_selected_text;


  callbacks->program.reg = gwp_plugin_program_register;
  callbacks->program.quit = gwp_plugin_program_quit;
}




static int extract_document_info (gint docid,
				  WPView **view_r,
				  WPDocument **doc_r)
{
  gpointer p = GINT_TO_POINTER (docid);

  if (IS_WP_MDI_DOCUMENT (p))
    {
      WPMDIDocument *mdi_doc = WP_MDI_DOCUMENT (p);
      if (view_r) (*view_r) = WP_VIEW (mdi->active_view);
      if (doc_r) (*doc_r) = wp_mdi_document_get_document (mdi_doc);
      return 0;
    }

  if (IS_WP_VIEW (p))
    {
      WPView *view = WP_VIEW (p);
      if (view_r) (*view_r) = view;
      if (doc_r) (*doc_r) = wp_view_get_document (view);
      return 0;
    }

  return -1;
}



/*
 *
 *
 *
 */

int gwp_plugin_document_create (gint context, gchar *title)
{
  printf ("==> gwp_plugin_document_create called, context=%d\n", context);
  setup_signals ();

  WPView *orig_view; WPDocument *orig_doc;
  if (extract_document_info (context, &orig_view, &orig_doc))
    return -1;

  WPPlugin *plugin = orig_doc->getPlugin ();
  g_return_val_if_fail (plugin, -1);

  WPMDIDocument *child = wp_mdi_document_new (plugin, title);
  gnome_mdi_add_child (mdi, GNOME_MDI_CHILD(child));
  gnome_mdi_add_view (mdi, GNOME_MDI_CHILD(child));

  int new_context = GPOINTER_TO_INT (child);
  return new_context;
}


int gwp_plugin_document_open (gint context, gchar */*title*/)
{
  printf ("==> gwp_plugin_document_open called, context=%d\n", context);
  setup_signals ();
  return -1;
}


gboolean gwp_plugin_document_close (gint docid)
{
  printf ("==> gwp_plugin_document_close called, docid=%d\n", docid);
  setup_signals ();
  return TRUE;
}


void gwp_plugin_text_append (gint docid, gchar *buffer, gint length)
{
  printf ("==> gwp_plugin_text_append called, docid=%d\n", docid);
  setup_signals ();

  WPView *view;
  WPDocument *doc;
  if (extract_document_info (docid, &view, &doc)) return;

  Slab root = doc->getRoot ();

  /* put the cursor at the end of the document */
  Spot save_point = doc->newDocumentSpot ();
  Spot point = wp_view_get_point (view);
  save_point->setSlab (point->getSlab ());
  save_point->setPosition (point->getPosition ());
  point->setSlab (root->lastLowestContainer ());
  point->setPosition (point->getSlab ()->getNumberOfChildren ());

  for (int i=0; i<length; i++)
    wp_view_insert_char (view, buffer[ i ]);

  /* put the cursor back where it was */
  point->setSlab (save_point->getSlab ());
  point->setPosition (save_point->getPosition ());
  doc->deleteDocumentSpot (save_point);
}


void gwp_plugin_text_insert (gint docid, gchar *buffer,
			     gint length, gint position)
{
  printf ("==> gwp_plugin_text_insert called, docid=%d\n", docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc)) return;

  printf ("    buffer='%s', length=%d, position=%d\n",
	  buffer, length, position);

  /* put the cursor at 'position' */
  Spot save_point = doc->newDocumentSpot ();
  Spot point = wp_view_get_point (view);
  save_point->setSlab (point->getSlab ());
  save_point->setPosition (point->getPosition ());
  doc->set_spot_to_index (point, position);

  for (int i=0; i<length; i++)
    wp_view_insert_char (view, buffer[ i ]);

  /* put the cursor back where it was */
  point->setSlab (save_point->getSlab ());
  point->setPosition (save_point->getPosition ());
  doc->deleteDocumentSpot (save_point);
}


void gwp_plugin_document_show (gint docid)
{
  printf ("==> gwp_plugin_document_show called, docid=%d\n", docid);
  setup_signals ();

  WPDocument *doc;
  if (extract_document_info (docid, NULL, &doc))
    return;

  g_return_if_fail (doc);

  WPMDIDocument *mdi_doc = doc->get_mdi_document ();
  g_return_if_fail (mdi_doc);

  /* get the first view of this document */
  GList *view_p = GNOME_MDI_CHILD (mdi_doc)->views;

  if (! view_p)
    {
      /* no views.  create one */
      gnome_mdi_add_view (mdi, GNOME_MDI_CHILD (mdi_doc));
      view_p = GNOME_MDI_CHILD (mdi_doc)->views;
      g_return_if_fail (view_p);
    }

  WPView *view = WP_VIEW (view_p->data);
  g_return_if_fail (view);

  gnome_mdi_set_active_view (mdi, GTK_WIDGET (view));
}


int gwp_plugin_document_current (gint context)
{
  printf ("==> gwp_plugin_document_current called, context=%d\n", context);
  setup_signals ();

  return context;
}


char *gwp_plugin_document_filename (gint docid)
{
  printf ("==> gwp_plugin_document_filename called, docid=%d\n", docid);
  setup_signals ();

  WPDocument *doc;
  if (extract_document_info (docid, NULL, &doc))
    return "";

  return doc->getDocumentName ();
}


/* XXX this doesn't work well for 2 byte characters... */
/* XXX what should be done about images, etc? */
char *gwp_plugin_text_get (gint docid)
{
  printf ("==> gwp_plugin_text_get called, docid=%d\n", docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return "";

  /* XXX who frees the return value? */
  /* make it static for now... */
  static char *buf = NULL;
  if (buf)
    free (buf); /* free the data from the last call */

  /* get two spots at either end of the document */
  Spot start = doc->newDocumentSpot ();
  Spot end = doc->newDocumentSpot ();
  Slab root = doc->getRoot ();
  end->setSlab (root->lastLowestContainer ());
  end->setPosition (end->getSlab ()->getNumberOfChildren ());

  buf = wp_get_text_between_spots (start, end);

  doc->deleteDocumentSpot (start);
  doc->deleteDocumentSpot (end);

  printf ("gwp_plugin_text_get -- returns:\n%s\n", buf);

  return buf;
}


gboolean gwp_plugin_program_quit ()
{
  printf ("==> gwp_plugin_program_quit called\n");
  setup_signals ();

  return TRUE;
}


void gwp_plugin_program_register (plugin_info *info)
{
  plugin_info *temp;
  setup_signals ();

  temp = info;
  info = (plugin_info *) g_malloc0 (sizeof (plugin_info));
  info->type = temp->type;
  info->menu_location = g_strdup (temp->menu_location);
  info->suggested_accelerator = g_strdup (temp->suggested_accelerator);
  info->plugin_name = g_strdup (temp->plugin_name);
  go_plugins = g_list_append (go_plugins, info);

  printf ("==> gwp_plugin_program_register called: '%s' '%s'\n",
	  info->plugin_name, info->menu_location);
}


gint gwp_plugin_document_get_position (gint docid)
{
  printf ("==> gwp_plugin_document_get_position called, docid=%d\n", docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return 0;

  return doc->spot_to_index (wp_view_get_point (view));
}


selection_range gwp_plugin_document_get_selection_range (gint docid)
{
  printf ("==> gwp_plugin_document_get_selected_range called, docid=%d\n",
	  docid);
  setup_signals ();

  selection_range r;
  r.start = 0;
  r.end = 0;


  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return r;

  Spot start = wp_view_get_point (view);
  Spot end = wp_view_get_mark (view);

  if (start == NULL || end == NULL)
    return r;

  /* make sure start and end are in the order suggested by their names */
  if (! wp_spot_order (start, end))
    {
      Spot t = start; start = end; end = t;
    }

  r.start = doc->spot_to_index (start);
  r.end = doc->spot_to_index (end);

  return r;
}


void gwp_plugin_document_set_selection_range (gint docid, selection_range sr)
{
  printf ("==> gwp_plugin_document_set_selected_range called, docid=%d\n",
	  docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return;

  printf ("    sr = {%d, %d}\n", sr.start, sr.end);

  Spot point = wp_view_get_point (view);
  Spot mark = wp_view_get_mark (view);

  doc->set_spot_to_index (point, sr.start);
  doc->set_spot_to_index (mark, sr.end);

  if (wp_spot_order (point, mark))
    {
      point->hopForwardLeaf ();
      mark->hopBackLeaf ();
    }
  else
    {
      point->hopBackLeaf ();
      mark->hopForwardLeaf ();
    }

  wp_view_hl_on (view);
  wp_view_refresh (view);
}


gchar *gwp_plugin_text_get_selected_text (gint docid)
{
  printf ("==> gwp_plugin_text_get_selected_text called, docid=%d\n", docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return "";

  static char *buf = NULL;
  if (buf)
    {
      free (buf); /* free the data from the last call */
      buf = NULL;
    }

  if (wp_view_hl_is_on (view))
    buf = wp_get_text_between_spots (wp_view_get_point (view),
				     wp_view_get_mark (view));
  else
    return "";

  return buf;
}


gchar *gwp_plugin_text_set_selected_text (gint docid,
					  gchar *data, gint length)
{
  printf ("==> gwp_plugin_text_set_selection_text called, docid=%d\n", docid);
  setup_signals ();

  WPView *view; WPDocument *doc;
  if (extract_document_info (docid, &view, &doc))
    return "";

  Spot point = wp_view_get_point (view);
  Spot mark = wp_view_get_mark (view);

  if (wp_spot_order (point, mark))
    {
      point->hopForwardLeaf ();
      mark->hopBackLeaf ();
    }
  else
    {
      point->hopBackLeaf ();
      mark->hopForwardLeaf ();
    }

  /* XXX gross hack to keep the font information
     the same on the replaced word(s) */
  if (wp_view_hl_is_on (view))
    {
      wp_view_hl_off (view);
      doc->getPlugin ()->updateUI (view);
      wp_view_hl_on (view);
    }
  else
    doc->getPlugin ()->updateUI (view);

  if (wp_view_hl_is_on (view))
    wp_view_delete_char (view);

  point->hopForwardLeaf ();

  for (int i=0; i<length; i++)
    wp_view_insert_char (view, data[ i ]);

  return data; /* ??? */
}
