#include <stdio.h>
#include <gmodule.h>

#include "wp_plugin.h"

#include "wp.h"
/*#include "wp_frame.h"*/
#include "wp_view.h"
#include "wp_document.h"

#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"


typedef struct
{
  WPView *view; /* to keep from duplicating menus */
} wordcount_data;


static char *getUIName ()
{
  return "Word Count";
}


static wordcount_data *get_data (WPView *view)
{
  wordcount_data *fud;

  fud =
    (wordcount_data *) wp_view_get_plugin_scratch_space (view, getUIName ());

  if (! fud)
    {
      fud = (wordcount_data *) calloc (1, sizeof (wordcount_data));
      wp_view_set_plugin_scratch_space (view, getUIName (), fud);
    }

  return fud;
}


void
do_word_count_callback (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  printf ("do_word_count_callback called\n");
  WPView *view = WP_VIEW (mdi->active_view);
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;

  printf ("  docname is '%s'\n", doc->getDocumentName ());

  Slab start, end;

  if (wp_view_hl_is_on (view))
    {
      printf ("  operating on region\n");
      start = wp_view_get_point (view)->getSlab ();
      end = wp_view_get_mark (view)->getSlab ();
    }
  else
    {
      printf ("  operating on entire document\n");
      Slab root = doc->getRoot ();
      start = root->firstLowestContainer ();
      end = root->lastLowestContainer ();
    }

  /* make sure they are in the right order */
  if (wp_slab_order (end, start))
    {
      Slab t = start; start = end; end = t;
    }


  /* count the leaves */
  int count = 0;

  for (Slab n=start;
       n != NULL && wp_slab_order (n, end);
       n = n->getNextLeaf ())
    if (n->classID () == WordSlabID)
      count ++;

  printf ("  result: %d\n", count);
}


static void newUI (WPView *view)
{
  /* insert a menu item under the Tools menu */

  if (!view) return;

  wordcount_data *fud = get_data (view);
  if (!fud) return;
  if (fud->view) return; /* already done */

  fud->view = view;

  GnomeUIInfo entry[2];
  entry[0].type = GNOME_APP_UI_ITEM;
  entry[0].label = getUIName ();
  entry[0].hint = NULL;
  entry[0].moreinfo = do_word_count_callback;
  entry[0].user_data = view;
  entry[0].unused_data = NULL;
  entry[0].pixmap_type = GNOME_APP_PIXMAP_STOCK;
  entry[0].pixmap_info = GNOME_STOCK_MENU_BLANK;
  entry[0].accelerator_key = 0;
  entry[1].type = GNOME_APP_UI_ENDOFINFO;

  GnomeApp *app = gnome_mdi_get_app_from_view (GTK_WIDGET (view));

  int menu_pos;
  char menu_item_path[ strlen (entry[0].label) + 20 ];
  sprintf (menu_item_path, "Tools/%s", entry[0].label);
  if (gnome_app_find_menu_pos (app->menubar, menu_item_path, &menu_pos))
    return; /* already done */

  gnome_app_insert_menus_with_data (app, "Tools/", entry, view);
}


static void deleteUI (WPView */*view*/)
{
}


static void hideUI (WPView */*view*/)
{
}


static void showUI (WPView */*view*/)
{
}


extern "C"
{
  int wp_plugin_fill_vtable (PluginVTable *pvt);
  char *g_module_check_init (GModule *module);
}


int wp_plugin_fill_vtable (PluginVTable *pvt)
{
  pvt->get_ui_name = getUIName;
  pvt->new_ui = newUI;
  pvt->delete_ui = deleteUI;
  pvt->show_ui = showUI;
  pvt->hide_ui = hideUI;
  return 0;
}


char *g_module_check_init (GModule */*module*/)
{
  printf ("wordcount -- g_module_check_init called\n");
  return NULL;
}
