#include <stdio.h>
#include <gmodule.h>

#include "wp_plugin.h"

#include "wp_view.h"
#include "wp_document.h"

#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"


static char *getUIName () { return "ASCII"; }
static int canLoad () { return 0; }
static int canInsert () { return 1; }
static int canSave () { return 1; }
static int canNew () { return 0; }


static int saveDocument (WPDocument *doc, char *filename)
{
  g_return_val_if_fail (doc, -1);
  g_return_val_if_fail (filename, -1);

  FILE *fp = fopen (filename, "w");
  if (! fp)
    {
      perror ("fopen for write");
      GtkWidget *dialog = gnome_error_dialog(_("Can't write file."));
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run(GNOME_DIALOG(dialog));
      return -1;
    }

  Spot start = doc->newDocumentSpot ();
  Spot end = doc->newDocumentSpot ();

  start->setSlab (doc->getRoot ()->firstLowestContainer ());
  start->setPosition (0);

  end->setSlab (doc->getRoot ()->lastLowestContainer ());
  end->setPosition (end->getSlab ()->getNumberOfChildren ());

  char *text = wp_get_text_between_spots (start, end);

  fprintf (fp, "%s", text);

  fclose (fp);
  return 0;
}


static int insertDocument (char *filename, WPView *view)
{
  printf ("asciiIO insertDocument called -- file:'%s' into '%s'\n",
	  filename,
	  wp_view_get_document (view)->getDocumentName ());

  FILE *fp = fopen (filename, "r");
  if (! fp)
    {
      perror ("fopen for read");
      GtkWidget *dialog = gnome_error_dialog(_("Can't read file."));
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run(GNOME_DIALOG(dialog));
      return -1;
    }

  WPDocument *doc = wp_view_get_document (view);
  doc->displayRefreshOff ();

  char c;
  while ((c = fgetc (fp)) != EOF)
    wp_view_insert_char (view, c);

  doc->displayRefreshOn ();

  fclose (fp);

  return 0;
}


extern "C"
{
  int wp_plugin_fill_vtable (PluginVTable *pvt);
  char *g_module_check_init (GModule *module);
}


int wp_plugin_fill_vtable (PluginVTable *pvt)
{
  pvt->get_ui_name = getUIName;
  /*pvt->new_ui = newUI;*/
  /*pvt->delete_ui = deleteUI;*/
  /*pvt->show_ui = showUI;*/
  /*pvt->hide_ui = hideUI;*/

  pvt->can_load = canLoad;
  pvt->can_insert = canInsert;
  pvt->can_save = canSave;
  pvt->can_new = canNew;

  pvt->insert_document = insertDocument;
  pvt->save_document = saveDocument;

  return 0;
}


char *g_module_check_init (GModule */*module*/)
{
  printf ("asciiIO -- g_module_check_init called\n");
  return NULL;
}
