#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <gmodule.h>
#include "wp_plugin.h"
#include "wp_font.h"
#include "wp_utils.h"
#include "wp_clipboard.h"

#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "wp.h"
#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"
#include "wp_document.h"

/*#ifdef BROKEN*/
# define ALLOW_TABLES 1
/*#endif*/ /* BROKEN */

/*#include "pixmaps/gwp-bold.xpm"*/
/*#include "pixmaps/gwp-italic.xpm"*/
/*#include "pixmaps/gwp-underline.xpm"*/
#include "pixmaps/gwp-image.xpm"
#include "pixmaps/gwp-list.xpm"
#if ALLOW_TABLES
# include "pixmaps/gwp-table.xpm"
#endif
#include "pixmaps/gwp-item.xpm"
#include "pixmaps/gwp-color.xpm"
/*#include "pixmaps/gwp-left.xpm"*/
/*#include "pixmaps/gwp-center.xpm"*/
/*#include "pixmaps/gwp-right.xpm"*/
#include "pixmaps/gwp-bullet0.xpm"

#include "plugin_list_helper.h"
#include "plugin_table_helper.h"
#include "plugin_paper_helper.h"

/**
 **  used by the xml code
 **/

typedef unsigned char *ucp;


/*
static unsigned char *itoa (int i)
{
  static char ret[ 2048 ];
  sprintf (ret, "%d", i);
  return (unsigned char *) ret;
}
*/


static unsigned char *ltoa (unsigned long int i)
{
  static char ret[ 2048 ];
  sprintf (ret, "%ld", i);
  return (unsigned char *) ret;
}


unsigned long int AtoI (const unsigned char *a)
{
  if (a == NULL) return 0;
  return strtol ((const char *) a, (char **) NULL, 10);
}



/**
 ** A parsing context.
 **/

typedef struct parseXmlContext {
  xmlDocPtr xmldoc;	/* Xml document */
  xmlNsPtr ns;		/* Main name space */
} parseXmlContext, *parseXmlContextPtr;


/**
 ** used for gtk callbacks
 **/

typedef struct
{
  WPView *view;
  char *s;
} view_and_string;


typedef struct
{
  WPView *view;
  int number;
} view_and_number;


typedef struct
{
  WPView *view;
  double number;
} view_and_double;


#define AUTOSAVETIME 160

/**
 ** plugin internal data structures
 **/

typedef struct
{
  WPDocument *doc;
  Spot spot;
  WPStyleManager *sm; /* keeps track of the styles for this doc */
  int autosave_counter;
} gwp_doc_data;


typedef struct
{
  WPView *view;
  WPStyle *current_insert_style;

  GtkWidget *filesel;
  GtkWidget *hbox;

  GtkWidget *fonts, *pointsizes;
  GtkWidget *bold, *italic, *underline;
  GtkWidget *left_align, *center_align, *right_align;
  GtkWidget *magnification;
  GHashTable *font_index_hash, *pointsize_index_hash;

  GtkWidget *table_popup;
  GtkWidget *list_popup;
  GtkWidget *other_popup;

  GtkWidget *page_setup_window;
} gwp_view_data;


static gwp_doc_data *new_gwp_doc_data (WPDocument *doc);
static void autosave (gwp_doc_data *sd);

gwp_doc_data *get_plugin_document_data (WPDocument *doc)
{
  if (!doc) return NULL;

  gwp_doc_data *sd = (gwp_doc_data *) doc->get_plugin_scratch_space ("gwpIO");
  if (!sd)
    {
      sd = new_gwp_doc_data (doc);
      doc->set_plugin_scratch_space ("gwpIO", sd);
    }

  return sd;
}


gwp_view_data *get_plugin_view_data (WPView *view)
{
  if (!view) return NULL;
  /*WPFrame *frame = wp_view_get_frame (view);*/
  /*if (!frame) return NULL;*/

  gwp_view_data *uid =
    (gwp_view_data *) wp_view_get_plugin_scratch_space (view, "gwpIO");
  if (!uid)
    {
      uid = (gwp_view_data *) malloc (sizeof (gwp_view_data));
      uid->view = view;
      uid->current_insert_style = NULL;
      uid->filesel = NULL;
      uid->hbox = NULL;
      uid->fonts = NULL;
      uid->pointsizes = NULL;
      uid->bold = NULL;
      uid->italic = NULL;
      uid->underline = NULL;
      uid->left_align = NULL;
      uid->center_align = NULL;
      uid->right_align = NULL;
      uid->magnification = NULL;
      uid->font_index_hash = NULL;
      uid->pointsize_index_hash = NULL;
      uid->table_popup = NULL;
      uid->list_popup = NULL;
      uid->other_popup = NULL;
      uid->page_setup_window = NULL;

      /*
      wp_view_set_plugin_scratch_space (view, "gwpIO", uid);
      */
      wp_view_set_plugin_scratch_space (view, "gwpIO", uid);
    }

  return uid;
}

/***
 ***
 ***   xml loading
 ***
 ***/

/*#define BUFFER_SIZE 8193*/

Slab read_xml_slab (WPDocument *doc, parseXmlContextPtr ctxt, xmlNodePtr tree);

WPStyle *read_xml_style (WPDocument *doc,
			 parseXmlContextPtr /*ctxt*/,
			 xmlNodePtr tree)
{
  gwp_doc_data *sd = get_plugin_document_data (doc);
  WPStyleManager *sm = sd->sm;

  /* save the style attributes as part of the current xml node (tree) */
  char *font = (char *) xmlGetProp (tree, (ucp) "Font");
  if (!font) font = "";

  unsigned long int foreground_red =
    AtoI (xmlGetProp (tree, (ucp) "ForegroundRed"));
  unsigned long int foreground_green =
    AtoI (xmlGetProp (tree, (ucp) "ForegroundGreen"));
  unsigned long int foreground_blue =
    AtoI (xmlGetProp (tree, (ucp) "ForegroundBlue"));
  unsigned long int background_red =
    AtoI (xmlGetProp (tree, (ucp) "BackgroundRed"));
  unsigned long int background_green =
    AtoI (xmlGetProp (tree, (ucp) "BackgroundGreen"));
  unsigned long int background_blue =
    AtoI (xmlGetProp (tree, (ucp) "BackgroundBlue"));
  int underline = 
    AtoI (xmlGetProp (tree, (ucp) "Underline"));

  if (background_red == 0 &&
      background_green == 0 &&
      background_blue == 0)
    {
      background_red = 0xA5A5;
      background_green = 0xA5A5;
      background_blue = 0xA5A5;
    }

  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = foreground_red;
  fg->green = foreground_green;
  fg->blue = foreground_blue;
  gdk_color_alloc (colormap, fg);

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = background_red;
  bg->green = background_green;
  bg->blue = background_blue;
  gdk_color_alloc (colormap, bg);

  WPAlignment alignment =
    style_string_to_alignment (xmlGetProp (tree, (ucp) "Alignment"));
  WPDirection direction =
    style_string_to_direction (xmlGetProp (tree, (ucp) "Direction"));
  WPSizePolicy xsizepolicy =
    style_string_to_size_policy (xmlGetProp (tree, (ucp) "XSizePolicy"));
  WPSizePolicy ysizepolicy =
    style_string_to_size_policy (xmlGetProp (tree, (ucp) "YSizePolicy"));

  unsigned long int fixed_left = AtoI (xmlGetProp (tree, (ucp) "Left"));
  unsigned long int fixed_up = AtoI (xmlGetProp (tree, (ucp) "Up"));
  unsigned long int fixed_right = AtoI (xmlGetProp (tree, (ucp) "Right"));
  unsigned long int fixed_down = AtoI (xmlGetProp (tree, (ucp) "Down"));

  unsigned long int margin_left =
    AtoI (xmlGetProp (tree, (ucp) "MarginLeft"));
  unsigned long int margin_up =
    AtoI (xmlGetProp (tree, (ucp) "MarginUp"));
  unsigned long int margin_right =
    AtoI (xmlGetProp (tree, (ucp) "MarginRight"));
  unsigned long int margin_down =
    AtoI (xmlGetProp (tree, (ucp) "MarginDown"));

  WPMarginRenderingPolicy margin_rendering =
    style_string_to_margin_rendering_policy
         (xmlGetProp (tree, (ucp) "MarginRendering"));
  WPWhenEmptyPolicy when_empty =
    style_string_to_when_empty_policy (xmlGetProp (tree, (ucp) "WhenEmpty"));
  unsigned long int max_children =
    AtoI (xmlGetProp (tree, (ucp) "MaxChildren"));

  /*
  style_group *sg =
    sm->getStyleGroupByName (xmlGetProp (tree, (ucp) "StyleGroup"));
  if (!sg)
    style_group *sg = sm->getStyleGroupByName ("word");
  */
  style_group *sg = sm->getStyleGroupByName ("word");

  WPStyle *s = sm->findOrCreateStyle (sg, font, fg, bg, alignment,
				      direction, xsizepolicy, ysizepolicy,
				      fixed_left, fixed_up, fixed_right,
				      fixed_down, margin_left, margin_up,
				      margin_right, margin_down,
				      max_children, margin_rendering,
				      when_empty, underline);

  return s;
}


void *get_logical_type (xmlNodePtr tree)
{
  unsigned const char *lts = xmlGetProp (tree, (ucp) "LT");
  int lt;
  lt = string_to_list_logical_type (lts);
  if (lt) return (void *) lt;
  lt = string_to_table_logical_type (lts);
  if (lt) return (void *) lt;
  return NULL;
}


#if 0
Slab read_xml_cont_slab (WPDocument *doc,
			 parseXmlContextPtr ctxt,
			 xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  Slab s = new _ContSlab (style);
  s->setPluginData (get_logical_type (tree));

  xmlNodePtr c;
  int num;

  for (num=0, c = tree->childs; c; c=c->next, num++)
    {
      Slab child = read_xml_slab (doc, ctxt, c);
      if (child == NULL)
	{
	  fprintf (stderr, "read_xml_slab failed for #%d\n", num);
	  continue;
	}
      if (s->insertChild (num, child, NULL, NULL, -1) != true)
	{
	  fprintf (stderr, "read_xml_cont_slab, insert #%d failed\n", num);
	}
    }

  return s;
}
#endif /* 0 */


#if 0
Slab read_xml_page_slab (WPDocument *doc,
			 parseXmlContextPtr ctxt,
			 xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  Slab s = new _DrawnSlab (style, NULL, NULL, NULL);
  s->setPluginData (get_logical_type (tree));

  xmlNodePtr c;
  int num;

  for (num=0, c = tree->childs; c; c=c->next, num++)
    {
      Slab child = read_xml_slab (doc, ctxt, c);
      if (s->insertChild (num, child, NULL, NULL, -1) != true)
	{
	  fprintf (stderr, "read_xml_cont_slab, insert #%d failed\n", num);
	  /* XXX error */
	}
    }

  return s;
}
#endif /* 0 */


Slab read_xml_lbt_slab (WPDocument *doc,
			parseXmlContextPtr ctxt,
			xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  int levels = 2;
  levels = AtoI (xmlGetProp (tree, (ucp) "Levels"));
  Slab s = new _LBTSlab (style, levels, 0); /* XXX 0 or 1 */
  s->setPluginData (get_logical_type (tree));
  return s;
}


Slab read_xml_word_slab (WPDocument *doc,
			 parseXmlContextPtr ctxt,
			 xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  char *content = (char *) xmlNodeGetContent(tree);
  if (! content) content = strdup("");

  Slab s = new _WordSlab (style, content, strlen (content));
  s->setPluginData (get_logical_type (tree));
  free(content);
  return s;
}


Slab read_xml_st_slab (WPDocument *doc,
		       parseXmlContextPtr ctxt,
		       xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  Slab s = new _STSlab (style);
  s->setPluginData (get_logical_type (tree));
  return s;
}


Slab read_xml_image_slab (WPDocument *doc,
			  parseXmlContextPtr ctxt,
			  xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  Slab s = NULL;

  int lt = (int) get_logical_type (tree);

  /* XXX move this into plugin_list_helper.C? */
  if (lt == lt_list_bullet)
    {
      /*
      GtkWidget *pixmapw = gnome_pixmap_new_from_xpm_d(gwp_bullet0);
      GnomePixmap *gpixmap = GNOME_PIXMAP (pixmapw);
      s = new _ImageSlab (style, gpixmap->pixmap, NULL);
      */

#     if 1
      GdkImlibImage *img;
      img = gdk_imlib_create_image_from_xpm_data (gwp_bullet0);
      s = new _ImageSlab (style, img, NULL);
#     endif /* 0 */
    }
  else /* an image from the file system */
    {
      char *imgfile = (char *) xmlGetProp (tree, (ucp) "FileName");
      if (!imgfile)
	{
	  fprintf (stderr, "error -- image with no filename\n");
	  return NULL;
	}

#     if 0
      GdkPixmap *pmap;
      GdkBitmap *mask;

      if (! gdk_imlib_load_file_to_pixmap (imgfile, &pmap, &mask))
	{
	  fprintf (stderr, "couldn't load image: '%s'\n", imgfile);
	  return NULL;
	}
      s = new _ImageSlab (style, pmap, imgfile);
#     endif /* 0 */

      GdkImlibImage *img = gdk_imlib_load_image (imgfile);
      if (!img)
	{
	  fprintf (stderr, "couldn't load image: '%s'\n", imgfile);
	  return NULL;
	}
      s = new _ImageSlab (style, img, imgfile);
    }

  s->setPluginData ((void *) lt);
  return s;
}


Slab read_xml_cont_slab (WPDocument *doc,
			 parseXmlContextPtr ctxt,
			 xmlNodePtr tree)
{
  WPStyle *style = read_xml_style (doc, ctxt, tree);
  Slab s = NULL;
  int lt = (int) get_logical_type (tree);
  gwp_doc_data *sd = get_plugin_document_data (doc);
  WPStyleManager *sm = sd->sm;

  /* XXX move this into plugin_list_helper.C? */
  if (lt == lt_list_bullet)
    s = new_list_item_bullet (sm, style, -1);
  else
    s = new _ContSlab (style, NULL, NULL, NULL);

  s->setPluginData (get_logical_type (tree));

  xmlNodePtr c;
  int num;

  for (num=0, c = tree->childs; c; c=c->next, num++)
    {
      Slab child = read_xml_slab (doc, ctxt, c);
      if (child == NULL)
	{
	  fprintf (stderr, "read_xml_slab failed for #%d\n", num);
	  continue;
	}
      if (s->insertChild (num, child, NULL, NULL, -1) != true)
	{
	  fprintf (stderr, "read_xml_cont_slab, insert #%d failed\n", num);
	}
    }

  return s;
}



Slab read_xml_slab (WPDocument *doc,
		    parseXmlContextPtr ctxt,
		    xmlNodePtr tree)
{
  Slab s = NULL;

  if (strcmp ((char *) tree->name, "Container") == 0)
    s = read_xml_cont_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Drawn") == 0)
    s = read_xml_cont_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Page") == 0)
    s = read_xml_cont_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Break") == 0)
    s = read_xml_lbt_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Word") == 0)
    s = read_xml_word_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Space") == 0)
    s = read_xml_st_slab (doc, ctxt, tree);
  else if (strcmp ((char *) tree->name, "Image") == 0)
    s = read_xml_image_slab (doc, ctxt, tree);
  else
    fprintf (stderr, "unknown slab type: '%s'\n", (char *) tree->name);

  return s;
}


static int readXmlWPDocument (WPDocument *doc,
			      parseXmlContextPtr ctxt,
			      xmlNodePtr tree)
{
  Slab s = read_xml_slab (doc, ctxt, tree);

  if (!s)
    {
      fprintf (stderr, "document load failed\n");
      return 0;
    }

  doc->setRoot (s);
  return 1;
}


/***
 ***   xml saving
 ***/



xmlNodePtr write_xml_slab (WPDocument *doc,
			   parseXmlContextPtr ctxt,
			   Slab slab);

void write_xml_style (xmlNodePtr cur,
		      parseXmlContextPtr /*ctxt*/,
		      WPStyle *s)
{
  /* save the style attributes as part of cur */
  char *font = s->getFontName ();

  GdkColor *foreground = s->getForeground ();
  GdkColor *background = s->getForeground ();
  unsigned long int foreground_red = 0;
  unsigned long int foreground_green = 0;
  unsigned long int foreground_blue = 0;
  unsigned long int background_red = 0;
  unsigned long int background_green = 0;
  unsigned long int background_blue = 0;
  if (foreground)
    {
      foreground_red = foreground->red;
      foreground_green = foreground->green;
      foreground_blue = foreground->blue;
    }
  if (background)
    {
      background_red = background->red;
      background_green = background->green;
      background_blue = background->blue;
    }

  WPAlignment alignment = s->getAlignment ();
  WPDirection direction = s->getDirection ();
  WPSizePolicy xsizepolicy = s->getXSizePolicy ();
  WPSizePolicy ysizepolicy = s->getYSizePolicy ();
  unsigned long int fixed_left = s->getFixedLeft ();
  unsigned long int fixed_up = s->getFixedUp ();
  unsigned long int fixed_right = s->getFixedRight ();
  unsigned long int fixed_down = s->getFixedDown ();
  unsigned long int margin_left = s->getMarginLeft ();
  unsigned long int margin_up = s->getMarginUp ();
  unsigned long int margin_right = s->getMarginRight ();
  unsigned long int margin_down = s->getMarginDown ();
  WPMarginRenderingPolicy margin_rendering = s->getMarginRendering ();
  WPWhenEmptyPolicy when_empty = s->getWhenEmpty ();
  unsigned long int max_children = s->getMaxChildren ();
  int underline = s->getUnderline ();

  if (strlen (font)) xmlSetProp (cur, (ucp) "Font", (ucp) font);
  if (foreground_red != 0 ||
      foreground_green != 0 ||
      foreground_blue != 0)
    {
      xmlSetProp (cur, (ucp) "ForegroundRed", ltoa (foreground_red));
      xmlSetProp (cur, (ucp) "ForegroundGreen", ltoa (foreground_green));
      xmlSetProp (cur, (ucp) "ForegroundBlue", ltoa (foreground_blue));
    }

  if (background_red != 0xA5A5 ||
      background_green != 0xA5A5 ||
      background_blue != 0xA5A5)
    {
      xmlSetProp (cur, (ucp) "BackgroundRed", ltoa (background_red));
      xmlSetProp (cur, (ucp) "BackgroundGreen", ltoa (background_green));
      xmlSetProp (cur, (ucp) "BackgroundBlue", ltoa (background_blue));
    }

  if (alignment != ALIGN_LOW)
    xmlSetProp (cur, (ucp) "Alignment",
		(ucp) style_alignment_to_string (alignment));
  if (direction != DIR_RIGHT)
    xmlSetProp (cur, (ucp) "Direction",
		(ucp) style_direction_to_string (direction));
  if (xsizepolicy != SIZE_POLICY_BY_CHILDREN)
    xmlSetProp (cur, (ucp) "XSizePolicy",
		(ucp) style_size_policy_to_string (xsizepolicy));
  if (ysizepolicy != SIZE_POLICY_BY_CHILDREN)
    xmlSetProp (cur, (ucp) "YSizePolicy",
		(ucp) style_size_policy_to_string (ysizepolicy));

  if (fixed_left) xmlSetProp (cur, (ucp) "Left", (ucp) itoa (fixed_left));
  if (fixed_up) xmlSetProp (cur, (ucp) "Up", (ucp) itoa (fixed_up));
  if (fixed_right) xmlSetProp (cur, (ucp) "Right", (ucp) itoa (fixed_right));
  if (fixed_down) xmlSetProp (cur, (ucp) "Down", (ucp) itoa (fixed_down));

  if (margin_left) xmlSetProp (cur, (ucp) "MarginLeft",
			       (ucp) itoa (margin_left));
  if (margin_up) xmlSetProp (cur, (ucp) "MarginUp",
			     (ucp) itoa (margin_up));
  if (margin_right) xmlSetProp (cur, (ucp) "MarginRight",
				(ucp) itoa (margin_right));
  if (margin_down) xmlSetProp (cur, (ucp) "MarginDown",
			       (ucp) itoa (margin_down));

  if (margin_rendering != MARGIN_RENDERING_POLICY_BLANK)
    xmlSetProp (cur, (ucp) "MarginRendering",
	   (ucp) style_margin_rendering_policy_to_string (margin_rendering));
  if (when_empty != WHEN_EMPTY_REMOVE)
    xmlSetProp (cur, (ucp) "WhenEmpty",
		(ucp) style_when_empty_policy_to_string (when_empty));
  if (max_children)
    xmlSetProp (cur, (ucp) "MaxChildren", (ucp) (ucp) itoa (max_children));
  if (underline)
    xmlSetProp (cur, (ucp) "Underline", (ucp) (ucp) itoa (underline));
}


void set_logical_type (xmlNodePtr tree, int lt)
{
  char *v;

  if (! lt)
    return;

  v = list_logical_type_to_string ((list_logical_type) lt);
  if (v)
    {
      xmlSetProp (tree, (ucp) "LT", (ucp) v);
      return;
    }
  v = table_logical_type_to_string ((table_logical_type) lt);
  if (v)
    {
      xmlSetProp (tree, (ucp) "LT", (ucp) v);
      return;
    }
}


xmlNodePtr write_xml_cont_slab (WPDocument *doc,
				parseXmlContextPtr ctxt,
				Slab slab)
{
  int i;
  int num = slab->getNumberOfChildren ();

  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns,
                                  (ucp) "Container", NULL);
  write_xml_style (cur, ctxt, slab->getStyle ());

  set_logical_type (cur, (int) slab->getPluginData ());

  for (i=0; i<num; i++)
    {
      Slab child = (Slab) slab->getChildAt (i);
      xmlNodePtr xmlchild = write_xml_slab (doc, ctxt, child);
      if (xmlchild)
	xmlAddChild (cur, xmlchild);
    }

  return cur;
}


xmlNodePtr write_xml_drawn_slab (WPDocument *doc,
				parseXmlContextPtr ctxt,
				Slab slab)
{
  int i;
  int num = slab->getNumberOfChildren ();

  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns,
                                  (ucp) "Drawn", NULL);

  write_xml_style (cur, ctxt, slab->getStyle ());

  set_logical_type (cur, (int) slab->getPluginData ());

  for (i=0; i<num; i++)
    {
      Slab child = (Slab) slab->getChildAt (i);
      xmlNodePtr xmlchild = write_xml_slab (doc, ctxt, child);
      if (xmlchild)
	xmlAddChild (cur, xmlchild);
    }

  return cur;
}


xmlNodePtr write_xml_page_slab (WPDocument *doc,
				parseXmlContextPtr ctxt,
				Slab slab)
{
  int i;
  int num = slab->getNumberOfChildren ();

  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns, (ucp) "Page", NULL);
  write_xml_style (cur, ctxt, slab->getStyle ());
  set_logical_type (cur, (int) slab->getPluginData ());

  for (i=0; i<num; i++)
    {
      Slab child = (Slab) slab->getChildAt (i);
      xmlNodePtr xmlchild = write_xml_slab (doc, ctxt, child);
      if (xmlchild)
	xmlAddChild (cur, xmlchild);
    }

  return cur;
}


xmlNodePtr write_xml_lbt_slab (WPDocument *,
			       parseXmlContextPtr ctxt,
			       Slab s)
{
  LBTSlab lbts = (LBTSlab) s;
  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns, (ucp) "Break", NULL);
  write_xml_style (cur, ctxt, s->getStyle ());
  set_logical_type (cur, (int) s->getPluginData ());
  xmlSetProp (cur, (ucp) "Levels", (ucp) itoa (lbts->getLevels ()));
  return cur;
}


xmlNodePtr write_xml_word_slab (WPDocument *,
				parseXmlContextPtr ctxt,
				Slab s)
{
  WordSlab ws = (WordSlab) s;
  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns,
			       (ucp) "Word",
			       (ucp) ws->getWord ());
  write_xml_style (cur, ctxt, s->getStyle ());
  set_logical_type (cur, (int) s->getPluginData ());

  return cur;
}


xmlNodePtr write_xml_st_slab (WPDocument *,
			      parseXmlContextPtr ctxt,
			      Slab s)
{
  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns, (ucp) "Space", NULL);
  write_xml_style (cur, ctxt, s->getStyle ());
  set_logical_type (cur, (int) s->getPluginData ());

  return cur;
}

xmlNodePtr write_xml_image_slab (WPDocument */*doc*/,
				 parseXmlContextPtr ctxt,
				 Slab s)
{
  ImageSlab ws = (ImageSlab) s;
  xmlNodePtr cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns, (ucp) "Image", NULL);
  set_logical_type (cur, (int) s->getPluginData ());

  if ((int) s->getPluginData () != lt_list_bullet)
    xmlSetProp (cur, (ucp) "FileName", (ucp) ws->get_image_file ());
  write_xml_style (cur, ctxt, ws->getStyle ());

  return cur;
}


xmlNodePtr write_xml_slab (WPDocument *doc,
			   parseXmlContextPtr ctxt,
			   Slab slab)
{
  if (!slab)
    return NULL;

  switch (slab->classID ())
    {
    case ContSlabID:
      return write_xml_cont_slab (doc, ctxt, slab);
    case LBTSlabID:
      return write_xml_lbt_slab (doc, ctxt, slab);
    case WordSlabID:
      return write_xml_word_slab (doc, ctxt, slab);
    case STSlabID:
      return write_xml_st_slab (doc, ctxt, slab);
    case ImageSlabID:
      return write_xml_image_slab (doc, ctxt, slab);
    default:
      /* can't cope */
      return NULL;
    }
}


static xmlNodePtr
writeXmlWPDocument (parseXmlContextPtr ctxt, WPDocument *doc)
{
  xmlNodePtr cur;
  xmlNodePtr child;

  cur = xmlNewDocNode (ctxt->xmldoc, ctxt->ns,
		    (ucp) "GwpDocument",
		    NULL);
  if (cur == NULL)
    return NULL;

  child = write_xml_slab (doc, ctxt, doc->getRoot ());
  xmlAddChild (cur, child);
  return cur;
}



/***
 ***
 ***  start of standard plugin functions
 ***
 ***/



static char *getFileExtension ()
{
  return ".gwp";
}


static char *getMimeType ()
{
  return "text/plain";
}


static char *getUIName ()
{
  return "gwp";
}


static int canLoad ()
{
  return 1;
}


static int canSave ()
{
  return 1;
}


static int canNew ()
{
  return 1;
}


static int deleteChild (WPView *view, Spot spot)
{
  WPDocument *doc = wp_view_get_document (view);

  Slab save_slab = spot->getSlab ();
  int save_pos = spot->getPosition ();

  /* XXX */
  {
  /* make sure spot is to the left of a slab */
  if (spot->getPosition () >= spot->getSlab()->getNumberOfChildren ())
    {
      Slab p;
      boolean found = false;
	  
      for (p=spot->getSlab ()->getNextNode ();
	   /*p = getNextNode (spot->getSlab ());*/
	   p != NULL;
	   /*p = getNextLeaf (p))*/
	   p = p->getNextNode ())
	{
	  if (p->getNumberOfChildren () > 0)
	    {
	      found = true;
	      break;
	    }
	}
	  
      if (found)
	{
	  spot->setSlab (p);
	  spot->setPosition (0);
	}
      else
	/* no slabs after of the cursor */
	return 0;
    }
  }


  Slab slab = spot->getSlab ();

  /* don't delete random parts of a list */
  if (find_list_containing_slab (slab, NULL))
    {
      /* cursor is somewhere in a list */

      if (slab_is_list_bullet (slab))
	{
	  if (! list_item_is_empty (slab))
	    {
	      printf ("list item isn't empty\n");
	      return 0;
	    }

	  return delete_list_item (doc, spot);
	}

      if (! slab_is_list_content (slab))
	{
	  printf ("don't delete random list parts\n");
	  return 0;
	}
    }

  if (find_table_containing_slab (slab, NULL))
    {
      /* cursor is somewhere in a table */
      if (! slab_is_table_content (slab))
	{
	  printf ("don't delete random table parts\n");
	  return 0;
	}
    }

  /* don't delete the last leaf in the document */
  if (slab->classID () != WordSlabID &&
      slab->getNumberOfChildren () == 1 &&
      slab->getNextNode () == NULL &&
      slab->getPrevNode () == NULL)
    {
      printf ("don't delete last leaf in doc\n");
      return 0;
    }

  /* don't delete a word that is next to a list, because
     it is possible to trap the cursor in the list */
  if (slab->getNextLeaf () == NULL &&
      slab->getPrevNode () != NULL &&
      slab->getPrevNode ()->getPluginData () == ((void *) lt_list) &&
      slab->getNumberOfChildren () == 1)
    {
      printf ("don't delete only thing between list and end\n");
      return 0;
    }

  if (slab->getPrevLeaf () == NULL &&
      slab->getNextNode () != NULL &&
      slab->getNextNode ()->getPluginData () == ((void *) lt_list) &&
      slab->getNumberOfChildren () == 1)
    {
      printf ("don't delete only thing between list and start\n");
      return 0;
    }

  int ret = doc->deleteChild (spot);
  if (ret == 0)
    {
      /* delete didn't happen, so put cursor back where it was */
      spot->setSlab (save_slab);
      spot->setPosition (save_pos);
    }
  return ret;
}


static int insertCharacter (WPView *view, Spot spot, unsigned long int c)
{
  g_return_val_if_fail (view, 0);
  g_return_val_if_fail (spot, 0);

  gwp_view_data *uid = get_plugin_view_data (view);
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return 0;

  gwp_doc_data *sd = get_plugin_document_data (doc);
  autosave (sd);


  Slab slab = spot->getSlab ();

  if ((find_list_containing_slab (slab, NULL)) &&
      (! slab_is_list_content (slab)))
    {
      fprintf (stderr, "can't change random part of a list\n");
      return 0;
    }

  if ((find_table_containing_slab (slab, NULL)) &&
      (! slab_is_table_content (slab)))
    {
      fprintf (stderr, "can't change random part of a table\n");
      return 0;
    }

  if (!uid->current_insert_style)
    uid->current_insert_style = slab->getStyle ();

  return doc->insertCharacter (spot, c, uid->current_insert_style);
}


static int insertSlab (WPView *view, Spot spot, Slab slab)
{
  WPDocument *doc = wp_view_get_document (view);
  return doc->insertSlab (spot, slab);
}

void showOtherUI (WPView */*view*/, GtkWidget *popup, GdkEventButton *geb);

static void buttonEvent (WPView *view, GdkEventButton *geb)
{
  WPDocument *doc = wp_view_get_document (view);
  gwp_doc_data *sd = get_plugin_document_data (doc);
  gwp_view_data *uid = get_plugin_view_data (view);
  Spot spot = doc->newDocumentSpot ();
  doc->setSpotToPoint (spot, view, (int) (geb->x), (int) (geb->y));
  Slab slab = spot->getSlab ();

  /* decide what the user clicked on, popup the correct menu */

  if (find_list_containing_slab (slab, NULL))
    showListUI (view, spot, sd->sm, uid->list_popup, geb);
  else if (find_table_containing_slab (slab, NULL))
    showTableUI (view, uid->table_popup, geb);
  else
    showOtherUI (view, uid->other_popup, geb);

  doc->deleteDocumentSpot (spot);
}


static int replaceWord (WPView *view, Spot spot,
			char *new_word, int new_word_length)
{
  if (!view) return 0;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return 0;
  return doc->replaceWord (spot, new_word, new_word_length);
}


static Slab createFollowingWord (WPView *view, Spot spot)
{
  WPDocument *doc = wp_view_get_document (view);
  gwp_view_data *uid = get_plugin_view_data (view);

  if (!uid->current_insert_style)
    uid->current_insert_style = spot->getSlab ()->getStyle ();

  Slab ns = doc->createFollowingWord (spot, uid->current_insert_style);
  return ns;
}


static gwp_doc_data *new_gwp_doc_data (WPDocument *doc)
{
  gwp_doc_data *sd = (gwp_doc_data*) calloc (1, sizeof (gwp_doc_data));
  sd->sm = new WPStyleManager ();
  sd->doc = doc;
  sd->spot = NULL;
  sd->autosave_counter = 0;
  return sd;
}


static WPDocument *loadDocument (char *filename)
{
  WPDocument *doc = new WPDocument (filename);

  parseXmlContext ctxt;
  xmlDocPtr res;
  xmlNsPtr ns;

  res = xmlParseFile (filename);
  if (res == NULL)
    return 0;

  if (res->root == NULL)
    {
      fprintf (stderr, " %s: tree is empty\n", filename);
      xmlFreeDoc (res);
      return 0;
    }

  ns = xmlSearchNsByHref (res, res->root,
			  (ucp) "http://www.superpants.com/");
  if (strcmp ((char *) res->root->name, "GwpDocument") || (ns == NULL))
    {
      fprintf (stderr, "%s: not a Gwp document file\n", filename);
      xmlFreeDoc (res);
      return 0;
    }

  ctxt.xmldoc = res;
  ctxt.ns = ns;

  doc->displayRefreshOff ();

  readXmlWPDocument (doc, &ctxt, res->root->childs);

  xmlFreeDoc (res);

  doc->displayRefreshOn ();

  return doc;
}


/* save without cleaning up autosave file */
static int internalSaveDocument (WPDocument *doc, char *filename)
{
  if (!doc)
    return 0;

  xmlDocPtr xml;
  xmlNsPtr ns;

  parseXmlContext ctxt;
  int ret;

  xml = xmlNewDoc ((ucp) "1.0");
  if (xml == NULL)
    {
      fprintf (stderr, "xmlNewDoc failed\n");
      return -1;
    }

  ns = xmlNewGlobalNs (xml,
		       (ucp) "http://www.superpants.com/",
		       (ucp) "gwp");
  ctxt.xmldoc = xml;
  ctxt.ns = ns;
  xml->root = writeXmlWPDocument (&ctxt, doc);

  xmlSetDocCompressMode (xml, 9);
  ret = xmlSaveFile (filename, xml);
  xmlFreeDoc (xml);
  if (ret < 0)
    return -1;
  return 0;
}


char *build_autosave_filename (char *filename)
{
  if (filename == NULL || strlen (filename) == 0)
    filename = "Untitled";

  char *autosave_filename;
  autosave_filename = (char *) malloc (strlen (filename) + 10);

  int fnl = strlen (filename);
  if (fnl > 4 &&
      filename[ fnl-4 ] == '.' &&
      filename[ fnl-3 ] == 'g' &&
      filename[ fnl-2 ] == 'w' &&
      filename[ fnl-1 ] == 'p')
    {
      /* if the filename is like file.gwp, make the
	 autosave name file~.gwp instead of file.gwp~.gwp */
      strcpy (autosave_filename, filename);
      autosave_filename[ fnl-4 ] = '\0';
      strcat (autosave_filename, "~.gwp");
    }
  else
    {
      strcpy (autosave_filename, filename);
      strcat (autosave_filename, "~.gwp");
    }

  return autosave_filename;
}


/* save and clean up autosave files */
static int saveDocument (WPDocument *doc, char *filename)
{
  g_return_val_if_fail (doc, -1);
  g_return_val_if_fail (filename, -1);

  char *docname = doc->getDocumentName ();
  if (!docname)
    docname = "gwp-unknown";

  char *autosave_filename = build_autosave_filename (docname);

  doc->setDocumentName (filename);

  int ret = internalSaveDocument (doc, filename);
  if (ret)
    {
      fprintf (stderr, "Error saving document: '%s'\n", filename);
      return ret;
    }

  unlink (autosave_filename);

  free (autosave_filename);

  return ret;
}


static void autosave (gwp_doc_data *sd)
{
  if (!sd) return;
  if (!sd->doc) return;

  sd->autosave_counter ++;
  if (sd->autosave_counter >= AUTOSAVETIME)
    {
      printf ("autosaving...\n");

      char *docname = sd->doc->getDocumentName ();
      if (!docname)
	docname = "gwp-unknown";

      char *autosave_filename = build_autosave_filename (docname);

      if (internalSaveDocument (sd->doc, autosave_filename))
	fprintf (stderr, "Warning: can't autosave document '%s'\n", docname);

      free (autosave_filename);
      sd->autosave_counter = 0;
    }
}


static WPDocument *newDocument (char *title)
{
  WPDocument *doc = new WPDocument (title);

  gwp_doc_data *sd = get_plugin_document_data (doc);

  WPStyleManager *sm = sd->sm;

  page_size_info psi;
  psi.width = 614; /* XXX */
  psi.height = 794;
  psi.leftM = 72;
  psi.upM = 72;
  psi.rightM = 72;
  psi.downM = 72;

  Slab root = wp_page_build (wppt_default, sm, &psi);

  doc->setRoot (root);

  return doc;
}

/********************************************************/
/*                  local UI stuff                      */
/********************************************************/



static void popup_cut_callback (GtkWidget *, gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  if (!view) return;
  printf ("popup cut\n");
  wp_clipboard_copy (view,
		     wp_view_get_mark (view),
		     wp_view_get_point (view));
  if (wp_view_hl_is_on (view))
    wp_view_delete_range (view);
}
static void popup_copy_callback (GtkWidget *, gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  if (!view) return;
  printf ("popup copy\n");
  wp_clipboard_copy (view,
		     wp_view_get_mark (view),
		     wp_view_get_point (view));
}
static void popup_paste_callback (GtkWidget *, gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  if (!view) return;
  printf ("popup paste\n");
  wp_clipboard_paste (view);
}


static GnomeUIInfo other_popup_info[] =
{
  { GNOME_APP_UI_ITEM, "Cut", NULL,
    popup_cut_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Copy", NULL,
    popup_copy_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Paste", NULL,
    popup_paste_callback, NULL, NULL },

  { GNOME_APP_UI_ENDOFINFO }
};


GtkWidget *newOtherUI (WPView *view)
{
  GtkWidget *popup = gnome_popup_menu_new (other_popup_info);
  GtkWidget *da = wp_view_get_drawing_area (view);
  gnome_popup_menu_attach (popup, da, view);
  return popup;
}


void deleteOtherUI (WPView */*view*/, GtkWidget *popup)
{
  gtk_widget_destroy (popup);
}


void showOtherUI (WPView *view, GtkWidget *popup, GdkEventButton *geb)
{
  gnome_popup_menu_do_popup (popup, NULL, NULL, geb, view);
}


typedef struct
{
  WPView *view;
  WPAlignment align;
} view_and_line_alignment;


static void
line_align_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_line_alignment *vala = (view_and_line_alignment *) client_data;
  WPView *view = vala->view;
  if (view == NULL) return;
  WPAlignment align = vala->align;
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

# ifdef PLUGIN_SPEW
  printf ("line_align_cb: '%s'\n", doc->getDocumentName ());
# endif

  gwp_doc_data *sd = get_plugin_document_data (doc);
  WPStyleManager *sm = sd->sm;

  /*Spot spot = wp_view_get_point (view);*/

  Slab start_line, end_line;
  Spot point, mark;

  if (wp_view_hl_is_on (view))
    {
      point = wp_view_get_point (view);
      mark = wp_view_get_mark (view);
    }
  else
    {
      point = wp_view_get_point (view);
      mark = wp_view_get_point (view);
    }

  /* get them in the right order */
  if (wp_spot_order (point, mark))
    {
      start_line = point->getSlab ()->getParent ();
      end_line = mark->getSlab ()->getParent ();
    }
  else
    {
      start_line = mark->getSlab ()->getParent ();
      end_line = point->getSlab ()->getParent ();
    }

  style_group *sg = sm->getStyleGroupByName ("line");

  Slab cur = start_line;
  Spot cur_spot = doc->newDocumentSpot ();
  for (cur = start_line; 
       cur && wp_slab_order (cur, end_line);
       cur = cur->getNextNode ())
    {
      WPStyle *s = cur->getStyle ();
      if (s->getAlignment () == align)
	continue;

      WPStyle *ns =
	sm->findOrCreateStyle (sg,
			       s->getFontName (), s->getForeground (),
			       s->getBackground (), align,
			       s->getDirection (), s->getXSizePolicy (),
			       s->getYSizePolicy (), s->getFixedLeft (),
			       s->getFixedUp (), s->getFixedRight (),
			       s->getFixedDown (), s->getMarginLeft (),
			       s->getMarginUp (), s->getMarginRight (),
			       s->getMarginDown (), s->getMaxChildren (),
			       s->getMarginRendering (),
			       s->getWhenEmpty (),
			       s->getUnderline ());

      cur_spot->setSlab ((Slab) (cur->getChildAt (0)));
      cur_spot->setPosition (0);
      if (! cur_spot->getSlab ()) continue;
      doc->setLineStyle (sm, cur_spot, ns->getName ());
    }

  doc->deleteDocumentSpot (cur_spot);
}


static void
insert_image_ok_cb (GtkWidget */*w*/, gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  if (view == NULL) return;
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

# ifdef PLUGIN_SPEW
  printf ("insert image ok: '%s'\n", doc->getDocumentName ());
# endif

  gwp_view_data *uid = get_plugin_view_data (view);
  GtkFileSelection *fs = GTK_FILE_SELECTION (uid->filesel);

  char *filename = gtk_file_selection_get_filename (fs);

# ifdef PLUGIN_SPEW
  printf ("  image filename: '%s'\n", filename);
# endif

  /*
  GdkPixmap *pmap;
  GdkBitmap *mask;
  if (gdk_imlib_load_file_to_pixmap (filename, &pmap, &mask))
  */

  GdkImlibImage *img = gdk_imlib_load_image (filename);

  if (img)
    {
      Spot spot = wp_view_get_point (view);

#     ifdef PLUGIN_SPEW
      printf ("gdk_imlib_load_file_to_pixmap returned 1\n");
#     endif

      if (!uid->current_insert_style)
	uid->current_insert_style = spot->getSlab ()->getStyle ();

      Slab s = new _ImageSlab (uid->current_insert_style, img, filename);
      doc->insertSlab (spot, s);
    }
# ifdef PLUGIN_SPEW
  else
    printf ("gdk_imlib_load_image returned NULL\n");
# endif

  gtk_widget_hide (GTK_WIDGET (fs));
}


static void
insert_image_cancel_cb (GtkWidget */*w*/, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

# ifdef PLUGIN_SPEW
  printf ("insert image cancel: '%s'\n", doc->getDocumentName ());
# endif

  gwp_view_data *uid = get_plugin_view_data (view);
  GtkFileSelection *fs = GTK_FILE_SELECTION (uid->filesel);

  gtk_widget_hide (GTK_WIDGET (fs));
}


static void
insert_image_cb (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

# ifdef PLUGIN_SPEW
  printf ("insert image: '%s'\n", doc->getDocumentName ());
# endif

  gwp_view_data *uid = get_plugin_view_data (view);

  if (!uid->filesel)
    {
      uid->filesel = gtk_file_selection_new("Select Image...");
      GtkFileSelection *filesel = GTK_FILE_SELECTION (uid->filesel);

      gtk_signal_connect (GTK_OBJECT (filesel->ok_button), "clicked",
			  (GtkSignalFunc) insert_image_ok_cb, view);

      gtk_signal_connect (GTK_OBJECT (filesel->cancel_button), "clicked",
			  (GtkSignalFunc) insert_image_cancel_cb, view);
    }

  gtk_widget_show (uid->filesel);
}


static void
insert_list_cb (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  gwp_doc_data *sd = get_plugin_document_data (doc);
  gwp_view_data *uid = get_plugin_view_data (view);
  WPStyleManager *sm = sd->sm;

  Spot spot = wp_view_get_point (view);

  if (! uid->current_insert_style)
    uid->current_insert_style =
      spot->getSlab ()->getStyle ();

  insert_new_list (doc, sm, uid->current_insert_style, spot);
  wp_view_refresh (view);
}


static void insert_new_list_item_into_view (WPView */*view*/,
					    int after_current)
{
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  gwp_doc_data *sd = get_plugin_document_data (doc);
  gwp_view_data *uid = get_plugin_view_data (view);
  WPStyleManager *sm = sd->sm;

  Spot spot = wp_view_get_point (view);

  if (! uid->current_insert_style)
    uid->current_insert_style = spot->getSlab ()->getStyle ();

  insert_new_list_item (sm, uid->current_insert_style, spot, after_current);
  wp_view_refresh (view);
}


static void
insert_list_item_cb (GtkWidget */*widget*/, gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  insert_new_list_item_into_view (view, 1);
}



#if ALLOW_TABLES
static void
insert_table_cb (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;
  gwp_view_data *uid = get_plugin_view_data (view);
  if (!uid) return;

  if (! uid->current_insert_style)
    uid->current_insert_style =
      wp_view_get_point (view)->getSlab ()->getStyle ();

  if (! uid->current_insert_style)
    return;

  gwp_doc_data *sd = get_plugin_document_data (doc);
  WPStyleManager *sm = sd->sm;

  Spot spot = wp_view_get_point (view);
  insert_new_table (doc, sm, uid->current_insert_style, spot, 4, 4);
  wp_view_refresh (view);
}
#endif /* ALLOW_TABLES */


static void regional_font_change (WPDocument *doc, WPView *view,
				  int (*tester) (Slab s, void *user_data),
				  WPStyle *(*setter) (WPStyleManager *sm,
						      WPStyle *s, int set,
						      void *user_data),
				  int set, void *user_data)
{
  gwp_doc_data *sd;
  WPStyleManager *sm;
  style_group *sg;

  sd = get_plugin_document_data (doc);
  sm = sd->sm;
  sg = sm->getStyleGroupByName ("word");

  Spot begin_hl, end_hl,
    point = wp_view_get_point (view),
    mark = wp_view_get_mark (view);

  /* get them in the right order */
  if (wp_spot_order (point, mark))
    {
      begin_hl = point;
      end_hl = mark;
    }
  else
    {
      begin_hl = mark;
      end_hl = point;
    }

  /* if the spots are on the edges of words, move them inward */
  begin_hl->hopForward ();
  end_hl->hopBack ();

  /* if the ends need spliting, do it here */
  if (begin_hl->getPosition () > 0 &&
      (*tester) (begin_hl->getSlab (), user_data) != set)
    {
      printf ("splitting first\n");
      if (doc->ensureBorder (begin_hl) == false)
	return;
      printf ("ok\n");
    }

  if (end_hl->getPosition () < end_hl->getSlab ()->getNumberOfChildren ()
      && (*tester) (end_hl->getSlab (), user_data) != set)
    {
      printf ("splitting last\n");
      if (doc->ensureBorder (end_hl) == false)
	return;
      printf ("ok\n");
    }

  /* make sure they ended up on the correct sides of the splits */
  begin_hl->hopForward ();
  end_hl->hopBack ();

  /*Spot cur = WP_NewSpot (doc);*/
  Spot cur = doc->newDocumentSpot ();
  cur->setSlab (begin_hl->getSlab ());
  cur->setPosition (begin_hl->getPosition ());

  doc->displayRefreshOff ();

  /* change the style of all slabs between the point and spot */
  while (wp_spot_order (cur, end_hl) > 0)
    {
      /*
      printf ("---------\n");
      printf ("cur=%p, end_hl=%p\n", cur, end_hl);
      printf ("cur->slab()=%p, end_hl->slab()=%p\n",
	      cur->getSlab (),
	      end_hl->getSlab ());
      printf ("cur:");
      cur->getSlab ()->dd ();
      printf ("\n");
      printf ("end_hl:");
      end_hl->getSlab ()->dd ();
      printf ("\n");
      */

      if ((*tester) (cur->getSlab (), user_data) != set)
	{
	  /*printf ("changing.\n");*/
	  WPStyle *s = cur->getSlab ()->getStyle ();
	  WPStyle *new_s = (*setter) (sm, s, set, user_data);
	  doc->setStyle (sm, cur, "word", new_s->getName());
	}
      /*
      else
	printf ("not changing.\n");
      */

      /*
      if (WP_MoveToNextSlab (doc, cur) == 0)
	break;
      */
      Slab next = cur->getSlab ()->getNextLeaf ();
      if (!next) break;
      cur->setSlab (next);
    }

  doc->displayRefreshOn ();
  doc->refresh ();

  /*WP_DeleteSpot (doc, cur);*/
  doc->deleteDocumentSpot (cur);
}

static WPStyle *set_style_fg_color (WPStyleManager *sm, WPStyle *s,
				    int /*set*/, void *user_data)
{
  GdkColor *set_fg = (GdkColor *) user_data;
  style_group *sg = sm->getStyleGroupByName ("word");

  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 s->getFontName (),
				 set_fg,
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  return new_s;
}

int slab_is_fg_color (Slab s, void *user_data)
{
  GdkColor *set = (GdkColor *) user_data;
  GdkColor *fg = s->getStyle ()->getForeground ();

  if (fg->red == set->red &&
      fg->green == set->green &&
      fg->blue == set->blue)
    return 1;

  return 0;
}


static void
set_fg_color_func (GnomeColorPicker *cp,
		   gint /*r*/,
		   gint /*g*/,
		   gint /*b*/,
		   gint /*a*/,
		   WPView *view)
{
  printf ("set_fg_color_func\n");
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;
  gwp_view_data *uid = get_plugin_view_data (view);
  if (!uid) return;
  gwp_doc_data *sd = get_plugin_document_data (doc);
  if (!sd) return;
  WPStyleManager *sm = sd->sm;
  if (!sm) return;

  gushort red, green, blue, alpha;
  gnome_color_picker_get_i16 (cp, &red, &green, &blue, &alpha);

  if (! uid->current_insert_style)
    uid->current_insert_style =
      wp_view_get_point (view)->getSlab ()->getStyle ();

  if (! uid->current_insert_style)
    return;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *c = (GdkColor *) malloc (sizeof (GdkColor));
  c->red = red;
  c->green = green;
  c->blue = blue;
  gdk_color_alloc (colormap, c);

  printf ("selected color: 0x%x 0x%x 0x%x\n", c->red, c->green, c->blue);

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_fg_color,
			  set_style_fg_color, 1, c);
  else
    uid->current_insert_style =
      set_style_fg_color (sm, uid->current_insert_style, 1, c);

}


static WPStyle *set_style_bg_color (WPStyleManager *sm, WPStyle *s,
				    int /*set*/, void *user_data)
{
  GdkColor *set_bg = (GdkColor *) user_data;
  style_group *sg = sm->getStyleGroupByName ("word");

  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 s->getFontName (),
				 s->getForeground (),
				 set_bg,
				 s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  return new_s;
}

int slab_is_bg_color (Slab s, void *user_data)
{
  GdkColor *set = (GdkColor *) user_data;
  GdkColor *bg = s->getStyle ()->getBackground ();

  if (bg->red == set->red &&
      bg->green == set->green &&
      bg->blue == set->blue)
    return 1;

  return 0;
}


static void
set_bg_color_func (GnomeColorPicker *cp,
		   gint /*r*/,
		   gint /*g*/,
		   gint /*b*/,
		   gint /*a*/,
		   WPView *view)
{
  printf ("set_bg_color_func\n");
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;
  gwp_view_data *uid = get_plugin_view_data (view);
  if (!uid) return;
  gwp_doc_data *sd = get_plugin_document_data (doc);
  if (!sd) return;
  WPStyleManager *sm = sd->sm;
  if (!sm) return;

  gushort red, green, blue, alpha;
  gnome_color_picker_get_i16 (cp, &red, &green, &blue, &alpha);

  if (! uid->current_insert_style)
    uid->current_insert_style =
      wp_view_get_point (view)->getSlab ()->getStyle ();

  if (! uid->current_insert_style)
    return;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *c = (GdkColor *) malloc (sizeof (GdkColor));
  c->red = red;
  c->green = green;
  c->blue = blue;
  gdk_color_alloc (colormap, c);

  printf ("selected color: 0x%x 0x%x 0x%x\n", c->red, c->green, c->blue);

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_bg_color,
			  set_style_bg_color, 1, c);
  else
    uid->current_insert_style =
      set_style_bg_color (sm, uid->current_insert_style, 1, c);
}


static void
change_color_cb (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  gwp_view_data *uid = get_plugin_view_data (view);


  GtkWidget *dialog;
  GtkWidget *fg_label, *bg_label;
  GtkWidget *hbox;

  dialog = gnome_dialog_new(_("Select Colors"),
			    GNOME_STOCK_BUTTON_OK,
			    GNOME_STOCK_BUTTON_CANCEL,
			    0);


  /*
  fg_button = gnome_color_selector_get_button(fg_gcs);
  bg_button = gnome_color_selector_get_button(bg_gcs);
  */

  GtkWidget *fg_button = gnome_color_picker_new ();
  GtkWidget *bg_button = gnome_color_picker_new ();
  if (uid && uid->current_insert_style)
    {
      GdkColor *fg = uid->current_insert_style->getForeground ();
      GdkColor *bg = uid->current_insert_style->getBackground ();
      gnome_color_picker_set_i16 (GNOME_COLOR_PICKER (fg_button),
				  fg->red, fg->green, fg->blue, 0xFFFF);
      gnome_color_picker_set_i16 (GNOME_COLOR_PICKER (bg_button),
				  bg->red, bg->green, bg->blue, 0xFFFF);
    }

  gtk_signal_connect (GTK_OBJECT (fg_button), "color_set",
		      GTK_SIGNAL_FUNC (set_fg_color_func), view);
  gtk_signal_connect (GTK_OBJECT (bg_button), "color_set",
		      GTK_SIGNAL_FUNC (set_bg_color_func), view);
  
  fg_label = gtk_label_new(_("Select Foreground:"));
  bg_label = gtk_label_new(_("Select Background:"));

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 0);

  gtk_box_pack_start(GTK_BOX(hbox), fg_label, TRUE, TRUE, 0);
  gtk_widget_show(fg_label);
  gtk_box_pack_start(GTK_BOX(hbox), fg_button, TRUE, TRUE, 0);
  gtk_widget_show(fg_button);

  gtk_widget_show(hbox);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 0);

  gtk_box_pack_start(GTK_BOX(hbox), bg_label, TRUE, TRUE, 0);
  gtk_widget_show(bg_label);
  gtk_box_pack_start(GTK_BOX(hbox), bg_button, TRUE, TRUE, 0);
  gtk_widget_show(bg_button);

  gtk_widget_show(hbox);

  gnome_dialog_set_default(GNOME_DIALOG(dialog), 0);
  gnome_dialog_set_close(GNOME_DIALOG(dialog), TRUE);
  gnome_dialog_run(GNOME_DIALOG(dialog));
}


static WPStyle *set_style_bold (WPStyleManager *sm, WPStyle *s,
				int set, void */*user_data*/)
{
  printf ("set_style_bold: %d\n", set);

  style_group *sg = sm->getStyleGroupByName ("word");

  char *fontName = s->getFontName();
  char *newFontName;
	      
  if (! set)
    newFontName = setFontWeight(fontName, "bold");
  else
    newFontName = setFontWeight(fontName, "medium");

  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 newFontName, s->getForeground (),
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  free (newFontName);

  return new_s;
}


int slab_is_bold (Slab s, void */*user_data*/)
{
  char *font = s->getStyle ()->getFontName ();
  if (font == NULL) return 0;
  char *slab_weight = getFontWeight (font);
  if (slab_weight == NULL) return 0;
  if (strcasecmp (slab_weight, "bold") == 0)
    {
      free (slab_weight);
      printf ("slab_is_bold: 0\n");
      return 0;
    }
  free (slab_weight);
  printf ("slab_is_bold: 1\n");
  return 1;
}


static void
bold_cb (GtkWidget *widget, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  int bold; /* 1 if we are setting bold, 0 if we are clearing it */

#ifdef PLUGIN_SPEW
  printf ("bold_cb: '%s'\n", doc->getDocumentName ());
#endif

  bold = ! (GTK_TOGGLE_BUTTON (widget)->active);
  printf ("bold_cb: bold = %d\n", bold);

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_bold,
			  set_style_bold, bold, NULL);
  else
    {
      gwp_doc_data *sd = get_plugin_document_data (doc);
      gwp_view_data *uid = get_plugin_view_data (view);
      Spot spot = wp_view_get_point (view);
      WPStyle *s = spot->getSlab ()->getStyle ();
      WPStyleManager *sm = sd->sm;

      WPStyle *new_s = set_style_bold (sm, s, bold, NULL);
      if (new_s)
	uid->current_insert_style = new_s;
    }
}


static WPStyle *set_style_italic (WPStyleManager *sm, WPStyle *s,
				  int set, void */*user_data*/)
{
  style_group *sg = sm->getStyleGroupByName ("word");

  char *fontName = s->getFontName();
  char *newFontName;
  char *font_family = getFontFamily(fontName);


  /* XXX this is gross.  we need a more generic solution to
     whether a font can be Oblique or Italic. */
  if (!strcasecmp(font_family, "helvetica"))
    {
      if (! set)
	newFontName = setFontSlant(fontName, "o");
      else
	newFontName = setFontSlant(fontName, "r");
    }
  else
    {
      if (! set)
	newFontName = setFontSlant(fontName, "i");
      else
	newFontName = setFontSlant(fontName, "r");
    }
	      
  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 newFontName, s->getForeground (),
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  free (newFontName);

  return new_s;
}


int slab_is_italic (Slab s, void */*user_data*/)
{
  char *font = s->getStyle ()->getFontName ();
  if (font == NULL) return 0;
  char *slab_slant = getFontSlant (font);
  if (slab_slant == NULL) return 0;
  if (strcasecmp (slab_slant, "r") == 0)
    {
      free (slab_slant);
      return 1;
    }
  free (slab_slant);
  return 0;
}


static void
italic_cb (GtkWidget *widget, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  int italic; /* 1 if we are setting italic, 0 if we are clearing it */

#ifdef PLUGIN_SPEW
  printf ("italic_cb: '%s'\n", doc->getDocumentName ());
#endif

  italic = ! (GTK_TOGGLE_BUTTON (widget)->active);

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_italic,
			  set_style_italic, italic, NULL);
  else
    {
      gwp_doc_data *sd = get_plugin_document_data (doc);
      gwp_view_data *uid = get_plugin_view_data (view);
      Spot spot = wp_view_get_point (view);
      WPStyle *s = spot->getSlab ()->getStyle ();
      WPStyleManager *sm = sd->sm;

      WPStyle *new_s = set_style_italic (sm, s, italic, NULL);
      if (new_s)
	uid->current_insert_style = new_s;
    }
}


static WPStyle *set_style_underline (WPStyleManager *sm, WPStyle *s,
				  int set, void */*user_data*/)
{
  printf ("set_style_underline: %d\n", set);
  style_group *sg = sm->getStyleGroupByName ("word");

  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 s->getFontName (), s->getForeground (),
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 ! set);

  return new_s;
}


int slab_is_underline (Slab s, void */*user_data*/)
{
  printf ("slab_is_underline: %d\n", s->getStyle ()->getUnderline ());
  return ! s->getStyle ()->getUnderline ();
}


static void
underline_cb (GtkWidget *widget, gpointer /*client_data*/)
{
  /*WPView *view = (WPView *) client_data;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  int underline; /* 1 if we are setting underline, 0 if we are clearing it */

# ifdef PLUGIN_SPEW
  printf ("underline_cb: '%s'\n", doc->getDocumentName ());
# endif

  underline = ! (GTK_TOGGLE_BUTTON (widget)->active);
  printf ("underline_cb: underline = %d\n", underline);


  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_underline,
			  set_style_underline, underline, NULL);
  else
    {
      gwp_doc_data *sd = get_plugin_document_data (doc);
      gwp_view_data *uid = get_plugin_view_data (view);
      Spot spot = wp_view_get_point (view);
      WPStyle *s = spot->getSlab ()->getStyle ();
      WPStyleManager *sm = sd->sm;

      WPStyle *new_s = set_style_underline (sm, s, underline, NULL);
      if (new_s)
	uid->current_insert_style = new_s;
    }
}


static WPStyle *set_style_family (WPStyleManager *sm, WPStyle *s,
				  int /*set*/, void *user_data)
{
  style_group *sg = sm->getStyleGroupByName ("word");
  char *foundry_family = (char *) user_data;

  char *fontName = s->getFontName();
  char *newFontName0;
  char *newFontName1;

  /* split something like "adobe-utopia" into "adobe" and "utopia" */
  char *newFontFoundry = (char *) malloc (strlen (foundry_family)+3);
  char *newFontFamily = (char *) malloc (strlen (foundry_family)+3);
  strcpy (newFontFoundry, foundry_family);
  char *e = strchr (newFontFoundry, '-');
  if (e == NULL)
    {
      fprintf (stderr, "gwpIO set_font_family, no '-' found\n");
      return NULL;
    }
  *e = '\0';
  strcpy (newFontFamily, foundry_family + (e-newFontFoundry) + 1);

  newFontName0 = setFontFoundry(fontName, newFontFoundry);
  newFontName1 = setFontFamily(newFontName0, newFontFamily);
	      
  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 newFontName1, s->getForeground (),
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  free (newFontName0);
  free (newFontName1);
  free (newFontFoundry);
  free (newFontFamily);

  return new_s;
}


int slab_is_family (Slab s, void *user_data)
{
  char *font = s->getStyle ()->getFontName ();
  char *foundry_family = (char *) user_data;
  if (font == NULL) return 0;
  char *slab_foundry = getFontFoundry (font);
  if (slab_foundry == NULL) return 0;
  char *slab_family = getFontFamily (font);
  if (slab_family == NULL) return 0;

  char *cur_foundry_family = (char *) malloc (strlen (slab_foundry) +
					      strlen (slab_family) + 5);

  sprintf (cur_foundry_family, "%s-%s", slab_foundry, slab_family);

  if (strcasecmp (cur_foundry_family, foundry_family) == 0)
    {
      free (slab_family);
      free (slab_foundry);
      free (cur_foundry_family);
      return 1;
    }
  free (slab_family);
  free (slab_foundry);
  free (cur_foundry_family);
  return 0;
}


static void
family_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_string *das = (view_and_string *) client_data;
  if (das == NULL) return;
  /*WPView *view = das->view;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  char *foundry_family = das->s;

# ifdef PLUGIN_SPEW
  printf ("family_cb: '%s'\n", doc->getDocumentName ());
# endif

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_family,
			  set_style_family, 1, foundry_family);
  else
    {
      gwp_doc_data *sd = get_plugin_document_data (doc);
      gwp_view_data *uid = get_plugin_view_data (view);
      Spot spot = wp_view_get_point (view);
      WPStyle *s = spot->getSlab ()->getStyle ();
      WPStyleManager *sm = sd->sm;

      WPStyle *new_s = set_style_family (sm, s, 1, foundry_family);
      if (new_s)
	uid->current_insert_style = new_s;
    }
}


static WPStyle *set_style_point_size (WPStyleManager *sm, WPStyle *s,
				      int /*set*/, void *user_data)
{
  style_group *sg = sm->getStyleGroupByName ("word");
  int psize = (int) user_data;

  char *fontName = s->getFontName();
  char *newFontName;

  newFontName = setFontPointSize (fontName, psize * 10);
	      
  WPStyle *new_s;
  new_s = sm->findOrCreateStyle (sg,
				 newFontName, s->getForeground (),
				 s->getBackground (), s->getAlignment(),
				 s->getDirection (), s->getXSizePolicy (),
				 s->getYSizePolicy (), s->getFixedLeft (),
				 s->getFixedUp (), s->getFixedRight (),
				 s->getFixedDown (), s->getMarginLeft (),
				 s->getMarginUp (), s->getMarginRight (),
				 s->getMarginDown (), s->getMaxChildren (),
				 s->getMarginRendering (),
				 s->getWhenEmpty (),
				 s->getUnderline ());

  free (newFontName);

  return new_s;
}


int slab_is_point_size (Slab s, void *user_data)
{
  char *font = s->getStyle ()->getFontName ();
  if (font == NULL) return 0;
  int psize = (int) user_data;

  int fontsize = (int) ((float) getFontPointSize (font) / 10.0);
  if (fontsize == 0)
    fontsize = getFontPixelSize (font);

  if (psize == fontsize)
    return 1;

  return 0;
}



static void
pointsize_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_number *dap = (view_and_number *) client_data;
  if (dap == NULL) return;
  /*WPView *view = dap->view;*/
  /*if (view == NULL) return;*/
  /*view = wp_frame_get_focus_view (wp_view_get_frame (view));*/
  WPView *view = WP_VIEW (mdi->active_view);
  if (view == NULL) return;
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL) return;

  int psize = dap->number;

# ifdef PLUGIN_SPEW
  printf ("pointsize_cb: '%s'\n", doc->getDocumentName ());
# endif

  if (wp_view_hl_is_on(view))
    regional_font_change (doc, view, slab_is_point_size,
			  set_style_point_size, 1, (void *) psize);
  else
    {
      gwp_doc_data *sd = get_plugin_document_data (doc);
      gwp_view_data *uid = get_plugin_view_data (view);
      Spot spot = wp_view_get_point (view);
      WPStyle *s = spot->getSlab ()->getStyle ();
      WPStyleManager *sm = sd->sm;

      WPStyle *new_s = set_style_point_size (sm, s, 1, (void *) psize);
      if (new_s)
	uid->current_insert_style = new_s;
    }
}

static void
add_font (gwp_view_data *uid,
	  GtkWidget *menu,
	  char *ui_name, char *foundry_family)
{
  GtkWidget *item;
  static int index = 0;

  item = gtk_menu_item_new_with_label(ui_name);
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_widget_show(item);

  view_and_string *vas = (view_and_string *) malloc (sizeof (view_and_string));
  vas->view = uid->view;
  vas->s = foundry_family;
  gtk_signal_connect(GTK_OBJECT(item),
		     "activate", GTK_SIGNAL_FUNC(family_cb),
		     vas);

  g_hash_table_insert(uid->font_index_hash,
		      foundry_family, (gpointer)index++);
}

static void
add_fonts(gwp_view_data *uid, GtkWidget *option)
{
  GtkWidget *menu = gtk_menu_new();

  if (uid->font_index_hash == NULL)
    uid->font_index_hash = g_hash_table_new (g_str_hash, g_str_equal);

  add_font(uid, menu, _("Adobe-Helvetica"), "adobe-helvetica");
  add_font(uid, menu, _("Adobe-Courier"), "adobe-courier");
  add_font(uid, menu, _("Bitstream-Charter"), "bitstream-charter");
  add_font(uid, menu, _("Bitstream-Courier"), "bitstream-courier");
  add_font(uid, menu, _("Adobe-Mincho"), "adobe-mincho");
  add_font(uid, menu, _("Adobe-Symbol"), "adobe-symbol");

  /*-isas-fangsong ti-medium-r-normal--16-160-72-72-c-160-gb2312.1980-0*/
  add_font(uid, menu, _("isas-fangsong ti"), "isas-fangsong ti");
  add_font(uid, menu, _("isas-song ti"), "isas-song ti");

  gtk_option_menu_set_menu (GTK_OPTION_MENU (option), menu);
}

static char* point_sizes[] = {
  "6","8","10","12","14","16","18","20","22","24",
  "28","32","36","40","50","60","72"
};
static int num_point_sizes = sizeof(point_sizes) / sizeof(point_sizes[0]);

static void
add_pointsizes (gwp_view_data *uid, GtkWidget *option)
{
  GtkWidget *menu = gtk_menu_new();
  int i;

  if (uid->pointsize_index_hash == NULL)
    {
      uid->pointsize_index_hash = g_hash_table_new (g_str_hash, g_str_equal);
    }

  for (i = 0; i < num_point_sizes; i ++)
    {
      view_and_number *van =
	(view_and_number*)malloc(sizeof(view_and_number));
      GtkWidget *item = gtk_menu_item_new_with_label(point_sizes[i]);
      gtk_menu_append(GTK_MENU(menu), item);
      gtk_widget_show(item);
      van->view = uid->view;
      /*van->number = atoi(point_sizes[i]) * 10;*/
      van->number = atoi(point_sizes[i]);
      gtk_signal_connect(GTK_OBJECT(item),
			 "activate", GTK_SIGNAL_FUNC(pointsize_cb),
			 van);
      g_hash_table_insert (uid->pointsize_index_hash,
			   point_sizes[i], (gpointer)i);
    }

  gtk_option_menu_set_menu(GTK_OPTION_MENU(option),
			   menu);
}


static void
magnification_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_double *van = (view_and_double *) client_data;
  if (! van) return;
  if (! van->view) return;

  printf ("switching view to magnification: %f\n", van->number);

  wp_view_set_scale_factor (van->view, van->number);
}


static char* percentages[] = {
  "50","75","100","120","150","200","300"
};
static int num_percentages = sizeof(percentages) / sizeof(percentages[0]);


static void
add_magnifications (WPView *view, GtkWidget *option)
{
  GtkWidget *menu = gtk_menu_new();
  int i;

  for (i = 0; i < num_percentages; i ++)
    {
      view_and_double *van =
	(view_and_double *) malloc (sizeof (view_and_double));
      char buf[ 10 ];
      GtkWidget *item;

      g_snprintf(buf, sizeof (buf), "%s%%", percentages[ i ]);

      item = gtk_menu_item_new_with_label (buf);

      gtk_menu_append (GTK_MENU (menu), item);
      gtk_widget_show (item);
      van->view = view;
      van->number = (double) (atoi (percentages[ i ])) / 100.0;
      gtk_signal_connect (GTK_OBJECT (item),
			  "activate", GTK_SIGNAL_FUNC (magnification_cb),
			  van);
    }

  gtk_option_menu_set_menu (GTK_OPTION_MENU (option), menu);
  /* set 100% as current scale */
  gtk_option_menu_set_history (GTK_OPTION_MENU (option), 2);
}


static void 
file_pagesetup_cmd_callback (GtkWidget */*widget*/,
			     gpointer client_data)
{
  WPView *view = (WPView *) client_data;
  if (!view) return;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;
  gwp_view_data *uid = get_plugin_view_data (view);
  if (!uid) return;
  gwp_doc_data *sd = (gwp_doc_data *) doc->get_plugin_scratch_space ("gwpIO");
  if (!sd) return;

  if (uid->page_setup_window == NULL)
    uid->page_setup_window = wp_new_page_setup_window (view, sd->sm);

  gtk_widget_show (uid->page_setup_window);
}



/***************************************************/
/*              end of local UI stuff              */
/***************************************************/


static void newUI (WPView *view)
{
  if (!view) return;
  gwp_view_data *uid = get_plugin_view_data (view);
  if (!uid) return;
  GtkWidget *plugin_toolbar = wp_view_get_plugin_toolbar (view);
  if (!plugin_toolbar) return;

  GtkWidget *da = wp_view_get_drawing_area (view);

  GtkWidget *pixmap, *element;
  GtkWidget *widget;
  /*GtkWidget *rtl, *ltr;*/
  GtkWidget *toolbar;

  if (uid->hbox != NULL)
    return; /* already did this */

  uid->hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(uid->hbox);
  gtk_container_add (GTK_CONTAINER (plugin_toolbar), uid->hbox);

  toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL,
			    GTK_TOOLBAR_ICONS);
  gtk_box_pack_start(GTK_BOX(uid->hbox), toolbar, TRUE, TRUE, 0);
  gtk_widget_show(toolbar);

  uid->fonts = gtk_option_menu_new();
  gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar),
			    uid->fonts,
			    _("Select Font"), NULL);
  gtk_widget_show(uid->fonts);
  add_fonts (uid, uid->fonts);

  uid->pointsizes = gtk_option_menu_new();
  gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar),
			    uid->pointsizes,
			    _("Select Point Size"), NULL);
  gtk_widget_show (uid->pointsizes);
  add_pointsizes (uid, uid->pointsizes);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  pixmap = gnome_pixmap_new_from_xpm_d(gwp_image);
  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("Image"),
				   _("Insert Image"),
				   NULL,
				   pixmap,
				   (GtkSignalFunc)insert_image_cb,
				   view);
  gtk_widget_show(widget);

  pixmap = gnome_pixmap_new_from_xpm_d(gwp_list);
  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("List"),
				   _("Insert List"),
				   NULL,
				   pixmap,
				   (GtkSignalFunc)insert_list_cb,
				   view);
  gtk_widget_show(widget);

  pixmap = gnome_pixmap_new_from_xpm_d(gwp_item);
  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("List Item"),
				   _("Insert List Item"),
				   NULL,
				   pixmap,
				   (GtkSignalFunc)insert_list_item_cb,
				   view);
  gtk_widget_show(widget);


# ifdef ALLOW_TABLES
  pixmap = gnome_pixmap_new_from_xpm_d (gwp_table);
  widget = gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
				    _("Table"),
				    _("Insert Table"),
				    NULL,
				    pixmap,
				    (GtkSignalFunc) insert_table_cb,
				    view);
  gtk_widget_show(widget);
# endif /* ALLOW_TABLES */


  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  /*pixmap = gnome_pixmap_new_from_xpm_d(gwp_bold);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_TEXT_BOLD);
  uid->bold = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
					  GTK_TOOLBAR_CHILD_TOGGLEBUTTON,
					  /*element*/ NULL,
					  _("Bold"),
					  _("Toggle bold"),
					  NULL,
					  pixmap,
					  (GtkSignalFunc) bold_cb,
					  view);
  gtk_widget_show(uid->bold);

  /*pixmap = gnome_pixmap_new_from_xpm_d(gwp_italic);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_TEXT_ITALIC);
  uid->italic = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
					    GTK_TOOLBAR_CHILD_TOGGLEBUTTON,
					    /*element*/ NULL,
					    _("Italic"),
					    _("Toggle italic"),
					    NULL,
					    pixmap,
					    (GtkSignalFunc) italic_cb,
					    view);
  gtk_widget_show(uid->italic);

  /*pixmap = gnome_pixmap_new_from_xpm_d(gwp_underline);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_TEXT_UNDERLINE);
  uid->underline = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
					       GTK_TOOLBAR_CHILD_TOGGLEBUTTON,
					       /*element*/NULL,
					       _("Underline"),
					       _("Toggle underline"),
					       NULL,
					       pixmap,
					       (GtkSignalFunc) underline_cb,
					       view);
  gtk_widget_show(uid->underline);

  pixmap = gnome_pixmap_new_from_xpm_d (gwp_color);
  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("Color"),
				   _("Change Colors"),
				   NULL,
				   pixmap,
				   (GtkSignalFunc) change_color_cb,
				   view);
  gtk_widget_show(widget);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  view_and_line_alignment *vala_low, *vala_center, *vala_high;
  vala_low =
    (view_and_line_alignment *) malloc (sizeof (view_and_line_alignment));
  vala_center =
    (view_and_line_alignment *) malloc (sizeof (view_and_line_alignment));
  vala_high =
    (view_and_line_alignment *) malloc (sizeof (view_and_line_alignment));

  vala_low->view = view;
  vala_center->view = view;
  vala_high->view = view;
  vala_low->align = ALIGN_LOW;
  vala_center->align = ALIGN_CENTER;
  vala_high->align = ALIGN_HIGH;

  /*pixmap = gnome_pixmap_new_from_xpm_d (gwp_left);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_ALIGN_LEFT);
  element = gtk_radio_button_new (NULL);
  uid->left_align = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
						GTK_TOOLBAR_CHILD_RADIOBUTTON,
						element,
						_("Left"),
						_("Left Justify Line"),
						NULL,
						pixmap,
						(GtkSignalFunc) line_align_cb,
						vala_low);
  gtk_widget_show(uid->left_align);

  /*pixmap = gnome_pixmap_new_from_xpm_d (gwp_center);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_ALIGN_CENTER);
  element = gtk_radio_button_new_from_widget (GTK_RADIO_BUTTON (element));
  uid->center_align = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
					       GTK_TOOLBAR_CHILD_RADIOBUTTON,
						  element,
						  _("Center"),
						  _("Center Line"),
						  NULL,
						  pixmap,
						  (GtkSignalFunc)line_align_cb,
						  vala_center);
  gtk_widget_show(uid->center_align);

  /*pixmap = gnome_pixmap_new_from_xpm_d (gwp_right);*/
  pixmap = gnome_stock_pixmap_widget (da, GNOME_STOCK_PIXMAP_ALIGN_RIGHT);
  element = gtk_radio_button_new_from_widget (GTK_RADIO_BUTTON (element));
  uid->right_align = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
						 GTK_TOOLBAR_CHILD_RADIOBUTTON,
						 element,
						 _("Right"),
						 _("Right Justify Line"),
						 NULL,
						 pixmap,
						 (GtkSignalFunc) line_align_cb,
						 vala_high);
  gtk_widget_show (uid->right_align);

  gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

  uid->magnification = gtk_option_menu_new ();
  gtk_toolbar_append_widget (GTK_TOOLBAR(toolbar),
			     uid->magnification,
			     _("Select Magnification"), NULL);
  gtk_widget_show (uid->magnification);
  add_magnifications (view, uid->magnification);

  gtk_widget_show (uid->magnification);

  uid->table_popup = newTableUI (view);
  uid->list_popup = newListUI (view);
  uid->other_popup = newOtherUI (view);


  /* add a page setup menu entry under 'file' */

  GnomeUIInfo entry[2];
  entry[0].type = GNOME_APP_UI_ITEM;
  entry[0].label = "Page Setup...";
  entry[0].hint = NULL;
  entry[0].moreinfo = file_pagesetup_cmd_callback;
  entry[0].user_data = view;
  entry[0].unused_data = NULL;
  entry[0].pixmap_type = GNOME_APP_PIXMAP_STOCK;
  entry[0].pixmap_info = GNOME_STOCK_MENU_PREF;
  entry[0].accelerator_key = 0;
  entry[1].type = GNOME_APP_UI_ENDOFINFO;

  GnomeApp *app = gnome_mdi_get_app_from_view (GTK_WIDGET (view));

  int menu_pos;
  char menu_item_path[ strlen (entry[0].label) + 20 ];
  sprintf (menu_item_path, "File/%s", entry[0].label);
  if (gnome_app_find_menu_pos (app->menubar, menu_item_path, &menu_pos))
    return; /* already done */

  /*gnome_app_insert_menus_with_data (app, "File/", entry, view);*/
  /* insert the menu item after the Print item */
  gnome_app_insert_menus_with_data (app, "File/Print", entry, view);
}


static void deleteUI (WPView *view)
{
  /*WPView *view = wp_frame_get_focus_view(frame);*/
  /*WPFrame *frame = wp_view_get_frame (view);*/
  /*WPDocument *doc = wp_view_get_document(view);*/
  gwp_view_data *uid = get_plugin_view_data (view);

  gtk_widget_destroy (uid->hbox);

  deleteTableUI (view, uid->table_popup);
  deleteListUI (view, uid->table_popup);
  deleteOtherUI (view, uid->table_popup);
}


static void hideUI (WPView *view)
{
  /*WPView *view = wp_frame_get_focus_view(frame);*/
  /*WPDocument *doc = wp_view_get_document (view);*/
  gwp_view_data *uid = get_plugin_view_data (view);

  gtk_widget_hide (uid->hbox);
}


static void updateUI (WPView *view)
{
  /*WPView *view = wp_frame_get_focus_view(frame);*/
  /*WPDocument *doc = wp_view_get_document (view);*/
  gwp_view_data *uid = get_plugin_view_data (view);
  WPStyle* style;
  char *fontName;
  char *weight;
  char *slant;
  char *foundry, *family;


  if (wp_view_hl_is_on(view)) /* something is currently selected */
    {
      /* clear out all state */
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->bold), FALSE);
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->italic), FALSE);
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->underline), FALSE);
    }
  else /* just update the toolbar based on the point's style */
    {
      char buf[100];
      int index;
      int size;
      /*style = WP_GetStyleOfSlab(doc, wp_view_get_point(view));*/
      Spot point = wp_view_get_point (view);
      style = point->getSlab ()->getStyle ();
      uid->current_insert_style = style;

      fontName = style->getFontName();
      weight = getFontWeight(fontName);
      slant = getFontSlant(fontName);

      /* do font family */
      foundry = getFontFoundry(fontName);
      family = getFontFamily(fontName);

      g_snprintf(buf, sizeof(buf),
		 "%s-%s", foundry, family);

      index = (int)g_hash_table_lookup(uid->font_index_hash, buf);
      gtk_option_menu_set_history(GTK_OPTION_MENU(uid->fonts), index);
      free(foundry);
      free(family);

      /* do font size */
      size = getFontPointSize(fontName);
      g_snprintf(buf, sizeof(buf),
		 "%d", size/10);

      index = (int)g_hash_table_lookup(uid->pointsize_index_hash, buf);
      gtk_option_menu_set_history(GTK_OPTION_MENU(uid->pointsizes), index);

      /* do font weight */
      if (!strcmp(weight, "bold"))
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->bold), TRUE);
      else
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->bold), FALSE);
      free(weight);
      
      /* do font slant */
      if (!strcmp(slant, "i") || !strcmp(slant, "o"))
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->italic), TRUE);
      else
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->italic), FALSE);
      free(slant);

      /* do underline */
      if (style->getUnderline ())
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->underline), TRUE);
      else
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->underline), FALSE);
    }

  style = wp_view_get_point(view)->getSlab()->getParent()->getStyle();

  switch (style->getAlignment())
    {
    case ALIGN_LOW:
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->left_align),
				  TRUE);
      break;
    case ALIGN_CENTER:
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->center_align),
				  TRUE);
      break;
    case ALIGN_HIGH:
      gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(uid->right_align),
				  TRUE);
      break;
    }
}


static void showUI (WPView *view)
{
  /*WPDocument *doc = wp_view_get_document (view);*/
  gwp_view_data *uid = get_plugin_view_data (view);

  if (!uid->hbox)
    newUI (view);

  updateUI (view);

  gtk_widget_show (uid->hbox);
}


/************************************************************/
/*                 plugin stuff                             */
/************************************************************/



extern "C"
{
  int wp_plugin_fill_vtable (PluginVTable *pvt);
  char *g_module_check_init (GModule *module);
}



int wp_plugin_fill_vtable (PluginVTable *pvt)
{
  pvt->get_file_extension = getFileExtension;
  pvt->get_mime_type = getMimeType;
  pvt->get_ui_name = getUIName;
  pvt->can_load = canLoad;
  pvt->can_save = canSave;
  pvt->can_new = canNew;
  pvt->load_document = loadDocument;
  pvt->save_document = saveDocument;
  pvt->new_document = newDocument;
  pvt->delete_child = deleteChild;
  pvt->insert_character = insertCharacter;
  pvt->replace_word = replaceWord;
  pvt->create_following_word = createFollowingWord;
  pvt->insert_slab = insertSlab;
  pvt->button_event = buttonEvent;
  pvt->new_ui = newUI;
  pvt->delete_ui = deleteUI;
  pvt->show_ui = showUI;
  pvt->hide_ui = hideUI;
  pvt->update_ui = updateUI;

  return 0;
}


char *g_module_check_init (GModule */*module*/)
{
  printf ("gwpIO -- g_module_check_init called\n");
  return NULL;
}
