
#include <gdk_imlib.h>
#include <gnome.h>
#include <libgnomeprint/gnome-print.h>
#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"

#include "wp_plugin.h"
#include "wp_word_slab.h"
#include "wp_plugin.h"
#include "wp_font.h"

#include "plugin_list_helper.h"

typedef struct
{
  WPView *view;
  Spot spot;
  WPStyleManager *sm;
} view_and_spot_and_sm;


/*#include "pixmaps/gwp-bullet0.xpm"*/



  /*
    .--------------cslab------------------------------------.
    |.-------------itemslab-------------------------------. |
    ||<CR0>.-bcslab-------. .-----scslab----------.       | |
    ||	   |.---bslab----.| | .-----lslab-------. |       | |
    ||	   || bwslab:O   || | | wslab:XXX       | |       | |
    ||     |`------------'| | `-----------------' |       | |
    ||     |              | | .----lslab--------. |       | |
    ||     `--------------' | | XXX             | |       | |
    ||                      | `-----------------' |       | |
    ||                      `---------------------'       | |
    |`----------------------------------------------------' |
    |.-------------itemslab-------------------------------. |
    ||<CR0>.-bcslab-------. .-----scslab----------.       | |
    ||	   |.---bslab----.| | .-----lslab-------. |       | |
    ||	   || bwslab:O   || | | wslab:XXX       | |       | |
    ||     |`------------'| | `-----------------' |       | |
    ||     |              | | .----lslab--------. |       | |
    ||     `--------------' | | XXX             | |       | |
    ||                      | `-----------------' |       | |
    ||                      `---------------------'       | |
    |`----------------------------------------------------' |
    `-------------------------------------------------------'
  */        



char *list_logical_type_to_string (list_logical_type a)
{
  switch (a)
    {
    case lt_list: return "List";
    case lt_list_item: return "ListItem";
    case lt_list_bullet_line: return "ListBulletLine";
    case lt_list_bullet: return "ListBullet";
    default: return NULL;
    }
}
list_logical_type string_to_list_logical_type (unsigned const char *s)
{
  if (!s) return lt_none;
  if (strcasecmp ((char *) s, "List") == 0)
    return lt_list;
  if (strcasecmp ((char *) s, "ListItem") == 0)
    return lt_list_item;
  if (strcasecmp ((char *) s, "ListBulletLine") == 0)
    return lt_list_bullet_line;
  if (strcasecmp ((char *) s, "ListBullet") == 0)
    return lt_list_bullet;
  return lt_none; /* none */
}



/* containing, but not the same as... */
Slab find_list_containing_slab (Slab slab, int *pos_ret)
{
  int pos;

  while (slab && slab->getParent ())
    {
      Slab pslab = slab->getParent ();
      pos = pslab->getIndexOfChild (slab);
      if (pos < 0)
	return NULL;
      if (pos_ret) (*pos_ret) = pos;

      slab = pslab;

      if (slab->getPluginData () == ((void *) lt_list))
	return slab;
    }

  return NULL;
}


/* return 1 if slab is editable content of a list item */
int slab_is_list_content (Slab slab)
{
  if (! slab) return 0;
  if (! slab->getParent ()) return 0;
  if (! slab->getParent ()->getParent ()) return 0;

  /* find the list */
  int pos;
  Slab listslab = find_list_containing_slab (slab, &pos);
  if (! listslab) return 0;

  /* find the item */
  Slab itemslab = (Slab) listslab->getChildAt (pos);
  if (! itemslab) return 0;

  /* find the column */
  Slab colslab = (Slab) itemslab->getChildAt (2);
  if (!colslab) return 0;

  /* don't edit the column */
  if (colslab == slab) return 0;

  /* anything inside the column can be edited */
  if (colslab->slabIsInside (slab)) return 1;

  /* otherwise... no */
  return 0;

  /*
  Slab pparent = slab->getParent ()->getParent ();
  if (! pparent->getParent ()) return 0;
  Slab ppparent = pparent->getParent ();

  if (ppparent->getPluginData () != (void *) lt_list_item) return 0;
  if (ppparent->getChildAt (2) != (void *) pparent) return 0;

  return 1;
  */
}


/* return 1 if slab is an image used as a list bullet */
int slab_is_list_bullet (Slab slab)
{
  if (slab->getPluginData () != (void *) lt_list_bullet)
    return 0;
  if (! slab->getParent ()) return 0;
  if (! slab->getParent ()->getParent ()) return 0;
  if (! slab->getParent ()->getParent ()->getParent ()) return 0;
  if (! slab->getParent ()->getParent ()->getParent ()->getParent ()) return 0;

  Slab parent = slab->getParent ();
  Slab ppparent = parent->getParent ()->getParent ();
  Slab pppparent = ppparent->getParent ();

  if (parent->getPluginData () != (void *) lt_list_bullet_line)
    return 0;
  if (ppparent->getPluginData () != (void *) lt_list_item)
    return 0;
  if (pppparent->getPluginData () != (void *) lt_list)
    return 0;

  return 1;
}



int list_item_is_empty (Slab s)
{
  while (s && s->getPluginData () != (void *) lt_list_item)
    s = s->getParent ();

  if (!s) return 1;

  Slab item_content = (Slab) s->getChildAt (2);
  if (!item_content) return 1;

  Slab leaf = item_content->firstLowestContainer ();
  if (!leaf) return 1;

  while (leaf && item_content->slabIsInside (leaf))
    {
      /* look for anything visible in the slab (a word
	 of length > 0, or a picture) */
      if ((leaf->classID () == WordSlabID &&
	   leaf->getNumberOfChildren () > 0) ||
	  (leaf->classID () == ImageSlabID))
	return 0;
      leaf = leaf->getNextLeaf ();
    }

  return 1;
}


char *list_bullet_asciidraw (WPDocument */*doc*/,
			     ContSlab slab,
			     int /*x*/, int /*y*/)
{
  int nested = 0;
  Slab s = slab;

  while (s)
    {
      s = find_list_containing_slab (s, NULL);
      nested ++;
    }

  int nestedmod = nested % 3;
  char buf[ 2 * nested + 10 ];

  buf[ 0 ] = '\0';
  for (int i=0; i<nested; i++)
    strcat (buf, "  ");

  if (nestedmod == 0)
    strcat (buf, ". ");
  else if (nestedmod == 1)
    strcat (buf, "- ");
  else
    strcat (buf, "* ");

  return strdup (buf);
}


static void list_bullet_psdraw (WPDocument */*doc*/,
				GnomePrintContext *pc,
				ContSlab slab,
				int x, int y)
{
  /* XXX these still come out really ugly.  someone fix this */

  /*

       P--a--b--c--Q  
       |    ___    |
       l   / _ \   d
       k  | |_| |  e
       j   \___/   f
       |           |
       S--i--h--g--R
		    
  */

  /* ew */

  /* corners */
  int Px = x - slab->getMLeft ();
  int Py = y - slab->getMUp ();
  int Qx = x + slab->getMRight ();
  int Qy = y - slab->getMUp ();
  int Rx = x + slab->getMRight ();
  int Ry = y + slab->getMDown ();
  int Sx = x - slab->getMLeft ();
  int Sy = y + slab->getMDown ();

  /* move it to the upper right corner of it's original space */
  Px += ((Qx - Px) / 2);
  Sx = Px;
  Py += ((Sy-Py) / 2);
  Qy = Py;

  /* centers */
  int bx = (Px+Qx)/2;
  int by = Py;
  int ex = Qx;
  int ey = (Qy+Ry)/2;
  int hx = (Px+Qx)/2;
  int hy = Ry;
  int kx = Sx;
  int ky = (Py+Sy)/2;

  /* the rest */
  int ax = (Px+bx)/2;
  int ay = Py;
  int cx = (bx+Qx)/2;
  int cy = Qy;
  int dx = Qx;
  int dy = (Qy+ey)/2;
  int fx = Rx;
  int fy = (ey+Ry)/2;
  int gx = (hx+Rx)/2;
  int gy = Ry;
  int ix = (Sx+hx)/2;
  int iy = hy;
  int jx = Sx;
  int jy = (ky+Sy)/2;
  int lx = Px;
  int ly = (Py+ky)/2;

  double matrix2[ 6 ] = {1, 0, 0, 1, 0, 0};

  gnome_print_concat (pc, matrix2);

  gnome_print_moveto (pc, bx, by);
  gnome_print_setrgbcolor (pc, 0, 0, 0.5);
  gnome_print_curveto (pc, cx, cy, dx, dy, ex, ey);
  gnome_print_curveto (pc, fx, fy, gx, gy, hx, hy);
  gnome_print_curveto (pc, ix, iy, jx, jy, kx, ky);
  gnome_print_curveto (pc, lx, ly, ax, ay, bx, by);

  gnome_print_stroke (pc);
}


static void list_bullet_xdraw (WPView *view, ContSlab slab, int x, int y)
{
  WPStyle *style = slab->getStyle ();

  int x0 = x - slab->getMLeft ();
  int y0 = y - slab->getMUp ();
  int x3 = x + slab->getMRight () /*- 1*/;
  int y3 = y + slab->getMDown () /*- 1*/;

  GdkColor *c = style->getForeground ();
  /*GdkFont *f = style->getFont (scale);*/

  wp_view_set_color (view, c);

  int nested = 0;
  Slab s = slab;

  while (s)
    {
      s = find_list_containing_slab (s, NULL);
      nested ++;
    }

  nested = nested % 3;

  if (nested == 0)
    {
      wp_view_draw_line (view, x0, y0, x3, y0);
      wp_view_draw_line (view, x3, y0, x3, y3);
      wp_view_draw_line (view, x3, y3, x0, y3);
      wp_view_draw_line (view, x0, y3, x0, y0);
    }

  if (nested == 1)
    {
      wp_view_draw_line (view, x0, y0, x3, y3);
      wp_view_draw_line (view, x3, y0, x0, y3);
    }

  if (nested == 2)
    wp_view_draw_arc (view,
		      1, /* filled */
		      x0, y0,
		      x3-x0, y3-y0,
		      0, /* angle1 */
		      360*64 /* angle2 */);
}


Slab new_list_item_bullet (WPStyleManager *sm, WPStyle *style, int /*debug*/)
{
  Slab s;
  style_group *wsg = sm->getStyleGroupByName ("word");

  WPStyle *bullet_style = sm->findOrCreateStyle
    (wsg,
     style->getFontName (),
     style->getForeground (), style->getBackground (), 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED, SIZE_POLICY_FIXED,
     0, 10, 10, 0,
     0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  s = new _ContSlab (bullet_style,
		     list_bullet_psdraw,
		     list_bullet_xdraw,
		     list_bullet_asciidraw);
  s->setPluginData ((void *) lt_list_bullet);

  return s;
}


Slab new_item_slab (WPStyleManager *sm, WPStyle *style,
		    int item_width, int debug)
{
  style_group *line_sg = sm->getStyleGroupByName ("line");

  WPStyle *item_style;
  WPStyle *bullet_style;
  /*WPStyle *word_style;*/

  item_style =
    sm->findOrCreateStyle (line_sg,
			   style->getFontName (),
			   style->getForeground (),
			   style->getBackground (),
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0, 0,
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE,
			   0);

  bullet_style =
    sm->findOrCreateStyle (line_sg,
			   style->getFontName (),
			   style->getForeground (),
			   style->getBackground (),
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_FIXED, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 12, 0, 0, 0, 0, 0,
			   0,
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE,
			   0);


  WPStyle *list_line0_style = 
    sm->createStyle (sm->getStyleGroupByName ("line"),
		     style->getFontName (),
		     style->getForeground (),
		     style->getBackground (),
		     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
		     SIZE_POLICY_BY_CHILDREN, 0, 0,

		     /*s->getFixedRight ()*/ item_width - 80, /* XXX */
		     /* 80 should be the width of the bullet image */

		     0, 0, 0, 0, 0, 0,
		     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  WPStyle *list_col0_style =
    sm->createStyle (sm->getStyleGroupByName ("column"),
		     style->getFontName (),
		     style->getForeground (),
		     style->getBackground (),
		     ALIGN_LOW, DIR_DOWN,
		     SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
		     0, 0, 0, 0, 5, 5, 5, 5, 0,
		     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);
 
  WPStyle *list_col1_style =
    sm->createStyle (sm->getStyleGroupByName ("column"),
		     style->getFontName (),
		     style->getForeground (),
		     style->getBackground (),
		     ALIGN_LOW, DIR_DOWN,
		     SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
		     0, 0, 0, 0, 5, 5, 5, 5, 0,
		     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  ContSlab itemslab = new _ContSlab (item_style,
				     NULL, NULL, NULL); /* item line */
  itemslab->setPluginData ((void *) lt_list_item);
  LBTSlab cr0slab = new _LBTSlab (style, 2, 0);
  /*cr0slab->setHasSize (0);*/
  ContSlab bslab = new _ContSlab (bullet_style,
				  NULL, NULL, NULL); /* bullet line */
  bslab->setPluginData ((void *) lt_list_bullet_line);
  ContSlab bcslab = new _ContSlab (list_col0_style,
				   NULL, NULL, NULL); /* bullet column */
  ContSlab scslab = new _ContSlab (list_col1_style,
				   NULL, NULL, NULL); /* sub column */
  ContSlab lslab = new _ContSlab (list_line0_style,
				  NULL, NULL, NULL); /* line */
  /*WordSlab bwslab = new _WordSlab (style, "O", 1);*/ /* bullet word */

# if 0
  GtkWidget *pixmapw = gnome_pixmap_new_from_xpm_d(gwp_bullet0);
  GnomePixmap *gpixmap = GNOME_PIXMAP (pixmapw);
  ImageSlab bwslab = new _ImageSlab (style,
				     gpixmap->pixmap,
				     NULL); /*bullet word*/
# endif /* 0 */

# if 0
  GdkImlibImage *img;
  img = gdk_imlib_create_image_from_xpm_data (gwp_bullet0);
  ImageSlab bwslab = new _ImageSlab (style,
				     img,
				     NULL); /*bullet word*/

  bwslab->setPluginData ((void *) lt_list_bullet);
# endif /* 0 */

  Slab bwslab = new_list_item_bullet (sm, style, debug);

  WordSlab wslab = new _WordSlab (style, "", 0); /*word */

  if (lslab->insertChild (0, wslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial word failed.\n");
  if (bslab->insertChild (0, bwslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial bullet failed.\n");
  if (scslab->insertChild (0, lslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial line failed.\n");
  if (itemslab->insertChild (0, cr0slab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of cr0 failed.\n");
  if (bcslab->insertChild (0, bslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial bullet line failed.\n");
  if (itemslab->insertChild (1, bcslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial bullet column failed.\n");
  if (itemslab->insertChild (2, scslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial sub column failed.\n");

  return itemslab;
}


int delete_list_item (WPDocument *doc, Spot spot)
{
  Slab slab = spot->getSlab ();

  Slab itemslab = slab->getParent ()->getParent ()->getParent ();
  Slab listslab = itemslab->getParent ();
  Spot delspot = doc->newDocumentSpot ();
  delspot->setSlab (listslab);
  delspot->setPosition (listslab->getIndexOfChild (itemslab));
  int ret = doc->deleteChild (delspot);
  doc->deleteDocumentSpot (delspot);
  spot->sink (); /* put the cursor somewhere sane */
  return ret;
}


void insert_new_list_item (WPStyleManager *sm,
			   WPStyle *style, Spot spot,
			   int after_current)
{
  int debug = -1;

  /* find the list that contains the point (if there is one) */
  int pos;
  Slab cslab = find_list_containing_slab (spot->getSlab (), &pos);
  if (cslab == NULL)
    {
      GtkWidget *dialog = gnome_error_dialog(_("Not in a List"));
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run(GNOME_DIALOG(dialog));
      return;
    }

  /*fprintf (stderr, "==>insert_new_list_item: pos=%d\n", pos);*/

  if (! cslab->getParent ())
    return;
  int item_width = cslab->getParent ()->getWidth ();
  Slab itemslab = new_item_slab (sm, style, item_width, debug);

  /* insert after the point, rather than before it */
  if (after_current && pos < cslab->getNumberOfChildren ())
    pos ++;

  if (cslab->insertChild (pos, itemslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of new item line failed.\n");
  else
    {
      spot->setPosition (0);
      spot->setSlab (itemslab->lastLowestContainer ());
    }
}


void insert_new_list (WPDocument *doc,
		      WPStyleManager *sm,
		      WPStyle *style,
		      Spot spot)
{
  if (doc == NULL) return;
  if (sm == NULL) return;
  if (spot == NULL) return;

  int debug = -1;

  style_group *column_sg = sm->getStyleGroupByName ("column");
  WPStyle *column_style;

  column_style =
    sm->findOrCreateStyle (column_sg,
			   style->getFontName (),
			   style->getForeground (),
			   style->getBackground (),
			   /*s->getAlignment ()*/ ALIGN_LOW,
			   /*s->getDirection ()*/ DIR_DOWN,
			   /*SIZE_POLICY_FIXED*/ SIZE_POLICY_BY_CHILDREN,
			   /*SIZE_POLICY_FIXED*/ SIZE_POLICY_BY_CHILDREN,
			   /*s->getFixedLeft ()*/ 0,
			   /*s->getFixedUp ()*/ 0,
			   /*s->getFixedRight ()*/ 0,
			   /*s->getFixedDown ()*/ 0,
			   /*s->getMarginLeft ()*/ /*5*/ 0,
			   /*s->getMarginUp ()*/ /*5*/ 0,
			   /*s->getMarginRight ()*/ /*5*/ 0,
			   /*s->getMarginDown ()*/ /*5*/ 0,
			   0,
			   MARGIN_RENDERING_POLICY_BLANK,
			   WHEN_EMPTY_REMOVE, 0);

  ContSlab cslab = new _ContSlab (column_style, NULL, NULL, NULL); /* list */
  cslab->setPluginData ((void *) lt_list);

  if (! spot->getSlab ()->getParent ())
    return;
  int item_width = spot->getSlab ()->getParent ()->getWidth ();

  Slab itemslab = new_item_slab (sm, style, item_width, debug);

  if (cslab->insertChild (0, itemslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial item line failed.\n");

  doc->insertSlab (spot, cslab);

  /* make sure we don't trap the cursor in the list */
  if (cslab->getPrevLeaf () == NULL)
    {
      spot->setPosition (0);
      printf ("inserting empty word between list and start\n");

      /*Slab ws = new _WordSlab (style, "", 0);*/
      Slab lb = new _LBTSlab (style, 2, 0);
      doc->insertSlab (spot, lb);
    }
  if (cslab->getNextLeaf () == NULL)
    {
      spot->setSlab (cslab);
      spot->setPosition (1);
      printf ("inserting empty word between list and end\n");

      /*Slab ws = new _WordSlab (style, "", 0);*/
      Slab lb = new _LBTSlab (style, 2, 0);
      doc->insertSlab (spot, lb);
    }

  spot->setPosition (0);
  spot->setSlab (itemslab->lastLowestContainer ());
}



static void list_insert_item_callback (GtkWidget *, gpointer client_data)
{
  view_and_spot_and_sm *vas = (view_and_spot_and_sm *) client_data;
  if (!vas) return;
  if (!vas->view) return;

  printf ("list insert item\n");

  WPDocument *doc = wp_view_get_document (vas->view);
  if (!doc) return;

  printf ("  doc is '%s'\n", doc->getDocumentName ());

  /* find the list */
  int pos;
  Slab listslab = find_list_containing_slab (vas->spot->getSlab (), &pos);
  if (! listslab) return;

  /* find the item */
  Slab itemslab = (Slab) listslab->getChildAt (pos);
  if (! itemslab) return;

  /* find the column */
  Slab colslab = (Slab) itemslab->getChildAt (2);
  if (! colslab) return;

  /* find a leaf */
  Slab leaf = colslab->firstLowestContainer ();
  if (! leaf) return;

  /* use the leaf's style */
  WPStyle *style = leaf->getStyle ();

  insert_new_list_item (vas->sm, style, vas->spot, 0);
  doc->refresh ();

  free (vas); /* XXX free here? */
}


static void list_delete_item_callback (GtkWidget *, gpointer client_data)
{
  view_and_spot_and_sm *vas = (view_and_spot_and_sm *) client_data;
  if (!vas) return;
  if (!vas->view) return;
  if (!vas->spot) return;

  printf ("list delete item\n");

  WPDocument *doc = wp_view_get_document (vas->view);
  if (!doc) return;

  printf ("  doc is '%s'\n", doc->getDocumentName ());

  if (! delete_list_item (doc, vas->spot))
    printf ("delete list item failed?\n");

  doc->refresh ();

  free (vas); /* XXX free here? */
}


static GnomeUIInfo list_popup_info[] =
{
  { GNOME_APP_UI_ITEM, "Insert List Item", NULL,
    list_insert_item_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Delete List Item", NULL,
    list_delete_item_callback, NULL, NULL },

  { GNOME_APP_UI_ENDOFINFO }
};


GtkWidget *newListUI (WPView *view)
{
  GtkWidget *popup = gnome_popup_menu_new (list_popup_info);
  GtkWidget *da = wp_view_get_drawing_area (view);
  gnome_popup_menu_attach (popup, da, NULL);
  return popup;
}


void deleteListUI (WPView */*view*/, GtkWidget *popup)
{
  gtk_widget_destroy (popup);
}


void showListUI (WPView *view, Spot spot, WPStyleManager *sm,
		 GtkWidget *popup, GdkEventButton *geb)
{
  view_and_spot_and_sm *vas =
    (view_and_spot_and_sm *) malloc (sizeof (view_and_spot_and_sm));
  vas->view = view;
  vas->spot = spot;
  vas->sm = sm;
  gnome_popup_menu_do_popup (popup, NULL, NULL, geb, vas);
}
