
#include <gnome.h>
/*#include "wp_frame.h"*/
#include "wp_view.h"
#include "wp_document.h"
#include "wp_plugin_manager.h"

#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"

#include "wp_style.h"
#include "wp_plugin.h"

#include "plugin_paper_helper.h"


typedef struct
{
  WPView *view;
  GtkWidget *dialog;
  wp_page_type wppt;
  WPStyleManager *sm;
  GtkWidget *paper_selector;
} view_and_dialog;



char *wp_page_type_to_string (wp_page_type wppt)
{
  switch (wppt)
    {
    case wppt_roman: return "Roman";
    case wppt_roman_2_col: return "Roman-2-Column";
    case wppt_roman_3_col: return "Roman-3-Column";
    case wppt_asian: return "Asian";
    case wppt_arabic: return "Arabic";
    case wppt_scroll: return "Scroll";
    case wppt_default: return "Default";
    case wppt_unknown:
    default:
      return "Unknown";
    }
}


wp_page_type string_to_wp_page_type (unsigned const char *s)
{
  if (!s) return wppt_unknown;
  if (strcasecmp ((char *) s, "Roman") == 0) return wppt_roman;
  if (strcasecmp ((char *) s, "Roman-2-Column") == 0) return wppt_roman_2_col;
  if (strcasecmp ((char *) s, "Roman-3-Column") == 0) return wppt_roman_3_col;
  if (strcasecmp ((char *) s, "Asian") == 0) return wppt_asian;
  if (strcasecmp ((char *) s, "Arabic") == 0) return wppt_arabic;
  if (strcasecmp ((char *) s, "Scroll") == 0) return wppt_scroll;
  if (strcasecmp ((char *) s, "Default") == 0) return wppt_default;
  return wppt_unknown;
}


static void wp_page_setup_roman_styles (WPStyleManager *sm,
					WPStyle **ret,
					page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     /*614*/ psi->width - psi->leftM - psi->rightM, /* fixed right */
     0, /* fixed down */
     psi->leftM /*10*/, /* fixed margin left */
     0 /*10*/, /* fixed margin up */
     psi->rightM /*10*/, /* fixed margin right */
     0 /*10*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_FILL, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_FIXED,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     /*648*/psi->height - psi->upM - psi->downM, /* fixed down */
     0, /* fixed margin left */
     /*72*/ psi->upM, /* fixed margin up */
     0, /* fixed margin right */
     /*72*/ psi->downM, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     /*469*/ psi->width - psi->leftM - psi->rightM - 1, /* fixed right */
     0, /* fixed down */
     0 /*72*/ /*psi->leftM*/, /* fixed margin left */
     0, /* fixed margin up */
     0 /*72*/ /*psi->rightM*/, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;
}


static void wp_page_setup_roman_2_styles (WPStyleManager *sm,
					  WPStyle **ret,
					  page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);


  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     /*614 */ psi->width - psi->leftM - psi->rightM, /* fixed right */
     0, /* fixed down */
     /*10*/ psi->leftM, /* fixed margin left */
     0 /*psi->upM*/, /* fixed margin up */
     /*10*/ psi->rightM, /* fixed margin right */
     0 /*psi->downM*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_FIXED,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     /*648*/ psi->height - psi->upM - psi->downM, /* fixed down */
     0, /* fixed margin left */
     psi->upM /*72*/, /* fixed margin up */
     0, /* fixed margin right */
     psi->downM /*72*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  int column_width = (psi->width - psi->leftM - psi->rightM - 1) / 2;

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     /*226*/ column_width - 18, /* fixed right */
     0, /* fixed down */
     0, /* fixed margin left */
     0, /* fixed margin up */
     18, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;

  /*sm->setDefaultStyleOfGroup ("word", word_0_style);*/
}


static void wp_page_setup_roman_3_styles (WPStyleManager *sm,
					  WPStyle **ret,
					  page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);


  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     psi->width - psi->leftM - psi->rightM /*614*/, /* fixed right */
     0, /* fixed down */
     psi->leftM /*10*/, /* fixed margin left */
     0 /*10*/, /* fixed margin up */
     psi->rightM /*10*/, /* fixed margin right */
     0 /*10*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_FIXED,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     psi->height - psi->upM - psi->downM /*648*/, /* fixed down */
     0, /* fixed margin left */
     psi->upM /*72*/, /* fixed margin up */
     0, /* fixed margin right */
     psi->downM /*72*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  int column_width = (psi->width - psi->leftM - psi->rightM - 1) / 3;

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     column_width - 18 /*150*/, /* fixed right */
     0, /* fixed down */
     0, /* fixed margin left */
     0, /* fixed margin up */
     18, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;
}


static void wp_page_setup_asian_styles (WPStyleManager *sm,
					WPStyle **ret,
					page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);


  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     psi->width /*614*/, /* fixed right */ /* XXX */
     0, /* fixed down */
     10, /* fixed margin left */
     10, /* fixed margin up */
     10, /* fixed margin right */
     10, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_FILL, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_LEFT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     psi->width - psi->leftM - psi->rightM - 1 /*469*/, /* fixed right */
     0, /* fixed down */
     psi->leftM /*72*/, /* fixed margin left */
     0, /* fixed margin up */
     psi->rightM /*72*/, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_FIXED,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     psi->height /*648*/, /* fixed down */
     10, /* fixed margin left */
     psi->upM /*72*/, /* fixed margin up */
     10, /* fixed margin right */
     psi->downM /*72*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0,
     0, 2, 0, 2, /* slight up and down margins to keep the glyphs apart */
     2,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;
}


static void wp_page_setup_arabic_styles (WPStyleManager *sm,
					 WPStyle **ret,
					 page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);

  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_LEFT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     psi->width /*614*/, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */ /* XXX */
     0, /* fixed down */
     10, /* fixed margin left */
     10, /* fixed margin up */
     10, /* fixed margin right */
     10, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_FILL, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_FIXED,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     psi->height - psi->upM - psi->downM /*648*/, /* fixed down */
     0, /* fixed margin left */
     psi->upM /*72*/, /* fixed margin up */
     0, /* fixed margin right */
     psi->downM /*72*/, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_LEFT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     psi->width - psi->leftM - psi->rightM - 1 /*469*/, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     0, /* fixed down */
     psi->leftM /*72*/, /* fixed margin left */
     0, /* fixed margin up */
     psi->rightM /*72*/, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_LEFT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;

  /*sm->setDefaultStyleOfGroup ("word", word_0_style);*/
}


static void wp_page_setup_scroll_styles (WPStyleManager *sm,
					 WPStyle **ret,
					 page_size_info *psi)
{
  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);

  style_group *dsg = sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->findOrCreateStyle 
    (dsg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  style_group *psg = sm->newStyleGroup ("page");
  WPStyle *page_0_style = sm->findOrCreateStyle 
    (psg, "", fg, bg, ALIGN_LOW, DIR_LEFT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     psi->width /*614*/, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */ /* XXX */
     0, /* fixed down */
     10, /* fixed margin left */
     10, /* fixed margin up */
     10, /* fixed margin right */
     10, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_FILL, WHEN_EMPTY_REMOVE, 0);

  style_group *csg = sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->findOrCreateStyle 
    (csg, "", fg, bg, ALIGN_LOW, DIR_DOWN, SIZE_POLICY_BY_CHILDREN,
     SIZE_POLICY_BY_CHILDREN,
     0, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     0, /* fixed down */
     0, /* fixed margin left */
     0, /* fixed margin up */
     0, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *lsg = sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->findOrCreateStyle
    (lsg, "", fg, bg, ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_FIXED,
     SIZE_POLICY_BY_CHILDREN,
     psi->width - psi->leftM - psi->rightM - 1 /*469*/, /* fixed left */
     0, /* fixed up */
     0, /* fixed right */
     0, /* fixed down */
     psi->leftM /*72*/, /* fixed margin left */
     0, /* fixed margin up */
     psi->rightM /*72*/, /* fixed margin right */
     0, /* fixed margin down */
     0, /* max children */
     MARGIN_RENDERING_POLICY_OUTLINE, WHEN_EMPTY_REMOVE, 0);

  style_group *wsg = sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->findOrCreateStyle
    (wsg, "-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*", fg, bg, 
     ALIGN_LOW, DIR_RIGHT, SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
     0, 0, 0, 0, 0, 0, 0, 0, 0,
     MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_REMOVE, 0);

  ret[ 0 ] = document_0_style;
  ret[ 1 ] = page_0_style;
  ret[ 2 ] = column_0_style;
  ret[ 3 ] = line_ltr_low_style;
  ret[ 4 ] = word_0_style;

  /*sm->setDefaultStyleOfGroup ("word", word_0_style);*/
}



void wp_page_setup_styles (wp_page_type wppt,
			   WPStyleManager *sm,
			   WPStyle **ret,
			   page_size_info *psi)
{
  WPStyle *styles[ 5 ];

  switch (wppt)
    {
    case wppt_roman:
      wp_page_setup_roman_styles (sm, styles, psi); break;
    case wppt_roman_2_col:
      wp_page_setup_roman_2_styles (sm, styles, psi); break;
    case wppt_roman_3_col:
      wp_page_setup_roman_3_styles (sm, styles, psi); break;
    case wppt_asian:
      wp_page_setup_asian_styles (sm, styles, psi); break;
    case wppt_arabic:
      wp_page_setup_arabic_styles (sm, styles, psi); break;
    case wppt_scroll:
      wp_page_setup_scroll_styles (sm, styles, psi); break;
    case wppt_unknown:
      return;
    default:
      wp_page_setup_roman_styles (sm, styles, psi);
      break;
    }

  if (ret)
    {
      ret[ 0 ] = styles[ 0 ];
      ret[ 1 ] = styles[ 1 ];
      ret[ 2 ] = styles[ 2 ];
      ret[ 3 ] = styles[ 3 ];
      ret[ 4 ] = styles[ 4 ];
    }
}


Slab wp_page_build (wp_page_type wppt,
		    WPStyleManager *sm,
		    page_size_info *psi)
{
  int debug = -1;
  WPStyle *styles[ 5 ];

  wp_page_setup_styles (wppt, sm, styles, psi);

  WPStyle *document_style = styles[ 0 ];
  WPStyle *page_style = styles[ 1 ];
  WPStyle *column_style = styles[ 2 ];
  WPStyle *line_style = styles[ 3 ];
  WPStyle *word_style = styles[ 4 ];

  ContSlab dslab = new _ContSlab (document_style, NULL, NULL, NULL);
  ContSlab pslab = new _ContSlab (page_style, NULL, NULL, NULL);
  ContSlab cslab = new _ContSlab (column_style, NULL, NULL, NULL);
  ContSlab lslab = new _ContSlab (line_style, NULL, NULL, NULL);
  WordSlab wslab = new _WordSlab (word_style, "", 0);

  /* spots and views are NULL at this point (NULL = empty glist). */
  if (dslab->insertChild (0, pslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial page failed.\n");
  if (pslab->insertChild (0, cslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial col failed.\n");
  if (cslab->insertChild (0, lslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial line failed.\n");
  if (lslab->insertChild (0, wslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial word failed.\n");

  return dslab;
}


int wp_page_convert_to (wp_page_type wppt,
			WPStyleManager *sm,
			page_size_info *psi,
			WPDocument *doc)
{
  WPStyle *styles[ 5 ];


  wp_page_setup_styles (wppt, sm, styles, psi);

  WPStyle *document_style = styles[ 0 ];
  WPStyle *page_style = styles[ 1 ];
  WPStyle *column_style = styles[ 2 ];
  WPStyle *line_style = styles[ 3 ];
  WPStyle *word_style = styles[ 4 ];

  doc->setStyles (sm,
		  document_style,
		  page_style,
		  column_style,
		  line_style,
		  word_style);

  doc->refresh ();

  return 1;
}


static void
wp_page_setup_apply (view_and_dialog *vad)
{
  if (!vad) return;

  /*WPView *view = vad->view;*/
  printf ("wp_page_setup_apply: type='%s'\n",
	  wp_page_type_to_string (vad->wppt));

  /*GtkWidget *paper_selector = vad->paper_selector;*/
  GnomePaperSelector *paper_selector =
    GNOME_PAPER_SELECTOR (vad->paper_selector);

  page_size_info psi;

  psi.width = /* XXX -- use converter */
    (int) (gnome_paper_selector_get_width (paper_selector) * 72.2);
  psi.height =
    (int) (gnome_paper_selector_get_height (paper_selector) * 72.2);

  /* so... uh... these went away...? */
  /*
  psi.leftM =
    (int) (gnome_paper_selector_get_left_margin (paper_selector) * 72.2);
  psi.upM =
    (int) (gnome_paper_selector_get_top_margin (paper_selector) * 72.2);
  psi.rightM =
    (int) (gnome_paper_selector_get_right_margin (paper_selector) * 72.2);
  psi.downM =
    (int) (gnome_paper_selector_get_bottom_margin (paper_selector) * 72.2);
  */

  /* XXX */
  psi.leftM = 72;
  psi.upM = 72;
  psi.rightM = 72;
  psi.downM = 72;

  printf ("paper size: %dx%d  (%d, %d, %d, %d)\n",
	  psi.width, psi.height, psi.leftM, psi.upM, psi.rightM, psi.downM);

  /* make sure the margins fit in the page */
  if (psi.width < psi.leftM + psi.rightM)
    psi.width = psi.leftM + psi.rightM + 1;
  if (psi.height < psi.upM + psi.downM)
    psi.height = psi.upM + psi.downM + 1;

  wp_page_convert_to (vad->wppt, vad->sm,
		      &psi,
		      wp_view_get_document (vad->view));
}


static void
wp_page_setup_ok_callback (GtkWidget */*w*/,
			   gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  GtkWidget *properties_dialog = vad->dialog;
  gtk_widget_hide (properties_dialog);
  wp_page_setup_apply (vad);
}


static void
wp_page_setup_apply_callback (GtkWidget */*w*/,
			      gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  wp_page_setup_apply (vad);
}


static void
wp_page_setup_cancel_callback (GtkWidget */*w*/,
				 gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  /*WPView *view = vad->view;*/
  GtkWidget *properties_dialog = vad->dialog;
  gtk_widget_hide (properties_dialog);
}


static void
page_type_roman_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_roman;
}

static void
page_type_roman_2_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_roman_2_col;
}

static void
page_type_roman_3_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_roman_3_col;
}

static void
page_type_asian_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_asian;
}

static void
page_type_arabic_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_arabic;
}

static void
page_type_scroll_cb (GtkWidget */*widget*/, gpointer client_data)
{
  view_and_dialog *vad = (view_and_dialog *) client_data;
  vad->wppt = wppt_scroll;
}


GtkWidget *
wp_new_page_setup_window (WPView *view, WPStyleManager *sm)
{
  GtkWidget *properties_dialog = gtk_dialog_new();
  GtkWidget *notebook = gtk_notebook_new();
  GtkWidget *button;
  view_and_dialog *vad = (view_and_dialog *) malloc (sizeof (view_and_dialog));

  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) wp_page_setup_ok_callback,
                      vad);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Apply");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) wp_page_setup_apply_callback,
                      vad);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) wp_page_setup_cancel_callback,
                      vad);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_show (button);


  {
    GtkWidget *frame;
    GtkWidget *label;

    frame = gtk_frame_new ("Page Layout");
    gtk_container_border_width (GTK_CONTAINER (frame), 10);
    /*gtk_widget_set_usize (frame, 200, 150);*/
    gtk_widget_show (frame);

    GtkWidget *page_type = gtk_option_menu_new ();
    gtk_container_add (GTK_CONTAINER (frame), page_type);
    gtk_widget_show (page_type);

    GtkWidget *menu = gtk_menu_new();

    vad->wppt = wppt_roman;

    label = gtk_menu_item_new_with_label ("roman");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_roman_cb), vad);

    label = gtk_menu_item_new_with_label ("roman 2 column");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_roman_2_cb), vad);

    label = gtk_menu_item_new_with_label ("roman 3 column");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_roman_3_cb), vad);

    label = gtk_menu_item_new_with_label ("asian");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_asian_cb), vad);

    label = gtk_menu_item_new_with_label ("arabic");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_arabic_cb), vad);

    label = gtk_menu_item_new_with_label ("scroll");
    gtk_menu_append (GTK_MENU (menu), label); gtk_widget_show (label);
    gtk_signal_connect (GTK_OBJECT (label),
			"activate",
			GTK_SIGNAL_FUNC (page_type_scroll_cb), vad);

    gtk_option_menu_set_menu (GTK_OPTION_MENU (page_type), menu);

    label = gtk_label_new ("Page Layout");
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);
  }


  {
    GtkWidget *frame;
    GtkWidget *label;
    GtkWidget *paper_selector;

    frame = gtk_frame_new ("paper size");
    gtk_container_border_width (GTK_CONTAINER (frame), 10);
    /*gtk_widget_set_usize (frame, 200, 150);*/
    gtk_widget_show (frame);

    paper_selector = gnome_paper_selector_new ();
    gtk_container_add (GTK_CONTAINER (frame), paper_selector);
    gtk_widget_show (paper_selector);
	  
    label = gtk_label_new ("Paper Size");
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);

    vad->paper_selector = paper_selector;
  }


  gtk_box_pack_start (GTK_BOX(GTK_DIALOG (properties_dialog)->vbox),
		      notebook, TRUE, TRUE, 0);
  gtk_widget_show (notebook);

  vad->view = view;
  vad->dialog = properties_dialog;
  vad->sm = sm;

  return properties_dialog;
}


  /*
  (define paper-types
    (list
     (list "US Letter" 612 792 72 72 72 72) ; 8.5" x 11"
     (list "US Legal" 612 1008 72 72 72 72) ; 8.5" x 14"
     (list "A3" 843 1192 80 80 80 80)  ; 297mm x 420mm
     (list "A4" 596 843 80 80 80 80) ; 210mm x 297mm
     (list "A5" 422 596 80 80 80 80) ; 148mm x 210mm
     (list "B4" 730 1034 60 60 60 60) ; 257mm x 364mm
     (list "B5" 500 710 80 80 80 80) ; 176mm x 250mm
     (list "B5 Japan" 517 730 80 80 80 80) ; 182mm x 257mm
     (list "Half Letter" 612 397 72 72 72 72) ; 5.5" x 8.5"
     (list "Executive" 523 758 72 72 72 72) ; 8.25" x 10.5"
     (list "Tabloid/Ledger" 794 1227 72 72 72 72) ; 11" x 17"
     (list "Monarch" 280 542 10 10 10 10) ; 3.875" x 7.5"
     (list "SuperB" 843 1227 80 80 80 80) ; 297mm x 432mm
     (list "Envelope Commercial #10" 298 686 5 5 5 5) ; 4.125" x 9.5"
     (list "Envelope Monarch #7 3/4" 280 542 5 5 5 5) ; 3.875" x 7.5"
     (list "Envelope DL" 312 625 5 5 5 5) ; 110mm x 220mm
     (list "Envelope C5" 460 650 5 5 5 5) ; 162mm x 229mm
     (list "EuroPostcard " 298 420 5 5 5 5) ; 105mm x 148mm
     ))
  */



#if 0
  int debug = -1;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5;
  bg->green = 0xA5A5;
  bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);

  /* black */
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000;
  fg->green = 0x0000;
  fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  /* blue */
  GdkColor *color1 = (GdkColor *) malloc (sizeof (GdkColor));
  color1->red = 0x0000;
  color1->green = 0x0000;
  color1->blue = 0xffff;
  gdk_color_alloc (colormap, color1);

  /* red */
  GdkColor *color2 = (GdkColor *) malloc (sizeof (GdkColor));
  color2->red = 0xffff;
  color2->green = 0x0000;
  color2->blue = 0x0000;
  gdk_color_alloc (colormap, color2);


  sm->newStyleGroup ("document");
  WPStyle *document_0_style = sm->newStyle ("document", "default");
# if 0
  WPStyle *document_1_style = sm->newStyle ("document", "asian");
# endif

  document_0_style->setAlignment (ALIGN_LOW);
  document_0_style->setDirection (DIR_DOWN);
  document_0_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  document_0_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  document_0_style->setBackground (bg);
  document_0_style->setForeground (fg);
# if 0
  document_1_style->setAlignment (ALIGN_LOW);
  document_1_style->setDirection (DIR_LEFT);
  document_1_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  document_1_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  document_1_style->setForeground (fg);
# endif


  sm->newStyleGroup ("page");

  WPStyle *page_0_style = sm->newStyle ("page", "letter");
  page_0_style->setAlignment (ALIGN_LOW);
  page_0_style->setDirection (DIR_RIGHT);
  page_0_style->setXSizePolicy (SIZE_POLICY_FIXED);
  page_0_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  page_0_style->setFixedRight (614);
  page_0_style->setBackground (bg);
  page_0_style->setForeground (fg);
  page_0_style->setMarginLeft (20);
  page_0_style->setMarginUp (20);
  page_0_style->setMarginRight (20);
  page_0_style->setMarginDown (20);
  page_0_style->setMarginRendering (MARGIN_RENDERING_POLICY_FILL);


  WPStyle *page_1_style = sm->newStyle ("page", "debug");
  page_1_style->setAlignment (ALIGN_LOW);
  /*page_1_style->setDirection (DIR_DOWN);*/
  page_1_style->setDirection (DIR_RIGHT);
  page_1_style->setXSizePolicy (SIZE_POLICY_FIXED);
  page_1_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  page_1_style->setFixedRight (300);
  page_1_style->setBackground (bg);
  page_1_style->setForeground (fg);
  page_1_style->setMarginLeft (20);
  page_1_style->setMarginUp (20);
  page_1_style->setMarginRight (20);
  page_1_style->setMarginDown (20);


  sm->newStyleGroup ("column");
  WPStyle *column_0_style = sm->newStyle ("column", "letter");
  column_0_style->setAlignment (ALIGN_LOW);
  column_0_style->setDirection (DIR_DOWN);
  column_0_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  column_0_style->setYSizePolicy (SIZE_POLICY_FIXED);
  column_0_style->setMarginUp (72);
  column_0_style->setMarginDown (72);
  /*column_0_style->setFixedDown (794);*/
  column_0_style->setFixedDown (648);
  column_0_style->setBackground (bg);
  column_0_style->setForeground (fg);

  WPStyle *column_1_style = sm->newStyle ("column", "asian");
  column_1_style->setAlignment (ALIGN_LOW);
  column_1_style->setDirection (DIR_LEFT);
  column_1_style->setXSizePolicy (SIZE_POLICY_FIXED);
  column_1_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  column_1_style->setMarginLeft (72);
  column_1_style->setMarginRight (72);
  /*column_1_style->setYSizePolicy (SIZE_POLICY_FIXED);*/
  column_1_style->setFixedRight (469);
  /*column_1_style->setFixedDown (300);*/
  column_1_style->setBackground (bg);
  column_1_style->setForeground (fg);

  WPStyle *column_2_style = sm->newStyle ("column", "debug");
  column_2_style->setAlignment (ALIGN_LOW);
  column_2_style->setDirection (DIR_DOWN);
  column_2_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  column_2_style->setYSizePolicy (SIZE_POLICY_FIXED);
  column_2_style->setMarginLeft (10);
  column_2_style->setMarginRight (10);
  column_2_style->setFixedDown (300);
  column_2_style->setBackground (bg);
  column_2_style->setForeground (fg);


  sm->newStyleGroup ("line");
  WPStyle *line_ltr_low_style = sm->newStyle ("line", "ltr, low justified");
  WPStyle *line_2_col_style = sm->newStyle ("line", "ltr, 2 col");
  WPStyle *line_3_col_style = sm->newStyle ("line", "ltr, 3 col");
  WPStyle *line_asian_style = sm->newStyle ("line", "asian");

  line_ltr_low_style->setAlignment (ALIGN_LOW);
  line_ltr_low_style->setDirection (DIR_RIGHT);
  line_ltr_low_style->setXSizePolicy (SIZE_POLICY_FIXED);
  line_ltr_low_style->setFixedRight (469);
  line_ltr_low_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  line_ltr_low_style->setBackground (bg);
  line_ltr_low_style->setForeground (fg);
  line_ltr_low_style->setMarginLeft (72);
  line_ltr_low_style->setMarginRight (72);
  line_2_col_style->setAlignment (ALIGN_LOW);
  line_2_col_style->setDirection (DIR_RIGHT);
  line_2_col_style->setXSizePolicy (SIZE_POLICY_FIXED);
  line_2_col_style->setFixedRight (226);
  line_2_col_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  line_2_col_style->setBackground (bg);
  line_2_col_style->setForeground (fg);
  line_2_col_style->setMarginLeft (9);
  line_2_col_style->setMarginRight (9);
  line_3_col_style->setAlignment (ALIGN_LOW);
  line_3_col_style->setDirection (DIR_RIGHT);
  line_3_col_style->setXSizePolicy (SIZE_POLICY_FIXED);
  line_3_col_style->setFixedRight (150);
  line_3_col_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  line_3_col_style->setBackground (bg);
  line_3_col_style->setForeground (fg);
  line_3_col_style->setMarginLeft (9);
  line_3_col_style->setMarginRight (9);
  line_asian_style->setAlignment (ALIGN_LOW);
  line_asian_style->setDirection (DIR_DOWN);
  line_asian_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  line_asian_style->setYSizePolicy (SIZE_POLICY_FIXED);
  line_asian_style->setFixedDown (649);
  line_asian_style->setBackground (bg);
  line_asian_style->setForeground (fg);
  line_asian_style->setMarginUp (72);
  line_asian_style->setMarginDown (72);

  sm->newStyleGroup ("word");
  WPStyle *word_0_style = sm->newStyle ("word", "Default");
#if 0
  WPStyle *word_1_style = sm->newStyle ("word", "Title");
  WPStyle *word_2_style = sm->newStyle ("word", "Header");
  WPStyle *word_3_style = sm->newStyle ("word", "right to left");
  WPStyle *word_4_style = sm->newStyle ("word", "Asian");
#endif /* 0 */

  /* word */

  word_0_style->setFontName ("-*-helvetica-*-r-normal-*-*-100-*-*-*-*-*-*");
  word_0_style->setBackground (bg);
  word_0_style->setForeground (fg);
  word_0_style->setBackground (bg);
  word_0_style->setDirection (DIR_RIGHT);
  word_0_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_0_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);

#if 0
  /*
  word_1_style->setFontName ("-*-helvetica-*-r-normal-*-*-160-*-*-*-*-*-*");
  */
  word_1_style->setFontFamily ("Helvetica");
  word_1_style->setFontSlant ("r");
  word_1_style->setFontSWidth ("normal");
  word_1_style->setFontPointSize (160);
  word_1_style->setBackground (bg);
  word_1_style->setForeground (color1);
  word_1_style->setBackground (bg);
  word_1_style->setDirection (DIR_RIGHT);
  word_1_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_1_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);

  /*
  word_2_style->setFontName ("-*-helvetica-*-r-normal-*-*-240-*-*-*-*-*-*");
  */
  word_2_style->setFontFamily ("Helvetica");
  word_2_style->setFontSlant ("r");
  word_2_style->setFontSWidth ("normal");
  word_2_style->setFontPointSize (240);
  word_2_style->setBackground (bg);
  word_2_style->setForeground (color2);
  word_2_style->setBackground (bg);
  word_2_style->setDirection (DIR_RIGHT);
  word_2_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_2_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);

  /*
  word_3_style->setFontName ("-*-helvetica-bold-r-normal-*-*-120-*-*-*-*-*-*");
  */
  word_3_style->setFontFamily ("Helvetica");
  word_3_style->setFontWeight ("bold");
  word_3_style->setFontSlant ("r");
  word_3_style->setFontSWidth ("normal");
  word_3_style->setFontPointSize (120);
  word_3_style->setBackground (bg);
  word_3_style->setForeground (fg);
  word_3_style->setBackground (bg);
  word_3_style->setDirection (DIR_LEFT);
  word_3_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_3_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);

  word_4_style->setFontFamily ("Mincho");
  /*
  word_4_style->setFontWeight ("*");
  word_4_style->setFontSlant ("*");
  word_4_style->setFontSWidth ("*");
  */
  word_4_style->setFontPointSize (240);
  word_4_style->setBackground (bg);
  word_4_style->setForeground (fg);
  word_4_style->setBackground (bg);
  word_4_style->setDirection (DIR_RIGHT);
  word_4_style->setXSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_4_style->setYSizePolicy (SIZE_POLICY_BY_CHILDREN);
  word_4_style->setMaxChildren (2);
#endif /* 0 */


  WPStyle *document_style = sm->getDefaultStyleOfGroup ("document");
  WPStyle *page_style = sm->getDefaultStyleOfGroup ("page");
  WPStyle *column_style = sm->getDefaultStyleOfGroup ("column");
  WPStyle *line_style = sm->getDefaultStyleOfGroup ("line");
  WPStyle *word_style = sm->getDefaultStyleOfGroup ("word");
  /*WPStyle *word_style = style0;*/


  ContSlab dslab = new _ContSlab (document_style, NULL, NULL, NULL);
  ContSlab pslab = new _ContSlab (page_style, NULL, NULL, NULL);
  ContSlab cslab = new _ContSlab (column_style, NULL, NULL, NULL);
  ContSlab lslab = new _ContSlab (line_style, NULL, NULL, NULL);
  WordSlab wslab = new _WordSlab (word_style, "", 0);


  /* spots and views are NULL at this point (NULL = empty glist). */
  if (dslab->insertChild (0, pslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial page failed.\n");
  if (pslab->insertChild (0, cslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial col failed.\n");
  if (cslab->insertChild (0, lslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial line failed.\n");
  if (lslab->insertChild (0, wslab, NULL, NULL, debug) == false)
    fprintf (stderr, "insert of initial word failed.\n");


  return dslab;
#endif /* 0 */
