
#include <gdk_imlib.h>
#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"

#include "wp_plugin.h"
#include "wp_word_slab.h"
#include "wp_plugin.h"
#include "wp_font.h"

#include "plugin_table_helper.h"

/* XXX this is flipped the wrong way.  the table holds columns

    .-----------table-------------------------------------------.
    |  <CR0>                                                    |
    | .----trow-----------------------------------------------. |
    | | <CR0> .------tcell----------. .------tcell----------. | |
    | |       |   <CR0>             | |   <CR0>             | | |
    | |       | .-----lslab-------. | | .-----lslab-------. | | |
    | |       | | wslab:XXX       | | | | wslab:XXX       | | | |
    | |       | `-----------------' | | `-----------------' | | |
    | |       | .----lslab--------. | | .----lslab--------. | | |
    | |       | | XXX             | | | | XXX             | | | |
    | |       | `-----------------' | | `-----------------' | | |
    | |       `---------------------' `---------------------' | |
    | `-------------------------------------------------------' |
    |                                                           |
    | .----trow-----------------------------------------------. |
    | | <CR0> .------tcell----------. .------tcell----------. | |
    | |       |   <CR0>             | |   <CR0>             | | |
    | |       | .-----lslab-------. | | .-----lslab-------. | | |
    | |       | | wslab:XXX       | | | | wslab:XXX       | | | |
    | |       | `-----------------' | | `-----------------' | | |
    | |       | .----lslab--------. | | .----lslab--------. | | |
    | |       | | XXX             | | | | XXX             | | | |
    | |       | `-----------------' | | `-----------------' | | |
    | |       `---------------------' `---------------------' | |
    | `-------------------------------------------------------' |
    |                                                           |
    `-----------table-------------------------------------------'

  */        



char *table_logical_type_to_string (table_logical_type a)
{
  switch (a)
    {
    case lt_table: return "Table";
    case lt_table_row: return "TableRow";
    case lt_table_cell: return "TableCell";
    default: return NULL;
    }
}
table_logical_type string_to_table_logical_type (unsigned const char *s)
{
  if (!s) return (table_logical_type) 0;
  if (strcasecmp ((char *) s, "Table") == 0)
    return lt_table;
  if (strcasecmp ((char *) s, "TableRow") == 0)
    return lt_table_row;
  if (strcasecmp ((char *) s, "TableCell") == 0)
    return lt_table_cell;
  return (table_logical_type) 0; /* none */
}


/* containing, but not the same as... */
Slab find_table_containing_slab (Slab slab, int *pos_ret)
{
  int pos;

  while (slab && slab->getParent ())
    {
      Slab pslab = slab->getParent ();
      pos = pslab->getIndexOfChild (slab);
      if (pos < 0)
	return NULL;
      if (pos_ret) (*pos_ret) = pos;

      slab = pslab;

      if (slab->getPluginData () == ((void *) lt_table))
	return slab;
    }

  return NULL;
}


/* return 1 if slab is editable content of a table item */
int slab_is_table_content (Slab slab)
{
  if (! slab) return 0;

  int col_pos;
  Slab tableslab = find_table_containing_slab (slab, &col_pos);
  if (! tableslab) return 0;

  Slab colslab = (Slab) tableslab->getChildAt (col_pos);
  if (! colslab) return 0;

  int num_children = colslab->getNumberOfChildren ();

  for (int i=0; i<num_children; i++)
    {
      Slab cell = (Slab) colslab->getChildAt (i);
      if (cell == slab) continue; /* don't edit a cell */
      if (cell->slabIsInside (slab))
	/* anything inside of a cell is editable */
	return 1;
    }

  return 0;
}


Slab new_table_cell (WPStyleManager *sm, WPStyle *style)
{
  style_group *line_sg = sm->getStyleGroupByName ("line");

  int debug = -1;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();
  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5; bg->green = 0xA5A5; bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000; fg->green = 0x0000; fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  if (!style) return NULL;
  /*
  GdkFont *font = style->getFont (1);
  if (!font) return NULL;
  int ascent = font->ascent;
  int descent = font->descent;
  */

  WPStyle *tcell_style;
  WPStyle *lslab_style;

  tcell_style =
    sm->findOrCreateStyle (line_sg, "",
			   fg, bg,
			   ALIGN_LOW, DIR_DOWN,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0,
			   4, 4, 4, 4,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_OUTLINE,
			   WHEN_EMPTY_KEEP, 0);

  lslab_style =
    sm->findOrCreateStyle (line_sg, "",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN,


			   SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0,

			   /*
			   SIZE_POLICY_FIXED,
			   0, ascent + 1, 0, descent + 1,
			   */

			   0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

# if 0
  wslab_style =
    sm->findOrCreateStyle (line_sg,
			   "-*-helvetica-medium-r-normal-*-*-120-*-*-*-*-*-*",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);
# endif /* 0 */

  ContSlab tcell = new _ContSlab (tcell_style, NULL, NULL, NULL);
  LBTSlab cr0slab = new _LBTSlab (style, 2, 0);
  ContSlab lslab = new _ContSlab (lslab_style, NULL, NULL, NULL);
  WordSlab wslab = new _WordSlab (style, "ok", 2);


  if (lslab->insertChild (0, wslab, NULL, NULL, debug) == false)
    fprintf (stderr, "table -- insert of initial word failed.\n");
  if (tcell->insertChild (0, cr0slab, NULL, NULL, debug) == false)
    fprintf (stderr, "table -- insert of initial cr failed.\n");
  if (tcell->insertChild (1, lslab, NULL, NULL, debug) == false)
    fprintf (stderr, "table -- insert of initial line failed.\n");

  tcell->setPluginData ((void *) lt_table_cell);

  return tcell;
}


Slab new_table_row (WPStyleManager *sm,
		    WPStyle *style,
		    int num_cells)
{
  style_group *line_sg = sm->getStyleGroupByName ("line");

  int debug = -1;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();
  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5; bg->green = 0xA5A5; bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000; fg->green = 0x0000; fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  WPStyle *trow_style;

  trow_style =
    sm->findOrCreateStyle (line_sg, "",
			   fg, bg,
			   ALIGN_LOW, DIR_DOWN,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  WPStyle *wslab_style =
    sm->findOrCreateStyle (line_sg,
			   "-*-helvetica-medium-r-normal-*-*-120-*-*-*-*-*-*",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  ContSlab trow = new _ContSlab (trow_style, NULL, NULL, NULL);
  LBTSlab cr0slab = new _LBTSlab (wslab_style, 2, 0);

  if (trow->insertChild (0, cr0slab, NULL, NULL, debug) == false)
    fprintf (stderr, "table -- insert of CRfailed.\n");

  int i;

  for (i=0; i<num_cells; i++)
    {
      Slab tcell = new_table_cell (sm, style);
      if (trow->insertChild (i+1, tcell, NULL, NULL, debug) == false)
	fprintf (stderr, "table -- insert of cell %d failed.\n", i);
    }

  trow->setPluginData ((void *) lt_table_row);

  return trow;
}


Slab new_table (WPStyleManager *sm,
		WPStyle *style,
		int num_cols, int num_rows)
{
  style_group *line_sg = sm->getStyleGroupByName ("line");

  int debug = -1;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();
  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5; bg->green = 0xA5A5; bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000; fg->green = 0x0000; fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  WPStyle *table_style;

  table_style =
    sm->findOrCreateStyle (line_sg, "",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  WPStyle *wslab_style =
    sm->findOrCreateStyle (line_sg,
			   "-*-helvetica-medium-r-normal-*-*-120-*-*-*-*-*-*",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  ContSlab table = new _ContSlab (table_style, NULL, NULL, NULL);

  LBTSlab cr0slab = new _LBTSlab (wslab_style, 2, 0);

  if (table->insertChild (0, cr0slab, NULL, NULL, debug) == false)
    fprintf (stderr, "table -- insert of initial word failed.\n");

  int i;

  for (i=0; i<num_rows; i++)
    {
      Slab trow = new_table_row (sm, style, num_cols);
      if (table->insertChild (i+1, trow, NULL, NULL, debug) == false)
	fprintf (stderr, "table -- insert of row %d failed.\n", i);
    }

  table->setPluginData ((void *) lt_table);

  return table;
}


void insert_new_table (WPDocument *doc, WPStyleManager *sm,
		       WPStyle *style,
		       Spot spot, int num_cols, int num_rows)
{
  if (sm == NULL) return;
  if (spot == NULL) return;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();
  GdkColor *bg = (GdkColor *) malloc (sizeof (GdkColor));
  bg->red = 0xA5A5; bg->green = 0xA5A5; bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, bg);
  GdkColor *fg = (GdkColor *) malloc (sizeof (GdkColor));
  fg->red = 0x0000; fg->green = 0x0000; fg->blue = 0x0000;
  gdk_color_alloc (colormap, fg);

  style_group *word_sg = sm->getStyleGroupByName ("word");
  WPStyle *style0 =
    sm->findOrCreateStyle (word_sg,
			   "-*-helvetica-medium-r-normal-*-*-120-*-*-*-*-*-*",
			   fg, bg,
			   ALIGN_LOW, DIR_RIGHT,
			   SIZE_POLICY_BY_CHILDREN, SIZE_POLICY_BY_CHILDREN,
			   0, 0, 0, 0, 0, 0, 0, 0,
			   0, /* max children */
			   MARGIN_RENDERING_POLICY_BLANK, WHEN_EMPTY_KEEP, 0);

  Slab table = new_table (sm, style, num_cols, num_rows);

  doc->insertSlab (spot, table);

  /* make sure we don't trap the cursor in the table */
  if (table->getPrevLeaf () == NULL)
    {
      spot->setPosition (0);
      printf ("inserting empty word between table and start\n");

      Slab ws = new _WordSlab (style0, "", 0);
      doc->insertSlab (spot, ws);
    }
  if (table->getNextLeaf () == NULL)
    {
      spot->setSlab (table);
      spot->setPosition (num_rows+1);
      printf ("inserting empty word between table and end\n");
      doc->createFollowingWord (spot, style0);
    }

  spot->setPosition (0);
  spot->setSlab (table->lastLowestContainer ());
}


static void table_insert_row_callback (GtkWidget *, gpointer /*client_data*/)
{
  printf ("table insert row\n");
}
static void table_insert_column_callback (GtkWidget *, gpointer /*client_d*/)
{
  printf ("table insert column\n");
}
static void table_delete_row_callback (GtkWidget *, gpointer /*client_data*/)
{
  printf ("table delete row\n");
}
static void table_delete_column_callback (GtkWidget *, gpointer /*client_d*/)
{
  printf ("table delete column\n");
}

static GnomeUIInfo table_popup_info[] =
{
  { GNOME_APP_UI_ITEM, "Insert Table Row", NULL,
    table_insert_row_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Insert Table Column", NULL,
    table_insert_column_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Delete Table Row", NULL,
    table_delete_row_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Delete Table Column", NULL,
    table_delete_column_callback, NULL, NULL },

  { GNOME_APP_UI_ENDOFINFO }
};


GtkWidget *newTableUI (WPView *view)
{
  GtkWidget *popup = gnome_popup_menu_new (table_popup_info);
  GtkWidget *da = wp_view_get_drawing_area (view);
  gnome_popup_menu_attach (popup, da, NULL);
  return popup;
}


void deleteTableUI (WPView */*view*/, GtkWidget *popup)
{
  gtk_widget_destroy (popup);
}


void showTableUI (WPView */*view*/, GtkWidget *popup, GdkEventButton *geb)
{
  gnome_popup_menu_do_popup (popup, NULL, NULL, geb, NULL);
}
