/*
 * Copyright  1997, 1998 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "config.h"
#include <gtk/gtk.h>
#include <glib.h>
#include <stdlib.h>
#include <signal.h>

#include "wp_document.h"
#include "wp_plugin_manager.h"
/*#include "wp_frame.h"*/
#include "wp_view.h"
#include "wp_clipboard.h"
#include "wp_prefs.h"
#include "wp_utils.h"

#include "wp_ui.h"

/* go/gedit plugin stuff */
#include "plugin.h"
#include "gwp_plugin_api.h"

#include "wp_mdi_document.h"

GnomeMDI *mdi;
/*GnomeMDIMode mdi_mode = GNOME_MDI_DEFAULT_MODE;*/
GnomeMDIMode mdi_mode = GNOME_MDI_NOTEBOOK;

static void quit_app_cb (GtkWidget */*widget*/, gpointer /*data*/)
{
  printf ("gwp: quit_app_cb called\n");
}


static int save_state (GnomeClient *client,
		       gint /*phase*/, GnomeRestartStyle /*save_style*/,
		       gint /*shutdown*/,
		       GnomeInteractStyle /*interact_style*/,
		       gint /*fast*/, gpointer client_data)
{
  char *argv[20];
  int i;
  gint xpos, ypos;

  /*gdk_window_get_origin (window->window, &xpos, &ypos);*/
  xpos = 100;
  ypos = 200;

  i = 0;
  argv[i++] = (char *)client_data;
  argv[i++] = "-x";
  argv[i++] = strdup (itoa (xpos));
  argv[i++] = "-y";
  argv[i++] = strdup (itoa (ypos));

  gnome_client_set_restart_command (client, i, argv);
  gnome_client_set_clone_command (client, 0, NULL);

  free(argv[2]);
  free(argv[4]);

  return TRUE;
}


int main (int argc, char *argv[])
{
  GnomeClient *client;

  int session_xpos = -1;
  int session_ypos = -1;
  const struct poptOption options[] =
  {
    {NULL, 'x', POPT_ARG_INT, &session_xpos, 0, NULL, NULL},
    {NULL, 'y', POPT_ARG_INT, &session_ypos, 0, NULL, NULL},
    {NULL, '\0', 0, NULL, 0}
  };


# ifdef BROKEN
  printf ("gwp built with --enable-broken\n");
# endif /* BROKEN */

  /*
  bindtextdomain (PACKAGE, GNOMELOCALEDIR);
  textdomain (PACKAGE);
  */

  /* initialize gnome libraries */
  /*gnome_init ("gwp", VERSION, argc, argv);*/
  gnome_init_with_popt_table ("gwp", VERSION, argc, argv, options, 0, NULL);

  /*client = gnome_client_new_default ();*/
  client = gnome_master_client ();

  gtk_object_ref (GTK_OBJECT (client));
  gtk_object_sink (GTK_OBJECT (client));
  gtk_signal_connect (GTK_OBJECT (client), "save_yourself",
		      GTK_SIGNAL_FUNC (save_state), argv[0]);
  gtk_signal_connect (GTK_OBJECT (client), "die",
		      GTK_SIGNAL_FUNC (quit_app_cb), argv[0]);


  /* stop so i can find problems */
  g_log_set_always_fatal ((GLogLevelFlags) 0xFFFF);

  /* persistent preferences */
  preferences = new WPPreferences ();


  /* load gwp .so style plugins */
  WPPluginManager::thePluginManager ()->scan_for_plugins ();

  /* load Go plugins */
  plugin_callback_struct callbacks;
  gwp_setup_plugin_callbacks (&callbacks);
  plugin_query_all (&callbacks);
  /* plugins can cause sigpipe */
  signal (SIGPIPE, SIG_IGN);


  mdi = GNOME_MDI (gnome_mdi_new ("gwp", "Gwp"));
  gnome_mdi_set_menubar_template (mdi, menubar_info);
  gnome_mdi_set_toolbar_template (mdi, toolbar_info);

  /* and document menu and document list paths */
  gnome_mdi_set_child_menu_path (mdi, _("File"));
  gnome_mdi_set_child_list_path (mdi, _("Children/"));

  /* connect signals */
  /*
  gtk_signal_connect(GTK_OBJECT(mdi),
		     "remove_child", GTK_SIGNAL_FUNC(remove_doc_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(mdi),
		     "destroy", GTK_SIGNAL_FUNC(cleanup_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(mdi),
		     "view_changed", GTK_SIGNAL_FUNC(view_changed_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(mdi),
		     "child_changed", GTK_SIGNAL_FUNC(child_changed_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(mdi),
		     "app_created", GTK_SIGNAL_FUNC(customize_app_cb), NULL);
  */


  /* set MDI mode */
  gnome_mdi_set_mode (mdi, mdi_mode);

  /* XXX restore state */
  /*    if (GNOME_CLIENT_CONNECTED (client)) { ... } */

  WPPlugin *plugin =
    WPPluginManager::thePluginManager ()->get_plugin_by_name ("gwp");
  if (plugin)
    {
      WPMDIDocument *child = wp_mdi_document_new (plugin,
						  wp_unique_document_name ());
      gnome_mdi_add_child (mdi, GNOME_MDI_CHILD (child));
      gnome_mdi_add_view (mdi, GNOME_MDI_CHILD (child));
    }
  else
    {
      fprintf (stderr, "can't find gwp plugin\n");
    }

  /*gnome_mdi_open_toplevel (mdi);*/

  /* pass control to gtk */
  gtk_main ();

  gtk_object_unref(GTK_OBJECT(client));
  return 0;
}


void hi (void)
{
  /* ugh.  fool linker */
  gtk_ctree_new (0,0);
  gtk_tree_set_view_lines (NULL, 0);
}
