/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <wp_associative.h>

//
// node functions
//

node::node(const char *name,const char *value)
 {
  parent=NULL;
  next=NULL;
  contents=NULL;
  name_value=NULL;
  if (name||value||*name||*value)
   set_nv(name,value);
 }

node::~node()
 {
  node *next,*tmp;;
  leave_parent();
  if (name_value)
   free(name_value);
  if (contents)
   for(tmp=contents;tmp;tmp=next)
    {
     next=tmp->next;
     delete tmp;
    }
 }

// name of this class
const char *node::class_name(void)
 {
  return "node";
 }
   
// get the name contained in the name/value pair
const char *node::name_of(void)
 {
  return name_value ? name_value : "";
 }

// get the value contained in the name/value pair
const char *node::value_of(void)
 {
  return name_value ? name_value+strlen(name_value)+1 : "";
 }

// set name+value
int node::set_nv(const char *name,const char *value)
 {
  char *old=name_value;
  int len=strlen(name);
  if (*name||*value)
   {
    name_value=(char *)malloc(len+strlen(value)+2);
    strcpy(name_value,name);
    strcpy(name_value+len+1,value);
   }
  else
   name_value=NULL;
  // delay the free since we may want to call this with one of
  // the old versions
  if (old)
   free(old);
  return 0;
 }
   
// copy this node and its contents recursively.
node *node::copy(void)
 {
  node *tmp;
  node *cloned=new node(name_of(),value_of());
  for(tmp=contents;tmp;tmp=tmp->next)
   // add a copy of the element to the end of the list
   cloned->add(tmp->copy(),0);
  return cloned;
 }
   
// tell our parent node to remove us.
int node::leave_parent(void)
 {
  if (parent)
   parent->remove_node(this);
  parent=NULL;
  return 0;
 }

// list functions
node *node::find_node(const char *name)
 {
  node *tmp;
  for(tmp=contents;tmp;tmp=tmp->next)
   if (!strcmp(name,tmp->name_of()))
    return tmp;
  return NULL;
 }
 
const char *node::find_value(const char *name,const char *def)
 {
  node *tmp;
  const char *v;
  for(tmp=contents;tmp;tmp=tmp->next)
   if (!strcmp(name,tmp->name_of()))
    return (v=tmp->value_of()) ? v : def;
  return def;
 }

int node::add(node *node,int flags)
 {
  if (!(flags & allow_duplicate_names))
   destroy(node->name_of());
  // extract from old home
  node->leave_parent();
  node->next=contents;
  contents=node;
  return 0;
 }

int node::remove_node(node *remove)
 {
  node *tmp,*trail=NULL;
  for(tmp=contents;tmp && (tmp!=remove);tmp=(trail=tmp)->next);
  if (!tmp)
   // no node removed
   return -1;
  if (trail)
   trail->next=tmp->next;
  else
   contents=tmp->next;
  // disconnect node
  tmp->next=NULL;
  tmp->parent=NULL;
  return 0;
 }

node *node::add(const char *name,const char *value,int flags)
 {
  node *tmp;
//  cerr<<"node::add name="<<name<<" value="<<value<<"\n";
  add(tmp=new node(name,value),flags);
  return tmp;
 }

int node::destroy(const char *name)
 {
  int count=0;
  node *tmp,*next;
  // seek and destroy node(s) with name 'name'
  for(tmp=contents;tmp;tmp=next)
   {
    next=tmp->next;
    if (tmp->name_of() && !strcmp(name,tmp->name_of()))
     {
      // it will remove itself upon deletion
      delete tmp;
      count++;
     }
   }
  return count;
 }

// child list traversal.
node *node::get_contents(void)
 {
  return contents;
 }

node *node::get_next(void)
 {
  return next;
 }

// print our tree
void node::dump(ostream &out,int depth)
 {
  (void)depth;
  node *tmp;
  out<<name_of()<<"="<<value_of()<<"\n";
  if (contents)
   {
    out<<"{\n";
    for(tmp=contents;tmp;tmp=tmp->next)
     tmp->dump(out,depth+1);
    cout<<"}\n";
   }
 }

