/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include <glib.h>
#include "wp_clipboard.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include <strings.h>
#include "wp_plugin.h"


static guchar *clipboard_buffer;


/* free what this returns. */
char *wp_convert_slab_to_clipboard (WPDocument *doc, Slab s)
{
  char *ret;

  switch (s->classID ())
    {
    case WordSlabID:
      {
	WordSlab ws = (WordSlab) s;
	char *text = ws->getWord ();
	int length = ws->getWordLength ();
	ret = (char *) malloc (length + 1);
	strncpy (ret, text, length);
	ret[length] = '\0';
	break;
      }
    case STSlabID:
      {
	ret = strdup (" ");
	break;
      }
    case LBTSlabID:
      {
	ret = strdup ("\n");
	break;
      }
    case ImageSlabID:
      {
	ret = strdup ("<Image>");
	break;
      }
    case ContSlabID:
      {
	ContSlab ds = (ContSlab) s;
	char *text = ds->drawASCII (doc, 0, 0);
	if (text)
	  ret = text;
	else
	  ret = strdup ("");
	break;
      }
    default:
      {
	char *text = "<unknown>";
	ret = (char *) malloc (strlen (text) + 1);
	strcpy (ret, text);
	break;
      }
    }

  return ret;
}


guchar *wp_clipboard_selection_copy(WPView *view, Spot mark, Spot point)
{
  if (!view) return (guchar *) strdup ("");
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return (guchar *) strdup ("");

  guchar *string;
  
  Slab ms = mark->getSlab ();
  int mp = mark->getPosition ();
  Slab ps = point->getSlab ();
  int pp = point->getPosition ();

  /* if they are out of order, swap them */
  if (! wp_slab_order (ms, ps))
    {
      Slab ts = ps; ps = ms; ms = ts;
      int tp = pp; pp = mp; mp = tp;
    }
  if (ms == ps && mp > pp)
    {
      Slab ts = ps; ps = ms; ms = ts;
      int tp = pp; pp = mp; mp = tp;
    }

  if (ms == ps && mp == pp)
    return (unsigned char *) strdup (""); /* no region selected */

  /* convert the slabs into text */
  if (ms == ps)
    {
      char *str = wp_convert_slab_to_clipboard (doc, ms);
      char *sub = str + mp;
      str[ pp ] = '\0';

      string = (guchar *) strdup(sub);
      free (str);
    }
  else
    {
      char *str;
      char *sub;
      char *tmp1, *tmp2;
      Slab s; /* pointer to current slab between first and last */

      /* get the first slab */
      str = wp_convert_slab_to_clipboard (doc, ms);
      sub = str + mp;
      tmp1 = strdup(sub);
      /* tmp1 = (char *) malloc(sizeof(char) * (strlen(sub) + 1));
      strcpy(tmp1, sub); */
      free (str);

      /* get the middle slabs */
      for (s=ms->getNextLeaf ();
	   (s != NULL) && (s != ps);
	   s = s->getNextLeaf ())
	{
	  str = wp_convert_slab_to_clipboard (doc, s);
	  tmp2 = (char *) malloc (sizeof(char) *
				  (strlen(str) + strlen(tmp1) + 1));
          strcpy(tmp2, tmp1);
          free(tmp1);
          strcat(tmp2, str);
          tmp1 = tmp2;
	  free (str);
	}

      /* get the last slab */
      str = wp_convert_slab_to_clipboard (doc, ps);
      str[ pp ] = '\0';
      tmp2 = (char *) malloc(sizeof(char) * (strlen(str) + strlen(tmp1) + 1));
      strcpy(tmp2, tmp1);
      free(tmp1);
      strcat(tmp2, str);
      string = (guchar *) tmp2;
      free (str);
    }

  return string;
}


void wp_clipboard_copy (WPView *view, Spot mark, Spot point)
{
  if (!view) return;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;

  if (wp_view_hl_is_on (view))
    clipboard_buffer = wp_clipboard_selection_copy(view, mark, point);
}


/* inserts at the views point */
void wp_clipboard_paste (WPView *view)
{
  /* XXX combine some of this with wp_view.C:view_selection_received */

  if (!view) return;
  WPDocument *doc = wp_view_get_document(view);
  if (!doc) return;

  unsigned int i;

  if (clipboard_buffer == NULL)
    {
      gtk_selection_convert (wp_view_get_drawing_area (view),
			     GDK_SELECTION_PRIMARY,
			     GDK_SELECTION_TYPE_STRING,
			     GDK_CURRENT_TIME);
      return;
    }

  doc->displayRefreshOff ();

  for (i=0; i<strlen((char *) clipboard_buffer); i++)
    wp_view_insert_char (view, clipboard_buffer[ i ]);

  doc->displayRefreshOn ();
  doc->refresh ();
}
