/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include "wp_slab.h"
#include "wp_document.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_lbt_slab.h"
#include "wp_st_slab.h"
#include "wp_image_slab.h"
#include "wp_character.h"
#include "wp_plugin.h"
#include "wp_font.h"
#include "wp_prefs.h"

GHashTable *WPDocument::documents = NULL;
guint32 WPDocument::untitled_document_count = 1; /* we start at 1. */


WPDocument::WPDocument (char *doc_name)
{
  read_only = 0;
  display_refresh = 1;

  debug = -1; /* set this to 0 for lots of debugging output,
		 or -1 for quiet */


  spots = NULL; /* an empty g_list */
  /*views = NULL;*/ /* an empty g_list */

  plugin = NULL;
  mdi_doc = NULL;

  if (!WPDocument::documents)
    WPDocument::documents = g_hash_table_new(g_str_hash,
					     g_str_equal);

  if (!doc_name)
    {
      char buf[100];
      sprintf (buf, "Untitled-%d", WPDocument::untitled_document_count++);
      g_hash_table_insert(WPDocument::documents, strdup(buf), this);
      document_name = strdup(buf);
    }
  else
    {
      g_hash_table_insert(WPDocument::documents, strdup(doc_name), this);
      document_name = strdup(doc_name);
    }

  root = NULL;

  plugin_scratchspace = g_hash_table_new (g_str_hash, g_str_equal);
}


WPDocument::~WPDocument ()
{
  /*plugin->endDocument (this);*/
}


void WPDocument::set_mdi_document (WPMDIDocument *set_mdi_doc)
{
  mdi_doc = set_mdi_doc;
}


WPMDIDocument *WPDocument::get_mdi_document ()
{
  return mdi_doc;
}


void WPDocument::setPlugin (WPPlugin *set_plugin)
{
  plugin = set_plugin;
}


void *WPDocument::get_plugin_scratch_space (char *plugin_name)
{
  void *space = g_hash_table_lookup (plugin_scratchspace, plugin_name);
  return space;
}


void WPDocument::set_plugin_scratch_space (char *plugin_name, void *space)
{
  g_hash_table_insert (plugin_scratchspace, plugin_name, space);
}


void WPDocument::dumpToScreen ()
{
  root->dumpToScreen (0);
}



void WPDocument::paint_slab (WPView *g, Slab top, GdkRectangle */*a*/)
{
  GList *temp_list = g_list_append (NULL, g);
  WpPoint p;
  if (top == NULL) top = root;

  /* XXX -- need a way to only paint the
     exposed area, instead of the whole thing. */

  p.x = root->getMLeft ();
  p.y = root->getMUp ();
  if (top->getParent () != NULL)
    p = top->getParent ()->getPointOfChild (top);

  top->draw (p.x, p.y, temp_list, debug);
  g_list_free (temp_list); /* XXX -- is this correct? */
}


void WPDocument::erase_slab (WPView *g, Slab top, GdkRectangle */*a*/)
{
  GList *temp_list = g_list_append (NULL, g);
  WpPoint p;
  if (top == NULL) top = root;

  /* XXX -- need a way to only paint the
     exposed area, instead of the whole thing. */

  p.x = 0;
  p.y = 0;
  if (top->getParent () != NULL)
    p = top->getParent ()->getPointOfChild (top);

  top->erase (p.x, p.y, temp_list);
  g_list_free (temp_list); /* XXX -- is this correct? */
}


GList *WPDocument::get_document_views ()
{
  if (mdi_doc)
    return GNOME_MDI_CHILD (mdi_doc)->views;
  return NULL;
}


void WPDocument::refresh ()
{
  if (preferences->draw_fast)
    for (GList *v = get_document_views (); v != NULL; v = v->next)
      wp_view_refresh ((WPView *) v->data);
  else
    {
      if (display_refresh)
	for (GList *v = get_document_views (); v != NULL; v = v->next)
	  {
	    WPView *view = (WPView *) v->data;
	    wp_view_clear_back_pixmap (view);
	    paint_slab (view, root, NULL);
	    wp_view_copy_back_pixmap (view);
	  }
    }
}



/*
 * Add a view of this document
 */

#if 0
void WPDocument::addView (WPView *g)
{
  /*printf ("doc adding view: %p\n", g);*/
  views = g_list_append (views, g);
}
#endif /* 0 */


/*
 * Remove a view of this document
 */

#if 0
void WPDocument::deleteView (WPView *g)
{
  /*printf ("doc deleting view: %p\n", g);*/
  views = g_list_remove (views, g);
}
#endif /* 0 */



/*
  Allows a view to mark a point in the document.  The document must keep
  track of *all* spots in the document so it can keep them from becoming
  invalid during editing.
*/

Spot WPDocument::newDocumentSpot ()
{
  Slab firstLowest;

  if (!root) return NULL;

  Spot new_spot = new _Spot ();
  spots = g_list_append (spots, new_spot);

  firstLowest = root->firstLowestContainer ();
  new_spot->setSlab (firstLowest);
  new_spot->setPosition (0);
  return new_spot;
}


void WPDocument::deleteDocumentSpot (Spot dead_spot)
{
  spots = g_list_remove (spots, dead_spot);
}



/**********************************************************************/
/*                                                                    */
/*                                                                    */
/**********************************************************************/


/*
  insertSlab -- insert a slab into the document tree
  insertCharacter -- insert a character into the document tree
  deleteChild -- delete a slab or character from the document tree
                 (deletes the child following the spot)

  These each take an argument "spot" which should have been
  created by calling newDocumentSpot.  spot determines where
  int the tree the insert or delete takes place.
  */


int WPDocument::insertSlab (Spot spot, Slab ns)
{
  Spot upLevel;
  unsigned long old_width = getDocumentWidth();
  unsigned long old_height = getDocumentHeight();

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  if (spot->getSlab()->ensureBorder (spot->getPosition (),
				     spots,
				     get_document_views (),
				     debug) == false)
    return 0;

  upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise();

  /* we know that 'spot' is next to a border, now figure out which
     side of the border it's on */

  int position = upLevel->getPosition ();
  if (spot->getPosition () > 0)
    position ++;

  Slab upLevelSlab = upLevel->getSlab ();

  if (upLevelSlab->insertChild (position, ns, spots,
				get_document_views (), debug) == false)
    {
      printf ("WPDocument::insertSlab upLevelSlab->insertChild failed\n");
      Slab next = spot->getSlab();
      if (next) next->trim_slab (spots, get_document_views (), debug);
      spot->getSlab ()->trim_slab (spots, get_document_views (), debug);
      return 0;
    }

  /* set 'spot' to be at the end of the new slab */

  spot->setSlab (ns);
  spot->setPosition (ns->getNumberOfChildren ());

  /* this crud should probably be somewhere in the slab code,
     but I can't get it to fit without causing problems. */

  Slab upLevelPrev = upLevelSlab->getPrevNode ();
  if (upLevelPrev != NULL &&
      upLevelPrev->classID () == ContSlabID &&
      upLevelPrev->lastLowestContainer ()->isLast () == 0 &&
      upLevelSlab->firstLowestContainer ()->isFirst () == 0)
    ((ContSlab) upLevelPrev)->attemptToStealChildren (spots,
						      get_document_views (),
						      debug);

  delete upLevel;

  if (! preferences->draw_fast)
    refresh ();

  /* if the document changed size thanks to this insertion,
     notify the views that their scrollbars need to update. */
  if (getDocumentWidth() != old_width || getDocumentHeight() != old_height)
    for (GList *v = get_document_views (); v != NULL; v = v->next)
      wp_view_sync_scrollbars((WPView*)v->data);

  return 1;
}


/* call this when something in a style has changed. */
/* returns 1 if anything changed. */
int WPDocument::styleChangeNotify (WPStyle *s)
{
  int ret = root->styleChangeNotify (s, spots, get_document_views (), debug);

  if (!preferences->draw_fast)
    refresh ();

  return ret;
}


/* XXX move this into plugs */
int WPDocument::setWordStyle (WPStyleManager *sm,
			      Spot spot, char *style_name)
{
  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  if (spot->getSlab()->classID() != WordSlabID &&
      spot->getSlab()->classID() != STSlabID &&
      spot->getSlab()->classID() != LBTSlabID)
    {
      fprintf (stderr,
	       "WPDocument::setStyle -- spot is not in a WordSlab\n");
      return 0;
    }

  WPStyle *style = sm->getStyleByName ("word", style_name);
  if (style == NULL) return 0;
  WordSlab ws = (WordSlab) spot->getSlab();
  ws->setStyle (style, spots, get_document_views (), debug);

  if (! preferences->draw_fast)
    refresh ();

  return 1;
}


/* XXX move this into plugs */
int WPDocument::setLineStyle (WPStyleManager *sm,
			      Spot spot, char *style_name)
{
  WpPoint p;

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  WPStyle *style = sm->getStyleByName ("line", style_name);
  if (style == NULL) return 0;
  WordSlab ws = (WordSlab) spot->getSlab();
  if (ws == NULL) return 0;
  ContSlab cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;

  if (preferences->draw_fast)
    {
      p.x = 0; p.y = 0;
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->erase (p.x, p.y, get_document_views ());
    }

  cs->setStyle (style, spots, get_document_views (), debug);

  if (preferences->draw_fast)
    {
      p.x = root->getMLeft ();
      p.y = root->getMUp ();
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->draw (p.x, p.y, get_document_views (), debug);
    }
  else
    refresh ();

  return 1;
}


/* XXX move this into plugs */
int WPDocument::setColumnStyle (WPStyleManager *sm,
				Spot spot, char *style_name)
{
  WpPoint p;

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  WPStyle *style = sm->getStyleByName ("column", style_name);
  if (style == NULL) return 0;
  WordSlab ws = (WordSlab) spot->getSlab();
  if (ws == NULL) return 0;
  ContSlab cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) cs->getParent ();
  if (cs == NULL) return 0;

  if (preferences->draw_fast)
    {
      p.x = 0; p.y = 0;
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->erase (p.x, p.y, get_document_views ());
    }

  cs->setStyle (style, spots, get_document_views (), debug);

  if (preferences->draw_fast)
    {
      p.x = root->getMLeft ();
      p.y = root->getMUp ();
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->draw (p.x, p.y, get_document_views (), debug);
    }
  else
    refresh ();

  return 1;
}


/* XXX move this into plugs */
int WPDocument::setPageStyle (WPStyleManager *sm,
			      Spot spot, char *style_name)
{
  WpPoint p;

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  WPStyle *style = sm->getStyleByName ("page", style_name);
  if (style == NULL) return 0;
  WordSlab ws = (WordSlab) spot->getSlab();
  if (ws == NULL) return 0;
  ContSlab cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) cs->getParent ();
  if (cs == NULL) return 0;

  if (preferences->draw_fast)
    {
      p.x = 0; p.y = 0;
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->erase (p.x, p.y, get_document_views ());
    }

  cs->setStyle (style, spots, get_document_views (), debug);

  if (preferences->draw_fast)
    {
      p.x = root->getMLeft ();
      p.y = root->getMUp ();
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->draw (p.x, p.y, get_document_views (), debug);
    }
  else
    refresh ();

  return 1;
}


/* XXX move this into plugs */
int WPDocument::setDocumentStyle (WPStyleManager *sm,
				  Spot spot, char *style_name)
{
  WpPoint p;

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  WPStyle *style = sm->getStyleByName ("document", style_name);
  if (style == NULL) return 0;
  WordSlab ws = (WordSlab) spot->getSlab();
  if (ws == NULL) return 0;
  ContSlab cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) ws->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) cs->getParent ();
  if (cs == NULL) return 0;
  cs = (ContSlab) cs->getParent ();
  if (cs == NULL) return 0;

  if (preferences->draw_fast)
    {
      p.x = 0; p.y = 0;
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->erase (p.x, p.y, get_document_views ());
    }

  cs->setStyle (style, spots, get_document_views (), debug);

  if (preferences->draw_fast)
    {
      p.x = root->getMLeft ();
      p.y = root->getMUp ();
      if (cs->getParent () != NULL)
	p = cs->getParent ()->getPointOfChild (cs);
      cs->draw (p.x, p.y, get_document_views (), debug);
    }
  else
    refresh ();

  return 1;
}


/* XXX move this into plugins */
int WPDocument::setStyle (WPStyleManager *sm,
			  Spot spot, char *group_name, char *style_name)
{
  if (strcmp (group_name, "word") == 0)
    return setWordStyle (sm, spot, style_name);
  else if (strcmp (group_name, "line") == 0)
    return setLineStyle (sm, spot, style_name);
  else if (strcmp (group_name, "column") == 0)
    return setColumnStyle (sm, spot, style_name);
  else if (strcmp (group_name, "page") == 0)
    return setPageStyle (sm, spot, style_name);
  else if (strcmp (group_name, "document") == 0)
    return setDocumentStyle (sm, spot, style_name);

  gdk_beep();

  return 0;
}


/* XXX */
int WPDocument::setStyles (WPStyleManager */*sm*/,
			   WPStyle *document_style,
			   WPStyle *page_style,
			   WPStyle *column_style,
			   WPStyle *line_style,
			   WPStyle *word_style)
{
  Slab cur;

  root->setStyle (document_style, spots, get_document_views (), -1);
  root->styleChangeNotify (root->getStyle (), spots,
			   get_document_views (), -1);

  cur = (ContSlab) root->getChildAt (0);
  for (;cur; cur = cur->getNextNode ())
    {
      cur->setStyle (page_style, spots, get_document_views (), -1);
      /*cur->styleChangeNotify (cur->getStyle (), spots, views, -1);*/
    }

  cur = (ContSlab) root->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  for (;cur; cur = cur->getNextNode ())
    {
      cur->setStyle (column_style, spots, get_document_views (), -1);
      /*cur->styleChangeNotify (cur->getStyle (), spots, views, -1);*/
    }

  cur = (ContSlab) root->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  for (;cur; cur = cur->getNextNode ())
    {
      cur->setStyle (line_style, spots, get_document_views (), -1);
      /*cur->styleChangeNotify (cur->getStyle (), spots, views, -1);*/
    }

  cur = (ContSlab) root->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  cur = (ContSlab) cur->getChildAt (0);
  for (;cur; cur = cur->getNextNode ())
    {
      cur->setStyle (word_style, spots, get_document_views (), -1);
      /*cur->styleChangeNotify (cur->getStyle (), spots, views, -1);*/
    }

  if (! preferences->draw_fast)
    refresh ();

  return 1;
}



/*** editing ***/



int WPDocument::replaceWord (Spot spot,
			     char *new_word,
			     int new_word_length)
{
  Slab slab = spot->getSlab ();
  if (!slab) return 0;
  if (slab->classID () != WordSlabID) return 0;
  WordSlab ws = (WordSlab) slab;
  int ret =  ws->replaceWord (new_word, new_word_length, spots,
			      get_document_views (), debug);

  if (! preferences->draw_fast)
    refresh ();

  return ret;
}


Slab WPDocument::createFollowingWord (Spot spot, WPStyle *style)
{
  /* cursor isn't currently in a wordslab -- make a
     new one and place spot at the beginning of it */
  Slab not_a_word = spot->getSlab ();
  not_a_word->ensureBorder (spot->getPosition(), spots,
			    get_document_views (), debug);
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();
  
  if (spot->getPosition () > 0)
    upLevel->setPosition (upLevel->getPosition () + 1);

  Slab nw = new _WordSlab (style);
  if (upLevel->getSlab()->insertChild (upLevel->getPosition(),
				       nw, spots,
				       get_document_views (), debug) == false)
    {
      printf ("WPDocument::createFollowingWord insertChild failed\n");
      delete (upLevel);
      delete (nw);
      return NULL;
    }

  spot->setSlab (nw);
  spot->setPosition (0);

  delete (upLevel);

  if (! preferences->draw_fast)
    refresh ();

  return nw;
}


int
WPDocument::insertCharacter (Spot spot,
			     unsigned long int c,
			     WPStyle *style)
{
  boolean ret;
  unsigned long old_width = getDocumentWidth();
  unsigned long old_height = getDocumentHeight();

  if (read_only)
    {
      gdk_beep();
      return 0;
    }

  /* see gdkkeysyms.h */

  if (c == GDK_space)
    {
      insertSlab (spot, new _STSlab (style));
    }
  else if (c == GDK_Linefeed || c == GDK_Return)
    {
      insertSlab (spot, new _LBTSlab (style, 2, 1));
    }
  else
    {
      Slab first, parent, ths;
      first = spot->getSlab ();
      parent = first->getParent ();

      /* make sure the cursor is in a WordSlab with the correct style */
      ths = spot->getSlab ();

      if (ths->classID () != WordSlabID ||
	  ((WordSlab)ths)->getStyle () != style)
	{
	  Slab nxt, prv;

	  /*nxt = getNextLeaf (ths);*/
	  /*prv = getPrevLeaf (ths);*/

	  nxt = ths->getNextNode ();
	  prv = ths->getPrevNode ();

	  if (spot->getPosition () == ths->getNumberOfChildren () &&
	      nxt != NULL &&
	      nxt->classID () == WordSlabID &&
	      ((WordSlab)nxt)->getStyle () == style)
	    {
	      spot->setSlab (nxt);
	      spot->setPosition (0);
	    }
	  else if (spot->getPosition () == 0 &&
		   prv != NULL &&
		   prv->classID () == WordSlabID &&
		   ((WordSlab)prv)->getStyle () == style)
	    {
	      spot->setSlab (prv);
	      spot->setPosition (prv->getNumberOfChildren ());
	    }
	  else
	    if (createFollowingWord (spot, style) == NULL)
	      return 0;
	}

      ret = spot->getSlab ()->insertChild (spot->getPosition (),
					   new _Character (c),
					   spots,
					   get_document_views (),
					   debug);


      if (ret == false)
	{
	  /* the word is too long for a line or max_children has
	     been reached.  create a follower and retry the insert */

	  if (createFollowingWord (spot, style) == NULL)
	    return 0;

	  ret = spot->getSlab ()->insertChild (spot->getPosition (),
					       new _Character (c),
					       spots,
					       get_document_views (), debug);
	  if (ret == false)
	    {
	      fprintf (stderr, "insert character failed.\n");
	      return 0;
	    }
	}


      /* if the document changed size thanks to this insertion,
	 notify the views that their scrollbars need to update. */
      if (getDocumentWidth() != old_width ||
	  getDocumentHeight() != old_height)
	{
	  for (GList *v = get_document_views (); v != NULL; v = v->next)
	    wp_view_sync_scrollbars((WPView*)v->data);
	}
    }

  if (! preferences->draw_fast)
    refresh ();

  if (debug >= 0)
    printf ("\n");

  return 1;
}


int WPDocument::deleteChild (Spot spot)
{
  if (read_only)
    {
      gdk_beep();
      return 0;
    } 

  unsigned long old_width = getDocumentWidth();
  unsigned long old_height = getDocumentHeight();

  Slab prnt = spot->getSlab ();

  if (prnt->deleteChild (spot->getPosition (),
			 spots, get_document_views (), debug, false) == NULL)
    return 0;

  /* this crud should probably be somewhere in the slab code,
     but I can't get it to fit without causing problems. */

  prnt = spot->getSlab ();
  Slab pprnt = prnt->getParent ();
  if (pprnt != NULL &&
      pprnt->classID () == ContSlabID &&
      pprnt->getNumberOfChildren () > 0 &&
      pprnt->lastLowestContainer ()->isLast () == 0 /*&&
      pprnt->getNextNode ()->firstLowestContainer ()->isFirst () == 0 */)
    ((ContSlab) pprnt)->attemptToStealChildren (spots,
						get_document_views (), debug);

  Slab upLevelPrev = prnt->getPrevNode ();
  if (upLevelPrev != NULL &&
      upLevelPrev->classID () == ContSlabID &&
      upLevelPrev->getNumberOfChildren () > 0 &&
      upLevelPrev->lastLowestContainer ()->isLast () == 0 &&
      prnt->firstLowestContainer ()->isFirst () == 0)
    ((ContSlab) upLevelPrev)->attemptToStealChildren (spots,
						      get_document_views (),
						      debug);

  /* if the document changed size thanks to this deletion,
     notify the views that their scrollbars need to update. */
  if (getDocumentWidth() != old_width ||
      getDocumentHeight() != old_height)
    {
      for (GList *v = get_document_views (); v != NULL; v = v->next)
	wp_view_sync_scrollbars ((WPView*) v->data);
    }

  if (! preferences->draw_fast)
    refresh ();

  return 1;
}


int WPDocument::ensureBorder (Spot spot)
{
  if (read_only)
    {
      gdk_beep();
      return 0;
    } 

  Slab slb = spot->getSlab ();
  int pos = spot->getPosition ();

  int ret = slb->ensureBorder (pos, spots, get_document_views (), debug);

  if (! preferences->draw_fast)
    refresh ();

  return ret;
}


void WPDocument::setSpotToPoint (Spot spot, WPView */*view*/, int x, int y)
{
  Spot t = root->getSpotOfPoint (x - root->getMLeft (),
				 y - root->getMUp ());

  if (t == NULL || spot == NULL) return;

  spot->setSlab (t->getSlab ());
  spot->setPosition (t->getPosition ());

  delete t;
}

void WPDocument::getPointFromSpot (Spot spot, WPView */*view*/,
				   int *x, int *y, int *height)
{
  WpPoint point;

  point = spot->getSlab()->getPointOfChild(spot->getPosition());

  *x = point.x;
  *y = point.y;
  /**height = spot->getSlab()->getParent()->getUp();*/
  *height = spot->getSlab()->getUp();
  if (*height == 0)
    (*height) = - (spot->getSlab()->getDown());
}

Slab WPDocument::getRoot()
{
  return root;
}

char*
WPDocument::getDocumentName ()
{
  if (document_name)
    return document_name;
  return "";
}


void WPDocument::setDocumentName (char *new_name)
{
  if (!new_name) return;

  if (document_name)
    free (document_name);
  document_name = strdup (new_name);
  gnome_mdi_child_set_name (GNOME_MDI_CHILD (mdi_doc), document_name);
}


unsigned long int WPDocument::getDocumentLeft ()
{
  return root->getMLeft ();
}

unsigned long int WPDocument::getDocumentUp ()
{
  return root->getMUp ();
}

unsigned long int WPDocument::getDocumentRight ()
{
  return root->getMRight ();
}

unsigned long int WPDocument::getDocumentDown ()
{
  return root->getMDown ();
}


unsigned long int WPDocument::getDocumentWidth ()
{
  return root->getMLeft () + root->getMRight ();
}


unsigned long int WPDocument::getDocumentHeight ()
{
  return root->getMUp () + root->getMDown ();
}

int WPDocument::getReadOnly()
{
  return read_only;
}

void WPDocument::setReadOnly(int flag)
{
  read_only = flag;
}

void WPDocument::displayRefreshOn ()
{
  display_refresh = 1;
}

void WPDocument::displayRefreshOff ()
{
  display_refresh = 0;
}

int WPDocument::displayRefreshIsOn ()
{
  return display_refresh;
}


int WPDocument::moveToPreviousCharacter (Spot spot)
{
  if (spot->getPosition () > 0)
    {
      spot->sink ();
      spot->setPosition (spot->getPosition () - 1);
      return 1;
    }

  if (!spot->getSlab ())
    return 0;

  Slab prev = spot->getSlab ()->getPrevLeaf ();
  if (prev == NULL)
    return 0;

  spot->setSlab (prev);
  spot->setPosition (prev->getNumberOfChildren ());
  moveToPreviousCharacter (spot);
  return 1;
}


int WPDocument::moveToNextCharacter (Spot spot)
{
  if (spot->getPosition () < spot->getSlab ()->getNumberOfChildren ())
    {
      spot->sink ();
      spot->setPosition (spot->getPosition () + 1);
      return 1;
    }

  if (! spot->getSlab ())
    return 0;

  Slab next = spot->getSlab ()->getNextLeaf ();
  if (next == NULL)
    return 0;

  spot->setSlab (next);
  spot->setPosition (0);
  moveToNextCharacter (spot);
  return 1;
}


int WPDocument::moveToPreviousWord (Spot spot)
{
  if (!spot) return 0;
  if (!spot->getSlab ()) return 0;

  Slab prev;
  for (prev = spot->getSlab ()->getPrevLeaf ();
       prev != NULL && prev->classID () != WordSlabID;
       prev = prev->getPrevLeaf ())
    ;

  if (prev == NULL) return 0;

  spot->setSlab (prev);
  spot->setPosition (prev->getNumberOfChildren ());
  return 1;
}


int WPDocument::moveToNextWord (Spot spot)
{
  if (!spot) return 0;
  if (!spot->getSlab ()) return 0;

  Slab next;
  for (next = spot->getSlab ()->getNextLeaf ();
       next != NULL && next->classID () != WordSlabID;
       next = next->getNextLeaf ())
    ;

  if (next == NULL) return 0;

  spot->setSlab (next);
  spot->setPosition (0);
  return 1;
}



int WPDocument::moveToStartOfLine (Spot spot)
{
  /* XXX should return 0 if spot doesn't move */
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();

  spot->setSlab (upLevel->getSlab ()->firstLowestContainer ());
  spot->setPosition (0);

  delete upLevel;

  if (spot->getSlab ()->classID () == LBTSlabID)
    {
      if (! spot->getSlab ()) return 0;

      Slab nxt = spot->getSlab ()->getNextLeaf ();
      if (nxt != NULL) /* don't crash if CR is last thing in doc */
	spot->setSlab (nxt);
      else
	spot->setPosition (1);

      return 1;
    }

  return 0;
}


int WPDocument::moveToEndOfLine (Spot spot)
{
  /* XXX should return 0 if spot doesn't move */
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();

  spot->setSlab (upLevel->getSlab ()->lastLowestContainer ());
  spot->setPosition (spot->getSlab ()->getNumberOfChildren ());

  delete upLevel;

  return 1;
}


int WPDocument::moveToPreviousLine (Spot spot)
{
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();
  WpPoint p = spot->getSlab ()->getPointOfChild (spot->getPosition ());

  Slab upPrev = upLevel->getSlab ()->getPrevNode ();

  if (upPrev == NULL)
    return 0;

  Spot newSpot = upPrev->getSpotOfPoint (p.x, p.y);

  if (! newSpot) return 0;
  if (! newSpot->getSlab ()) return 0;

  spot->setSlab (newSpot->getSlab ());
  spot->setPosition (newSpot->getPosition ());

  delete newSpot;
  delete upLevel;

  return 1;
}


int WPDocument::moveToNextLine (Spot spot)
{
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();
  WpPoint p = spot->getSlab ()->getPointOfChild (spot->getPosition ());

  Slab upNext = upLevel->getSlab ()->getNextNode ();

  if (upNext == NULL)
    return 0;

  Spot newSpot = upNext->getSpotOfPoint (p.x, p.y);

  if (! newSpot) return 0;
  if (! newSpot->getSlab ()) return 0;

  spot->setSlab (newSpot->getSlab ());
  spot->setPosition (newSpot->getPosition ());

  delete newSpot;
  delete upLevel;

  return 1;
}


int WPDocument::moveToPreviousLine (Spot spot,
				    int preferred_x_pos, int preferred_y_pos)
{
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();

  Slab upLevelParent = upLevel->getSlab ()->getParent ();
  if (! upLevelParent) return 0;

  WpPoint p = upLevelParent->getPointOfChild (upLevel->getSlab ());

  Slab upPrev = upLevel->getSlab ()->getPrevNode ();

  if (upPrev == NULL)
    return 0;

  Spot newSpot;
  switch (upLevelParent->getStyle ()->getDirection ())
    {
    case DIR_LEFT:
    case DIR_RIGHT:
      newSpot = upPrev->getSpotOfPoint (0, preferred_y_pos - p.y);
      break;
    default:
      newSpot = upPrev->getSpotOfPoint (preferred_x_pos - p.x, 0);
      break;
    }

  if (! newSpot) return 0;
  if (! newSpot->getSlab ()) return 0;

  spot->setSlab (newSpot->getSlab ());
  spot->setPosition (newSpot->getPosition ());

  delete newSpot;
  delete upLevel;

  return 1;
}


int WPDocument::moveToNextLine (Spot spot,
				int preferred_x_pos, int preferred_y_pos)
{
  Spot upLevel = new _Spot (spot->getSlab (), spot->getPosition ());
  upLevel->rise ();

  Slab upLevelParent = upLevel->getSlab ()->getParent ();
  if (! upLevelParent) return 0;

  WpPoint p = upLevelParent->getPointOfChild (upLevel->getSlab ());

  Slab upNext = upLevel->getSlab ()->getNextNode ();

  if (upNext == NULL)
    return 0;

  Spot newSpot;
  switch (upLevelParent->getStyle ()->getDirection ())
    {
    case DIR_LEFT:
    case DIR_RIGHT:
      newSpot = upNext->getSpotOfPoint (0, preferred_y_pos - p.y);
      break;
    default:
      newSpot = upNext->getSpotOfPoint (preferred_x_pos - p.x, 0);
      break;
    }

  if (! newSpot) return 0;
  if (! newSpot->getSlab ()) return 0;

  spot->setSlab (newSpot->getSlab ());
  spot->setPosition (newSpot->getPosition ());

  delete newSpot;
  delete upLevel;

  return 1;
}


/* this makes sure that all spots are actually
   contained in the document.  this shouldn't be
   needed, but it is easier than tracking down a couple
   of nasty bugs right now.

   returns 0 if all spots are ok.
   if there was a problem with a spot, it moves the spot
   to the beginning of the document and return 1 */
int WPDocument::sanityCheckSpots ()
{
  int ret = 0;

  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot spot = (Spot) g_list_nth (spots, spotnum)->data;

      if (spot == NULL)
	{
	  printf ("WPDocument -- sanityCheckSpots -- null spot.\n");
	  continue;
	}

      Slab slab = spot->getSlab ();
      int pos = spot->getPosition ();

      if (slab == NULL || pos < 0 || pos > slab->getNumberOfChildren ())
	{
	  spot->setSlab (root->firstLowestContainer ());
	  spot->setPosition (0);
	  fprintf (stderr, "OOPS!  a spot got lost. (0) moving it to "
		   "the beginning of the document.\n");
	  ret = 1;
	}

      while (slab->getParent () != NULL)
	slab = slab->getParent ();

      if (slab == root)
	return 0;

      spot->setSlab (root->firstLowestContainer ());
      spot->setPosition (0);
      fprintf (stderr, "OOPS!  a spot got lost.  (1) moving it to "
	       "the beginning of the document.\n");
      ret = 1;
    }

  return ret;
}


/* free what this returns */
char *wp_get_text_between_spots (Spot start, Spot end)
{
  /* XXX add an flags to this function that controls
     if the character count has to match indexs.  combine
     this with wp_convert_slab_to_clipboard.  if the
     character index's don't matter, we can put things
     like <image> and list bullets in, otherwise, we can't */

  int buf_len = 0;
  int buf_alloced = 10;
  char *buf = (char *) malloc (buf_alloced);

  /* make sure start and end are in the order suggested by their names */
  if (! wp_spot_order (start, end))
    {
      Spot t = start; start = end; end = t;
    }

  Slab cur;
  int pos;
  int get = 0;

  /* convert text from tree format to a char buffer format */

  for (cur = start->getSlab (), pos = start->getPosition ();
       cur && wp_slab_order (cur, end->getSlab ());
       cur = cur->getNextLeaf (), pos = 0)
    {
      char *word;
      int word_len;

      /* convert current slab to text */
      if (cur->classID () == WordSlabID)
	{
	  WordSlab ws = (WordSlab) cur;
	  word = ws->getWord ();
	  word_len = ws->getWordLength ();
	}
      else if (cur->classID () == STSlabID)
	{
	  word = " ";
	  word_len = 1;
	}
      else if (cur->classID () == LBTSlabID)
	{
	  word = "\n";
	  word_len = 1;
	}
      /*
      else if (cur->classID () ==  ImageSlabID)
	{
	  word = ".";
	  word_len = 1;
	}
      */
      /*
      else if (cur->classID () ==  ContSlabID)
	{
	  ContSlab ds = (ContSlab) s;
	  word = ds->drawASCII (doc, 0, 0);
	  word_len = strlen (word);
	}
      */
      else
	{
	  /* not a textual leaf -- skip it */
	  int i;
	  int num = cur->getNumberOfChildren ();

	  static char *word_buf = NULL;
	  static int word_buf_len = 0;

	  if (word_buf == NULL)
	    {
	      word_buf = (char *) malloc (num+1);
	      word_buf_len = num;
	    }
	  else if (num > word_buf_len)
	    {
	      word_buf = (char *) realloc (word_buf, num+1);
	      word_buf_len = num;
	    }

	  for (i=0; i<num; i++)
	    word_buf[ i ] = '.';
	  word_buf[ num ] = '\0';

	  word = word_buf;
	  word_len = num;

	  /*
	  cur = cur->getNextLeaf ();
	  pos = 0;
	  continue;
	  */
	}

      /* figure out how many characters we want from this slab */
      if (cur == end->getSlab ())
	get = end->getPosition ();
      else
	get = word_len;

      if (word_len - pos < get)
	get = word_len - pos;

      /* make sure there is enough room to copy */
      if (buf_len + word_len >= buf_alloced)
	{
	  buf_alloced += 1024;
	  buf = (char *) realloc (buf, buf_alloced);
	}

#     if SPEW
      printf ("      get_text_between_spots cur=");
      cur->dd ();
      printf (", get=%d\n", get);
#     endif

      /* copy what we can */
      int i;
      for (i=0; i<get; i++)
	buf[ buf_len++ ] = word[ pos + i ];
      buf[ buf_len ] = '\0';
    }

  return buf;
}


int WPDocument::spot_to_index (Spot spot)
{
  Slab slab = spot->getSlab ();
  int index = spot->getPosition ();

# if SPEW
  printf ("      spot_to_index, spot->position=%d, spot->slab=", index);
  slab->dd ();
  printf ("\n");
# endif

  slab = slab->getPrevLeaf ();

  while (slab)
    {
#     if SPEW
      printf ("      spot_to_index, slab=");
      slab->dd ();
#     endif

      index += slab->getNumberOfChildren ();

#     if SPEW
      printf (", index=%d\n", index);
#     endif
      slab = slab->getPrevLeaf ();
    }

  return index;
}


void WPDocument::set_spot_to_index (Spot spot, int index)
{
  Slab cur = getRoot ()->firstLowestContainer ();

  while (cur && index > cur->getNumberOfChildren ())
    {
#     if SPEW
      printf ("      set_spot_to_index, index=%d, cur=", index);
      cur->dd ();
      printf (", num_children=%d\n", cur->getNumberOfChildren ());
#     endif

      index -= cur->getNumberOfChildren ();
      cur = cur->getNextLeaf ();
    }

  if (cur)
    {
      if (index > cur->getNumberOfChildren ())
	{
	  index = cur->getNumberOfChildren ();
	  printf ("        forcing index to %d\n", index);
	}
      spot->setSlab (cur);
      spot->setPosition (index);

#     if SPEW
      printf ("      set_spot_to_index done: index=%d, cur=", index);
      cur->dd ();
      printf ("\n");
#     endif
    }
  else
    {
      /* something is bogus.  set spot to start of doc */
      printf ("      WPDocument::set_spot_to_index -- something is busted.\n");
      spot->setSlab (getRoot ()->firstLowestContainer ());
      spot->setPosition (0);
    }
}


/*
** return 1 if 'a' is closer to the beginning of the document
** than 'b'.  otherwise, return 0.
*/
int wp_spot_order (Spot a, Spot b)
{
  if (a->getSlab () == b->getSlab ())
    {
      if (a->getPosition () < b->getPosition ())
	return 1;
      return 0;
    }

  return wp_slab_order (a->getSlab (), b->getSlab ());
}


/*
** return 1 if 'a' is closer to the beginning of the document
** than 'b'.  otherwise, return 0.
** if 'a' and 'b' are the same, return 1.
*/
int wp_slab_order (Slab a, Slab b)
{
  if (!a) return 0;
  if (!b) return 0;

  int a_depth = a->getDepth ();
  int b_depth = b->getDepth ();

  if (a_depth != b_depth)
    {
      int min_depth = (a_depth < b_depth) ? a_depth : b_depth;

      while (min_depth < a_depth)
	{
	  a = a->getParent ();
	  a_depth --;
	}

      while (min_depth < b_depth)
	{
	  b = b->getParent ();
	  b_depth --;
	}
    }

  /* make sure we are not at the root */
  if (a_depth == 1)
    return 1;

  if (a == b)
    return 1;

  while (a->getParent () != b->getParent ())
    {
      a = a->getParent ();
      b = b->getParent ();
    }

  Slab p = a->getParent ();

  int a_index = p->getIndexOfChild (a);
  int b_index = p->getIndexOfChild (b);

  if (a_index <= b_index)
    return 1;

  return 0;
}


/*
**  return 1 if b is inclusively between a and c
**  return 0 otherwise.
*/
int wp_slab_is_between (Slab a, Slab b, Slab c)
{
  if (a == b || b == c)
    return 1;

  if (wp_slab_order (a, b))
    return wp_slab_order (b, c);

  if (wp_slab_order (c, b))
    return wp_slab_order (b, a);

  return 0;
}
