/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Camplugin, MA 02139, USA.
 */

#include <gtk/gtk.h>
#include "wp_plugin_manager.h"
#include "wp_plugin.h"
#include "wp.h"
#include "wp_mdi_document.h"

static WPPlugin *load_file_plugin = NULL;
static WPPlugin *insert_file_plugin = NULL;
static WPPlugin *save_file_plugin = NULL;


WPPlugin *lookup_plugin_by_filename (char *filename,
				     int load_flag,
				     int insert_flag,
				     int save_flag);


void
file_load_by_extension_callback (GtkWidget */*w*/,
				 gpointer   /*client_data*/)
{
  load_file_plugin = NULL;
}


void
file_insert_by_extension_callback (GtkWidget */*w*/,
				   gpointer   /*client_data*/)
{
  insert_file_plugin = NULL;
}


void
file_save_by_extension_callback (GtkWidget */*w*/,
				 gpointer   /*client_data*/)
{
  save_file_plugin = NULL;
}


void
file_load_by_plugin_callback (GtkWidget *w,
			      gpointer   /*client_data*/)
{
  load_file_plugin =
    (WPPlugin*) g_hash_table_lookup (WPPluginManager::plugins_hash,
				     gtk_widget_get_name (w));
}


void
file_insert_by_plugin_callback (GtkWidget *w,
				gpointer   /*client_data*/)
{
  insert_file_plugin =
    (WPPlugin*) g_hash_table_lookup (WPPluginManager::plugins_hash,
				     gtk_widget_get_name (w));
}


void
file_save_by_plugin_callback (GtkWidget *w,
			      gpointer   /*client_data*/)
{
  save_file_plugin =
    (WPPlugin*) g_hash_table_lookup (WPPluginManager::plugins_hash,
				     gtk_widget_get_name (w));
}




static GtkWidget *get_load_menu ()
{
  GList* plugins = WPPluginManager::plugins;
  GList* cur;
  GtkWidget* load_menu;
  GtkWidget* ext_item;

  load_menu = gtk_menu_new();
  
  ext_item = gtk_menu_item_new_with_label(_("By File Extension"));
  gtk_menu_append(GTK_MENU(load_menu), ext_item);
  gtk_widget_show(ext_item);
  gtk_signal_connect(GTK_OBJECT(ext_item),
		     "activate",
		     (GtkSignalFunc)file_load_by_extension_callback,
		     NULL);

  for (cur = plugins; cur != NULL; cur = cur->next)
    {
      WPPlugin *plugin = (WPPlugin *) cur->data;

      GtkWidget *item;

      if (! plugin->canLoad ())
	continue;

      item = gtk_menu_item_new_with_label (plugin->getUIName ());
      gtk_menu_append (GTK_MENU (load_menu), item);
      gtk_widget_show (item);
      gtk_widget_set_name (item, plugin->getUIName ());
      gtk_signal_connect (GTK_OBJECT (item),
			  "activate",
			  (GtkSignalFunc) file_load_by_plugin_callback,
			  NULL);
    }

  return load_menu;
}


static GtkWidget *get_insert_menu ()
{
  GList* plugins = WPPluginManager::plugins;
  GList* cur;
  GtkWidget* insert_menu;
  GtkWidget* ext_item;

  insert_menu = gtk_menu_new();
  
  ext_item = gtk_menu_item_new_with_label(_("By File Extension"));
  gtk_menu_append (GTK_MENU (insert_menu), ext_item);
  gtk_widget_show (ext_item);
  gtk_signal_connect(GTK_OBJECT(ext_item),
		     "activate",
		     (GtkSignalFunc)file_insert_by_extension_callback,
		     NULL);

  for (cur = plugins; cur != NULL; cur = cur->next)
    {
      WPPlugin *plugin = (WPPlugin *) cur->data;

      GtkWidget *item;

      if (! plugin->canInsert ())
	continue;

      item = gtk_menu_item_new_with_label (plugin->getUIName ());
      gtk_menu_append (GTK_MENU (insert_menu), item);
      gtk_widget_show (item);
      gtk_widget_set_name (item, plugin->getUIName ());
      gtk_signal_connect (GTK_OBJECT (item),
			  "activate",
			  (GtkSignalFunc) file_insert_by_plugin_callback,
			  NULL);
    }

  return insert_menu;
}


static GtkWidget *get_save_menu ()
{
  GList* plugins = WPPluginManager::plugins;
  GList* cur;
  GtkWidget* save_menu;
  GtkWidget* ext_item;

  save_menu = gtk_menu_new();

  ext_item = gtk_menu_item_new_with_label(_("By File Extension"));
  gtk_menu_append(GTK_MENU(save_menu), ext_item);
  gtk_widget_show(ext_item);
  gtk_signal_connect(GTK_OBJECT(ext_item),
		     "activate",
		     (GtkSignalFunc)file_save_by_extension_callback,
		     NULL);

  for (cur = plugins;
       cur != NULL;
       cur = cur->next)
    {
      WPPlugin *plugin = (WPPlugin*)cur->data;
      GtkWidget *item;

      if (! plugin->canSave ())
	continue;

      item = gtk_menu_item_new_with_label (plugin->getUIName ());
      gtk_menu_append (GTK_MENU (save_menu), item);
      gtk_widget_show (item);
      gtk_widget_set_name (item, plugin->getUIName ());
      gtk_signal_connect (GTK_OBJECT (item),
			  "activate",
			  (GtkSignalFunc) file_save_by_plugin_callback,
			  NULL);
    }

  return save_menu;
}


WPPlugin *lookup_plugin_by_filename (char *filename,
				     int load_flag,
				     int insert_flag,
				     int save_flag)
{
  GList* plugins = WPPluginManager::plugins;
  GList* cur;

  if (!filename)
    return NULL;

  for (cur = plugins; cur != NULL; cur = cur->next)
    {
      WPPlugin *plugin = (WPPlugin *) cur->data;
      if (!plugin)
	continue;

      if (load_flag && (! plugin->canLoad ()))
	continue;
      if (insert_flag && (! plugin->canInsert ()))
	continue;
      if (save_flag && (! plugin->canSave ()))
	continue;

      char *plugin_ext = plugin->getFileExtension ();
      char *filename_ext = strrchr (filename, '.');

      if (plugin_ext == NULL || filename_ext == NULL)
	continue;

      if (strcasecmp (plugin_ext, filename_ext) == 0)
	return plugin;
    }

  return NULL;
}
	      

static void
file_open_ok_callback(GtkWidget */*w*/,
		      gpointer client_data)
{
  GtkFileSelection *fs;
  char *filename;
  char *raw_filename;

  fs = GTK_FILE_SELECTION (client_data);
  filename = gtk_file_selection_get_filename (fs);
  raw_filename = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));
  
  gtk_widget_hide (GTK_WIDGET (fs));

  if (! load_file_plugin)
    {
      /* if load_file_plugin is null, then the user wants
	 gwp to guess the plugin to use by the file name extension */
      load_file_plugin = lookup_plugin_by_filename (filename, 1, 0, 0);
    }

  if (! load_file_plugin)
    {
      GtkWidget *dialog = gnome_error_dialog (_("Unknown File Type."));
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run (GNOME_DIALOG (dialog));
      return;
    }


  WPMDIDocument *child = wp_mdi_document_load (load_file_plugin, filename);
  gnome_mdi_add_child (mdi, GNOME_MDI_CHILD (child));
  gnome_mdi_add_view (mdi, GNOME_MDI_CHILD (child));
}


static void
file_insert_ok_callback(GtkWidget */*w*/,
			gpointer client_data)
{
  GtkFileSelection *fs;
  char *filename;
  char *raw_filename;

  fs = GTK_FILE_SELECTION (client_data);
  filename = gtk_file_selection_get_filename (fs);
  raw_filename = gtk_entry_get_text (GTK_ENTRY (fs->selection_entry));
  
  gtk_widget_hide (GTK_WIDGET (fs));

  if (! insert_file_plugin)
    {
      /* if insert_file_plugin is null, then the user wants
	 gwp to guess the plugin to use by the file name extension */
      insert_file_plugin = lookup_plugin_by_filename (filename, 0, 1, 0);
    }

  if (! insert_file_plugin)
    {
      GtkWidget *dialog = gnome_error_dialog (_("Unknown File Type."));
      gtk_window_set_modal (GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run (GNOME_DIALOG (dialog));
      return;
    }

  WPView *view = WP_VIEW (mdi->active_view);
  insert_file_plugin->insertDocument (filename, view);
}


static void
file_save_ok_callback (GtkWidget */*w*/,
		       gpointer client_data)
{
  GtkFileSelection *fs;
  char *filename;
  char *raw_filename;

  fs = GTK_FILE_SELECTION (client_data);
  filename = gtk_file_selection_get_filename (fs);
  raw_filename = gtk_entry_get_text (GTK_ENTRY(fs->selection_entry));

  gtk_widget_hide(GTK_WIDGET(fs));  

  if (! save_file_plugin)
    {
      /* if save_file_plugin is null, then the user wants
	 gwp to guess the plugin to use by the file name extension */
      save_file_plugin = lookup_plugin_by_filename (filename, 0, 0, 1);
    }

  if (! save_file_plugin)
    {
      GtkWidget *dialog = gnome_error_dialog (_("Unknown File Type."));
      gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
      gnome_dialog_run (GNOME_DIALOG (dialog));
      return;
    }

  WPView *view = WP_VIEW (mdi->active_view);
  WPDocument *doc = wp_view_get_document (view);
  save_file_plugin->saveDocument (doc, filename);
}


static void
file_cancel_callback(GtkWidget */*w*/,
		     gpointer client_data)
{
  GtkWidget *widget = (GtkWidget*) client_data;
  gtk_widget_hide (widget);
}


void 
file_insert_cmd_callback (GtkWidget */*widget*/,
			gpointer /*client_data*/)
{
  GtkWidget *insert_options, *hbox, *label, *option_menu, *insert_menu;
  static GtkWidget *filesel = NULL;

  if (!filesel)
    {
      filesel = gtk_file_selection_new ("Insert Document...");

      insert_options = gtk_frame_new ("Insert Options");
      gtk_frame_set_shadow_type (GTK_FRAME (insert_options),
				 GTK_SHADOW_ETCHED_IN);

      hbox = gtk_hbox_new (FALSE, 1);
      gtk_container_border_width (GTK_CONTAINER (hbox), 5);
      gtk_container_add (GTK_CONTAINER (insert_options), hbox);
      gtk_widget_show (hbox);
	  
      label = gtk_label_new ("Determine file type:");
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
      gtk_widget_show (label);
	  
      option_menu = gtk_option_menu_new ();
      gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
      gtk_widget_show (option_menu);
	  
      insert_menu = get_insert_menu ();
  
      gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), insert_menu);

      GtkFileSelection *gfs = GTK_FILE_SELECTION (filesel);
	  
      gtk_container_add (GTK_CONTAINER (gfs->main_vbox), insert_options);
      gtk_widget_show(insert_options);
	  
      gtk_signal_connect (GTK_OBJECT (gfs->cancel_button), "clicked",
			  (GtkSignalFunc) file_cancel_callback, filesel);
	  
      gtk_signal_connect (GTK_OBJECT (gfs->ok_button), "clicked",
			  (GtkSignalFunc) file_insert_ok_callback, filesel);
    }

  gtk_widget_show(filesel);
}


void 
file_open_cmd_callback (GtkWidget */*widget*/,
			  gpointer /*client_data*/)
{
  GtkWidget *open_options, *hbox, *label, *option_menu, *load_menu;
  static GtkWidget *filesel = NULL;

  if (!filesel)
    {
      filesel = gtk_file_selection_new ("Open Document...");

      open_options = gtk_frame_new ("Open Options");
      gtk_frame_set_shadow_type (GTK_FRAME (open_options),
				 GTK_SHADOW_ETCHED_IN);

      hbox = gtk_hbox_new (FALSE, 1);
      gtk_container_border_width (GTK_CONTAINER (hbox), 5);
      gtk_container_add (GTK_CONTAINER (open_options), hbox);
      gtk_widget_show (hbox);
	  
      label = gtk_label_new ("Determine file type:");
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
      gtk_widget_show (label);
	  
      option_menu = gtk_option_menu_new ();
      gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
      gtk_widget_show (option_menu);
	  
      load_menu = get_load_menu ();
  
      gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), load_menu);

      GtkFileSelection *gfs = GTK_FILE_SELECTION (filesel);
	  
      gtk_container_add (GTK_CONTAINER (gfs->main_vbox), open_options);
      gtk_widget_show(open_options);
	  
      gtk_signal_connect (GTK_OBJECT (gfs->cancel_button), "clicked",
			  (GtkSignalFunc) file_cancel_callback, filesel);
	  
      gtk_signal_connect (GTK_OBJECT (gfs->ok_button), "clicked",
			  (GtkSignalFunc) file_open_ok_callback, filesel);
    }

  gtk_widget_show(filesel);
}


void 
file_save_cmd_callback (GtkWidget */*widget*/,
			gpointer /*client_data*/)
{
}


void 
file_saveas_cmd_callback (GtkWidget */*widget*/,
			  gpointer /*client_data*/)
{
  GtkWidget *save_options, *hbox, *label, *option_menu, *save_menu;
  static GtkWidget *filesel;

  WPView *view = WP_VIEW (mdi->active_view);
  WPDocument *doc = wp_view_get_document (view);
  if (doc == NULL)
    return;

  if (!filesel)
    {
      filesel = gtk_file_selection_new("Save Document...");
	  
      save_options = gtk_frame_new ("Save Options");
      gtk_frame_set_shadow_type (GTK_FRAME (save_options),
				 GTK_SHADOW_ETCHED_IN);
	  
      hbox = gtk_hbox_new (FALSE, 1);
      gtk_container_border_width (GTK_CONTAINER (hbox), 5);
      gtk_container_add (GTK_CONTAINER (save_options), hbox);
      gtk_widget_show (hbox);
	  
      label = gtk_label_new ("Determine file type:");
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);
      gtk_widget_show (label);
	  
      option_menu = gtk_option_menu_new ();
      gtk_box_pack_start (GTK_BOX (hbox), option_menu, TRUE, TRUE, 0);
      gtk_widget_show (option_menu);

      save_menu = get_save_menu ();

      gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), save_menu);

      GtkFileSelection *gfs = GTK_FILE_SELECTION (filesel);

      gtk_container_add (GTK_CONTAINER (gfs->main_vbox), save_options);
      gtk_widget_show(save_options);
	  
      gtk_signal_connect (GTK_OBJECT (gfs->cancel_button), "clicked",
			  (GtkSignalFunc) file_cancel_callback, filesel);
      
      gtk_signal_connect (GTK_OBJECT (gfs->ok_button), "clicked",
			  (GtkSignalFunc) file_save_ok_callback, filesel);
    }

  gtk_widget_show(filesel);
}
