/*
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include "wp_image_slab.h"
#include "wp_view.h"
#include "wp_utils.h"


_ImageSlab::_ImageSlab (WPStyle *set_style,
			GdkImlibImage *set_img, char *set_image_file)
  : _Slab (set_style)
{
  /* ooh... depth might cause us some problems here... */

  left = 0;
  down = 0;

  image_file = NULL;
  if (set_image_file)
    image_file = set_image_file;

  img = set_img;

  /*pix = gdk_imlib_copy_image (img);*/
  pix = NULL;
  if (gdk_imlib_render(img, img->rgb_width, img->rgb_height))
    pix = gdk_imlib_copy_image (img);

  gdk_window_get_size (pix, &right, &up);

  calculateSize (-1);
}


_ImageSlab::~_ImageSlab ()
{
  /* XXX delete img */
  /* XXX delete pix */
  free (image_file);
}


Slab _ImageSlab::newOfThisType ()
{
  ImageSlab s = new _ImageSlab (style, img, image_file);
  s->setPluginData (getPluginData ());
  return s;
}


GdkPixmap *_ImageSlab::getPixmap ()
{
  return pix;
}


void _ImageSlab::calculateSize (int /*debug*/)
{
  /* size doesn't change. */
}


void _ImageSlab::draw (int x, int y, GList *views, int /*debug*/)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;

      /*System->out->println ("ImageSlab drawing at " + x + ", " + y);*/

      wp_view_copy_from_pixmap (g, pix, 0, 0, 
				left+right, up+down,
				x-left, y-up);

      /* draw highlighting (if any) */
      if (wp_view_hl_is_on (g) &&
	  wp_slab_is_between (wp_view_get_point (g)->getSlab (),
			      this,
			      wp_view_get_mark (g)->getSlab ()))
	{
	  wp_view_set_color (g, wp_view_get_cursor_color (g));
	  wp_view_draw_rect (g, x-left, y-up,
			     left+right, up+down);
	  wp_view_draw_rect (g, x-left+1, y-up+1,
			     left+right-2, up+down-2);

	}
    }
}


void _ImageSlab::erase (int x, int y, GList *views)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;
      wp_view_set_color (g, style->getBackground ());
      wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
    }
}


boolean _ImageSlab::insertChild (int /*i*/,
			      Object /*what*/,
			      GList */*spots*/,
			      GList */*views*/,
			      int /*debug*/)
{
  printf ("error -- ImageSlab->insertChild called\n");
  return false;
}


Object _ImageSlab::deleteChild (int i,
				GList *spots,
				GList *views,
				int debug,
				boolean ignore_lost_spots)
{
  if (debug >= 0)
    {
      printf ("%sdeleteChild this=", tabber (debug));
      dd ();
      printf ("i=%d\n", i);
    }

  if (i != 0)
    {
      printf ("error -- ImageSlab->deleteChild, i=%d", i);
      return NULL;
    }

  return parent->deleteChild (parent->getIndexOfChild (this),
			      spots, views, debug, ignore_lost_spots);
}


boolean _ImageSlab::childSizeChange (Slab /*child*/, GList */*spots*/,
				  int /*dLeft*/, int /*dUp*/,
				  int /*dRight*/, int /*dDown*/,
				  GList */*views*/,
				  int /*debug*/)
{
  printf ("error -- ImageSlab->childSizeChange called\n");
  return false;
}


boolean _ImageSlab::ensureBorder (int i, GList */*spots*/,
			       GList */*views*/, int /*debug*/)
{
  if (i<0 || i>1)
    {
      printf ("error  -- ImageSlab->ensureBorder, i=%d\n", i);
      return false;
    }

  return true;
}


int _ImageSlab::getIndexOfChild (Object /*what*/)
{
  printf ("error -- ImageSlab->getIndexOfChild called\n");
  return -1;
}


Slab _ImageSlab::getChildAt (int /*i*/)
{
  printf ("error -- ImageSlab->getIndexOfChild called\n");
  return NULL;
}


void _ImageSlab::insertChildAt (Object /*what*/, int /*i*/)
{
  printf ("error -- ImageSlab->insertChildAt called\n");
}


void _ImageSlab::dumpToScreen (int tab)
{
  for (int t=0; t<tab; t++) printf (" ");
  printf ("Image(%d,%d,%d,%d)\n", left, up, right, down);
}


void _ImageSlab::dd ()
{
  printf ("[<img>]");
}


int _ImageSlab::isLast ()
{
  return 0;
}


int _ImageSlab::isFirst ()
{
  return 0;
}


Slab _ImageSlab::firstLowestContainer () {return this;}
Slab _ImageSlab::lastLowestContainer () {return this;}


Slab _ImageSlab::getNextNode ()
{
  if (parent == NULL)
    return NULL;
  return parent->getNextNodeUp (this, 1);
}
Slab _ImageSlab::getNextNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _ImageSlab::getNextNodeDown (int depth)
{
  if (depth == 0)
    return this;
  return NULL;
}


Slab _ImageSlab::getPrevNode ()
{
  return parent->getPrevNodeUp (this, 1);
}
Slab _ImageSlab::getPrevNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _ImageSlab::getPrevNodeDown (int /*depth*/)
{
  return this;
}


int _ImageSlab::getNumberOfChildren ()
{
  return 1;
}


WpPoint _ImageSlab::getPointOfChild (Slab child)
{
  return getPointOfChild (getIndexOfChild (child));
}


WpPoint _ImageSlab::getPointOfChild (int childNum)
{
  int x=0, y=0;

  if (parent != NULL)
    {
      WpPoint my_point = parent->getPointOfChild (this);
      x = my_point.x;
      y = my_point.y;
    }

  if (childNum == 0)
    return newWpPoint (x, y);

  if (childNum == 1)
    return newWpPoint (x+right, y);

  printf ("error -- ImageSlab->getPointOfChild: childNum != 0\n");

  return newWpPoint (0, 0);
}


Spot _ImageSlab::getSpotOfPoint (int px, int /*py*/)
{
  if (px < left + right)
    return new _Spot (this, 0);

  return new _Spot (this, 1);
}

