/*
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include "wp_lbt_slab.h"
#include "wp_utils.h"

_LBTSlab::_LBTSlab (WPStyle *set_style, int set_levels, int set_has_size)
  : _Slab (set_style)
{
  style = set_style;
  levels = set_levels;

  if (set_has_size)
    has_size = 1;
  else
    has_size = 0;

  calculateSize (-1);
}


Slab _LBTSlab::newOfThisType ()
{
  LBTSlab s = new _LBTSlab (style, levels, has_size);
  s->setPluginData (getPluginData ());
  return s;
}


void _LBTSlab::setHasSize (int v)
{
  if (v)
    has_size = 1;
  else
    has_size = 0;
}


void _LBTSlab::calculateSize (int debug)
{
  GdkFont *f = style->getFont (1); /* 1 so we get points */

  if (has_size && f)
    {
      if (debug >= 0)
	{
	  printf ("%scalculateSize this=", tabber (debug));
	  dd ();
	  printf ("(l,u,r,d): %d %d %d %d\n", left, up, right, down);
	}

      left = 0;
      right = 0;
      up = f->ascent;
      down = f->descent;
    }
  else
    {
      left = 0;
      right = 0;
      up = 0;
      down = 0;
    }
}


void _LBTSlab::draw (int /*x*/, int /*y*/, GList */*views*/, int /*debug*/)
{
  /*printf ("LBTSlab drawing at " + x + ", " + y);*/
}


void _LBTSlab::erase (int /*x*/, int /*y*/, GList */*views*/)
{
}


boolean _LBTSlab::insertChild (int /*i*/, Object /*what*/, GList */*spots*/,
			       GList */*views*/, int /*debug*/)
{
  printf ("error -- LBTSlab->insertChild called\n");
  return false;
}


Object _LBTSlab::deleteChild (int i, GList *spots,
			      GList *views,
			      int debug, boolean ignore_lost_spots)
{
  if (debug >= 0)
    {
      printf ("%sdeleteChild this=", tabber (debug));
      dd ();
      printf ("i=%d\n", i);
    }

  if (i != 0)
    {
      printf ("error -- LBTSlab->deleteChild, i=%d", i);
      return NULL;
    }

  return parent->deleteChild (parent->getIndexOfChild (this),
			      spots, views, debug, ignore_lost_spots);
}


boolean _LBTSlab::childSizeChange (Slab /*child*/, GList */*spots*/,
				   int /*dLeft*/, int /*dUp*/,
				   int /*dRight*/, int /*dDown*/,
				   GList */*views*/, int /*debug*/)
{
  printf ("error -- LBTSlab->childSizeChange called\n");
  return false;
}


boolean _LBTSlab::ensureBorder (int i, GList */*spots*/,
				GList */*views*/, int /*debug*/)
{
  if (i<0 || i>1)
    {
      printf ("error -- LBTSlab->ensureBorder, ");
      printf ("i=%d\n", i);
      return false;
    }

  return true;
}


int _LBTSlab::getIndexOfChild (Object /*what*/)
{
  printf ("error -- LBTSlab::getIndexOfChild called");
  return -1;
}


Object _LBTSlab::getChildAt (int /*i*/)
{
  printf ("error -- LBTSlab->getIndexOfChild called");
  return NULL;
}


void _LBTSlab::insertChildAt (Object /*what*/, int /*i*/)
{
  printf ("error -- LBTSlab->insertChildAt called");
}


void _LBTSlab::dumpToScreen (int tab)
{
  for (int t=0; t<tab; t++) printf ("+-");
  printf ("LineBreak(%d,%d,%d,%d):%d pp:%p",
	  left, up, right, down, levels, plugin_data);

  printf (" <this=%p><parent=%p>\n", this, parent);
}


void _LBTSlab::dd ()
{
  printf ("<br:%d>", levels);
}


int _LBTSlab::isLast ()
{
  return 0;
}


int _LBTSlab::isFirst ()
{
  /*return 2;*/
  return levels;
}


Slab _LBTSlab::firstLowestContainer () {return this;}
Slab _LBTSlab::lastLowestContainer () {return this;}


Slab _LBTSlab::getNextNode ()
{
  if (parent == NULL)
    return NULL;
  return parent->getNextNodeUp (this, 1);
}
Slab _LBTSlab::getNextNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _LBTSlab::getNextNodeDown (int depth)
{
  if (depth == 0)
    return this;
  return NULL;
}


Slab _LBTSlab::getPrevNode ()
{
  return parent->getPrevNodeUp (this, 1);
}
Slab _LBTSlab::getPrevNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _LBTSlab::getPrevNodeDown (int /*depth*/)
{
  return this;
}


int _LBTSlab::getNumberOfChildren ()
{
  return 1;
}


WpPoint _LBTSlab::getPointOfChild (Slab child)
{
  return getPointOfChild (getIndexOfChild ((Object) child));
}


WpPoint _LBTSlab::getPointOfChild (int childNum)
{
  int x=0, y=0;

  if (parent != NULL)
    {
      WpPoint my_point = parent->getPointOfChild (this);
      x = my_point.x;
      y = my_point.y;
    }

  if (childNum == 0 || childNum == 1)
    return newWpPoint (x, y);

  printf ("error -- LBTSlab->getPointOfChild: ");
  printf ("childNum != 0\n");

  return newWpPoint (0, 0);
}


Spot _LBTSlab::getSpotOfPoint (int px, int /*py*/)
{
  if (px < 0)
    return new _Spot (this, 0);

  return new _Spot (this, 1);
}
