/*
 * Copyright  1999 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gnome.h>

#include "wp_mdi_document.h"

#include <stdio.h>
#include <string.h>

#include "wp_plugin.h"

static void       wp_mdi_document_class_init       (WPMDIDocumentClass *);
static void       wp_mdi_document_init             (WPMDIDocument *);
static GtkWidget *wp_mdi_document_create_view      (GnomeMDIChild *);
static void       wp_mdi_document_destroy          (GtkObject *);

static GnomeMDIChildClass *parent_class = NULL;


guint wp_mdi_document_get_type ()
{
  static guint doc_type = 0;
	
  if (!doc_type)
    {
      GtkTypeInfo doc_info =
      {
	"WPMDIDocument",
	sizeof (WPMDIDocument),
	sizeof (WPMDIDocumentClass),
	(GtkClassInitFunc) wp_mdi_document_class_init,
	(GtkObjectInitFunc) wp_mdi_document_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL,
      };
		
      doc_type = gtk_type_unique (gnome_mdi_child_get_type (), &doc_info);
    }
	
  return doc_type;
}


static GtkWidget *wp_mdi_document_create_view (GnomeMDIChild */*child*/)
{
  GtkWidget *view = wp_view_new ();
  /*gtk_widget_show (view);*/
  return view;
}


static void wp_mdi_document_destroy (GtkObject *obj)
{
  printf ("wp_mdi_document_destroy called\n");

  WPMDIDocument *mdi_doc;
  mdi_doc = WP_MDI_DOCUMENT(obj);
  WPDocument *doc = wp_mdi_document_get_document (mdi_doc);
  delete (doc);
}


WPDocument *wp_mdi_document_get_document (WPMDIDocument *mdi_doc)
{
  return mdi_doc->doc;
}


static void wp_mdi_document_class_init (WPMDIDocumentClass *klass)
{
  GtkObjectClass *object_class;
  GnomeMDIChildClass *child_class;

  object_class = (GtkObjectClass *) klass;
  child_class = GNOME_MDI_CHILD_CLASS (klass);

  object_class->destroy = wp_mdi_document_destroy;

  child_class->create_view =
    (GnomeMDIChildViewCreator) (wp_mdi_document_create_view);

  /*
  child_class->get_config_string =
    (GnomeMDIChildConfigFunc) (wp_mdi_document_get_config_string);
  */

  /*klass->document_changed = wp_mdi_document_real_changed;*/

  parent_class =
    (GnomeMDIChildClass *) gtk_type_class (gnome_mdi_child_get_type ());
}


static void wp_mdi_document_init (WPMDIDocument *mdi_doc)
{
  mdi_doc->doc = NULL;
  /*gnome_mdi_child_set_menu_template (GNOME_MDI_CHILD (document), doc_menu);*/
}


WPMDIDocument *wp_mdi_document_new (WPPlugin *plugin, char *title)
{
  g_return_val_if_fail (plugin, NULL);

  WPMDIDocument *mdi_doc;
  mdi_doc = (WPMDIDocument *) gtk_type_new (wp_mdi_document_get_type ());
  mdi_doc->doc = plugin->newDocument (title);
  mdi_doc->doc->set_mdi_document (mdi_doc);

  GNOME_MDI_CHILD (mdi_doc)->name = strdup (mdi_doc->doc->getDocumentName ());

  return mdi_doc;
}


WPMDIDocument *wp_mdi_document_load (WPPlugin *plugin, char *filename)
{
  g_return_val_if_fail (plugin, NULL);

  WPMDIDocument *mdi_doc;
  mdi_doc = (WPMDIDocument *) gtk_type_new (wp_mdi_document_get_type ());
  mdi_doc->doc = plugin->loadDocument (filename);
  mdi_doc->doc->set_mdi_document (mdi_doc);

  GNOME_MDI_CHILD (mdi_doc)->name = strdup (mdi_doc->doc->getDocumentName ());

  mdi_doc->doc->refresh ();

  return mdi_doc;
}
