#include <stdio.h>
#include <stdlib.h>

#include <gnome.h>

#include "wp_prefs.h"


WPPreferences *preferences = NULL;


WPPreferences::WPPreferences ()
{
  draw_fast = gnome_config_get_int ("/gwp/Preferences/DrawFast=0");
  properties_dialog = NULL;
  notebook = NULL;
  button = NULL;

  GdkColormap *colormap = gtk_widget_get_default_colormap ();

  margin_color = (GdkColor *) malloc (sizeof (GdkColor));
  margin_color->red = 0x9000;
  margin_color->green = 0x9000;
  margin_color->blue = 0x9000;
  gdk_color_alloc (colormap, margin_color);
}


WPPreferences::~WPPreferences ()
{
}


void WPPreferences::backup ()
{
  b_draw_fast = draw_fast;
  b_margin_color = margin_color;
}


void WPPreferences::apply ()
{
  draw_fast = b_draw_fast;
  margin_color = b_margin_color;
  printf ("draw_fast is %d\n", draw_fast);

  gnome_config_set_int("/gwp/Preferences/DrawFast", draw_fast);
  gnome_config_sync ();
}


void 
preferences_ok_callback (GtkWidget */*widget*/, gpointer client_data)
{
  GtkWidget *properties_dialog = (GtkWidget *) client_data;
  gtk_widget_hide (properties_dialog);
  preferences->apply ();
}


void 
preferences_apply_callback (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  preferences->apply ();
}


void 
preferences_cancel_callback (GtkWidget */*widget*/, gpointer client_data)
{
  GtkWidget *properties_dialog = (GtkWidget *) client_data;
  gtk_widget_hide (properties_dialog);
}


void 
draw_fast_toggle_callback (GtkWidget *widget, gpointer /*client_data*/)
{
  GtkCheckButton *cb = GTK_CHECK_BUTTON (widget);
  /*preferences->b_draw_fast = ! preferences->b_draw_fast;*/

  preferences->b_draw_fast = cb->toggle_button.active;
}


void WPPreferences::showDialog ()
{
  if (properties_dialog)
    {
      gtk_widget_show (properties_dialog);
      return;
    }

  properties_dialog = gtk_dialog_new();
  notebook = gtk_notebook_new();

  preferences->backup ();

  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) preferences_ok_callback,
                      properties_dialog);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Apply");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) preferences_apply_callback,
                      properties_dialog);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) preferences_cancel_callback,
                      properties_dialog);

  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (properties_dialog)->action_area),
		      button, TRUE, TRUE, 0);
  gtk_widget_show (button);

  {
    GtkWidget *label;
    GtkWidget *frame;
	  
    frame = gtk_frame_new ("Display");
    gtk_container_border_width (GTK_CONTAINER (frame), 10);
    gtk_widget_set_usize (frame, 200, 150);
    gtk_widget_show (frame);

    GtkWidget *draw_fast_toggle =
      gtk_check_button_new_with_label ("Draw Fast");
    gtk_container_add (GTK_CONTAINER (frame), draw_fast_toggle);
    gtk_widget_show (draw_fast_toggle);
    gtk_signal_connect (GTK_OBJECT (draw_fast_toggle), "toggled",
			(GtkSignalFunc) draw_fast_toggle_callback,
			properties_dialog);
    gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (draw_fast_toggle),
				 preferences->b_draw_fast);

    /*
    label = gtk_label_new ("Display0");
    gtk_container_add (GTK_CONTAINER (frame), label);
    gtk_widget_show (label);
    */
	  
    label = gtk_label_new ("Display");
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);
  }

  gtk_box_pack_start (GTK_BOX(GTK_DIALOG (properties_dialog)->vbox),
		      notebook, TRUE, TRUE, 0);
  gtk_widget_show(notebook);

  gtk_widget_show(properties_dialog);
}


void 
preferences_cmd_callback (GtkWidget */*widget*/, gpointer /*client_data*/)
{
  preferences->showDialog ();
}
