
#include <ctype.h>
#include "wp_print.h"
#include "wp_slab.h"
#include "wp_cont_slab.h"
#include "wp_word_slab.h"
#include "wp_image_slab.h"
#include "wp_font.h"


void gwp_print_slab (WPDocument *doc, GnomePrintContext *pc, Slab slab);


/* adjust for which page we are on */
static void gwp_print_adjust_XY (Slab s, long int *x, long int *y)
{
  Slab h = s;
  Slab page_parent;
  long int page_x, page_y;
  long int page_height;

  /*
  while (h->getDepth () > 3 ||
	 h->getStyle ()->getYSizePolicy () != SIZE_POLICY_FIXED)
  */
  while (h->getDepth () > 2)
    {
      h = h->getParent ();
      if (h == NULL)
	break;
    }

  if (! h) /* duh.  something is hosed. */
    return;

  page_height = h->getHeight ();

  /*
  while (s->getDepth () > 2)
    s = s->getParent ();
  */

  page_parent = h->getParent ();
  if (page_parent == NULL)
    {
      page_x = 0;
      page_y = 0;
    }
  else
    {
      WpPoint page_point = page_parent->getPointOfChild (h);
      page_x = page_point.x - h->getMLeft ();
      page_y = page_point.y - h->getMUp ();
    }

  (*x) -= page_x;
  (*y) -= page_y;

  /* PS coords are in first quadrent */
  (*y) = page_height - (*y);
}


#if 0
static int gwp_print_nextIsNotOnSamePage (Slab s)
{
  Slab n = s->getNextNode ();

  if (n == NULL)
    return 1;

  while (s->getDepth () > 2)
    s = s->getParent ();

  while (n->getDepth () > 2)
    n = n->getParent ();

  return s != n;
}
#endif /* 0 */ 

void gwp_print_cont_slab (WPDocument *doc, GnomePrintContext *pc, Slab slab)
{
  int i;
  int num = slab->getNumberOfChildren ();

  for (i=0; i<num; i++)
    {
      Slab child = (Slab) slab->getChildAt (i);
      gwp_print_slab (doc, pc, child);
    }

  ContSlab s = (ContSlab) slab;

  Slab parent = slab->getParent ();
  WpPoint p;

  if (parent == NULL)
    {
      p.x = 0;
      p.y = 0;
    }
  else
    p = parent->getPointOfChild (slab);

  gwp_print_adjust_XY (slab, &p.x, &p.y);

  gnome_print_gsave (pc);
  /*gnome_print_moveto (pc, p.x, p.y);*/
  s->drawPS (doc, pc, p.x, p.y);
  gnome_print_grestore (pc);
}

void gwp_print_page_slab (WPDocument *doc, GnomePrintContext *pc, Slab slab)
{
  gwp_print_cont_slab (doc, pc, slab);
  gnome_print_showpage (pc);
}

void gwp_print_lbt_slab (WPDocument *, GnomePrintContext *, Slab)
{
  return;
}

void gwp_print_word_slab (WPDocument *, GnomePrintContext *pc, Slab slab)
{
  GnomeFont *font;
  WordSlab ws = (WordSlab) slab;
  char *text = ws->getWord ();
  char *fontname = ws->getStyle ()->getFontName ();
  char *fontfam = getFontFamily (fontname); /* XXX */
  float fontsize = (float) getFontPointSize (fontname) / 10.0;
  Slab parent = slab->getParent ();
  WpPoint p;

  if (parent == NULL)
    {
      fprintf (stderr, "gwp_print -- null parent.\n");
      return;
    }

  p = parent->getPointOfChild (slab);
  gwp_print_adjust_XY (slab, &p.x, &p.y);

  /* XXX GROSS!  the font stuff needs to be generally reworked.
     The font name that gnome_font_new takes is case sensitive */
  if (fontfam && fontfam[ 0 ])
    fontfam[ 0 ] = toupper (fontfam[ 0 ]);

  gnome_print_moveto (pc, p.x, p.y);
  font = gnome_font_new (fontfam, fontsize);
  if (font == NULL)
    font = gnome_font_new ("Helvetica", fontsize);
  if (font == NULL)
    return;

  /* set the color */
  GdkColor *c = slab->getStyle ()->getForeground ();
  gnome_print_setrgbcolor (pc,
			   (float) (c->red) / (float) 0xFFFF,
			   (float) (c->green) / (float) 0xFFFF,
			   (float) (c->blue) / (float) 0xFFFF);

  gnome_print_setfont (pc, font);
  gnome_print_show (pc, text);

  if (ws->getStyle ()->getUnderline ())
    {
      gnome_print_newpath (pc);
      gnome_print_moveto (pc, p.x - ws->getMLeft (), p.y + 1);
      gnome_print_lineto (pc, p.x + ws->getMRight (), p.y + 1);
      gnome_print_lineto (pc, p.x + ws->getMRight (), p.y + 2);
      gnome_print_lineto (pc, p.x - ws->getMLeft (), p.y + 2);
      gnome_print_closepath (pc);
      gnome_print_fill (pc);
      gnome_print_stroke (pc);
    }

  gtk_object_unref (GTK_OBJECT (font));
  free (fontfam);
}

void gwp_print_st_slab (WPDocument *, GnomePrintContext *, Slab)
{
  return;
}

void gwp_print_image_slab (WPDocument */*doc*/,
			   GnomePrintContext *pc,
			   Slab slab)
{
  ImageSlab s = (ImageSlab) slab;
  Slab parent = slab->getParent ();
  WpPoint p;

  if (parent == NULL)
    {
      fprintf (stderr, "gwp_print -- null parent.\n");
      return;
    }

  p = parent->getPointOfChild (slab);
  gwp_print_adjust_XY (slab, &p.x, &p.y);

  char *data = (char *) s->get_rgb_data ();
  int width = s->get_rgb_width ();
  int height = s->get_rgb_height ();
  int rowstride = 3 * width;

  double matrix4[6] = {width, 0, 0, height, p.x, p.y};

  gnome_print_gsave (pc);
  gnome_print_concat (pc, matrix4);
  gnome_print_moveto (pc, p.x, p.y);
  gnome_print_rgbimage (pc, data, width, height, rowstride);
  gnome_print_grestore (pc);
}


#if 0
void gwp_print_drawn_slab (WPDocument *doc, GnomePrintContext *pc, Slab slab)
{
  int i;
  int num = slab->getNumberOfChildren ();

  for (i=0; i<num; i++)
    {
      Slab child = (Slab) slab->getChildAt (i);
      gwp_print_slab (doc, pc, child);
    }

  DrawnSlab s = (DrawnSlab) slab;

  Slab parent = slab->getParent ();
  WpPoint p;

  if (parent == NULL)
    {
      fprintf (stderr, "gwp_print -- null parent.\n");
      return;
    }

  p = parent->getPointOfChild (slab);
  gwp_print_adjust_XY (slab, &p.x, &p.y);

  gnome_print_gsave (pc);
  /*gnome_print_moveto (pc, p.x, p.y);*/
  s->drawPS (doc, pc, p.x, p.y);
  gnome_print_grestore (pc);
}
#endif /* 0 */


void gwp_print_slab (WPDocument *doc, GnomePrintContext *pc, Slab slab)
{
  if (!slab)
    return;

  switch (slab->classID ())
    {
    case ContSlabID:
      gwp_print_cont_slab (doc, pc, slab);
      break;
    case LBTSlabID:
      gwp_print_lbt_slab (doc, pc, slab);
      break;
    case WordSlabID:
      gwp_print_word_slab (doc, pc, slab);
      break;
    case STSlabID:
      gwp_print_st_slab (doc, pc, slab);
      break;
    case ImageSlabID:
      gwp_print_image_slab (doc, pc, slab);
      break;
    default:
      /* can't cope */
      break;
    }
}


void gwp_print (GnomePrinter *printer, WPDocument *doc)
{
  GnomePrintContext *pc = gnome_print_context_new (printer);
  gwp_print_slab (doc, pc, doc->getRoot ());
  gnome_print_context_close (pc);
  gtk_object_unref (GTK_OBJECT (pc));
}
