/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright (C) 1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wp_slab.h"
#include <assert.h>

/**
 * _Slab::_Slab: constructor
 * @set_style: style of new Slab
 * 
 * Slab is created.  If @set_style has fixed dimensions, slab's
 * sizes are set accordingly.  Non-fixed dimensions are set to 0
 * 
 * Return Value: new Slab
 **/
_Slab::_Slab (WPStyle *set_style)
{
  parent = NULL;

  left = 0;
  right = 0;
  up = 0;
  down = 0;
  /*
  xSizePolicy = SIZE_POLICY_BY_CHILDREN;
  ySizePolicy = SIZE_POLICY_BY_CHILDREN;
  */
  style = set_style;
  plugin_data = NULL;


  if (style->getXSizePolicy () == SIZE_POLICY_FIXED)
    {
      left = style->getFixedLeft ();
      right = style->getFixedRight ();
    }
  if (style->getYSizePolicy () == SIZE_POLICY_FIXED)
    {
      up = style->getFixedUp ();
      down = style->getFixedDown ();
    }
}


/**
 * _Slab:destructor
 * 
 * does nothing
 * 
 * Return Value: none
 **/
_Slab::~_Slab ()
{
}


/**
 * _Slab::setStyle:set slab's style pointer
 * @set_style: slab's new style
 * @spots: list of spots
 * @views: list of views
 * @debug: verbosity
 * 
 * Changes which style a slab has.  The slab's size and
 * position are recalculated, screen updates are done.
 **/
void _Slab::setStyle (WPStyle *set_style, GList *spots,
		      GList *views, int debug)
{
  WPStyle *save_style;
  int sleft, sup, sright, sdown; /* save size */
  int dleft, dup, dright, ddown; /* size change */

  /**********************/
  /*** force a rewrap ***/
  /**********************/

  save_style = style;
  sleft = left;
  sup = up;
  sright = right;
  sdown = down;

  style = set_style;

  calculateSize (debug);

  dleft = left - sleft;
  dup = up - sup;
  dright = right - sright;
  ddown = down - sdown;

  left = sleft;
  up = sup;
  right = sright;
  down = sdown;

  if (parent)
    {
      int ret = parent->childSizeChange (this, spots,
					 dleft, dup, dright, ddown,
					 views, (debug >= 0)?debug+1:-1);
      if (ret == false)
	{
	  style = save_style;
	  calculateSize (debug);
	  return;
	}
    }

  WpPoint p;

  if (parent)
    p = parent->getPointOfChild (this);
  else
    {
      p.x=0; p.y=0;
    }

  erase (p.x, p.y, views);
  draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
}


/**
 * _Slab::getStyle: get Slab's style
 * 
 * Return Value: Slab's style
 **/
WPStyle *_Slab::getStyle ()
{
  return style;
}


/**
 * _Slab::getLeft:get Slab's left dimension
 * 
 * returns the amount to which this Slab extends
 * left of it's center point, not including it's
 * left margin.
 * 
 * Return Value: Slab's left dimension (in pixels)
 **/
int _Slab::getLeft ()
{
  return left;
}


/**
 * _Slab::getUp:get Slab's up dimension
 * 
 * returns the amount to which this Slab extends
 * upward from it's center point, not including it's
 * top margin
 * 
 * Return Value: Slab's up dimension (in pixels)
 **/
int _Slab::getUp ()
{
  return up;
}


/**
 * _Slab::getRight:get Slab's right dimension
 * 
 * returns the amount to which this Slab extends
 * rigth of it's center point, not including it's
 * right margin
 * 
 * Return Value: Slab's right dimension (in pixels)
 **/
int _Slab::getRight ()
{
  return right;
}


/**
 * _Slab::getDown:get Slab's down dimension
 *
 * returns the amount to which this Slab extends
 * below it's center point, not including it's margins.
 * 
 * Return Value: 
 **/
int _Slab::getDown ()
{
  return down;
}


/**
 * _Slab::getWidth:get slab width
 * 
 * Adds the left and right dimensions of this slab, not
 * including margins.
 * 
 * Return Value: Width of this slab, not including margins.
 **/
int _Slab::getWidth ()
{
  return getLeft () + getRight ();
}


/**
 * _Slab::getHeight:get slab height
 * 
 * Adds the up and down dimensions of the slab, not including
 * the margins.
 * 
 * Return Value: Height of this slab, not including margins.
 **/
int _Slab::getHeight ()
{
  return getUp () + getDown ();
}


/**
 * _Slab::getMLeft:get slab left including margins
 * 
 * Adds Slab's left dimension with its left margin
 * 
 * Return Value: Slab's left dimension, including margin
 **/
int _Slab::getMLeft ()
{
  return left + style->getMarginLeft ();
}


/**
 * _Slab::getMUp:get Slab's up dimension, including margins
 * 
 * Adds Slab's up dimension with its up margin
 * 
 * Return Value: Slab's up dimension, including margin
 **/
int _Slab::getMUp ()
{
  return up + style->getMarginUp ();
}


/**
 * _Slab::getMRight:get Slab's right dimension, including margins
 * 
 * Adds Slab's right dimension with its right margin
 * 
 * Return Value: Slab's right dimension, including margin
 **/
int _Slab::getMRight ()
{
  return right + style->getMarginRight ();
}


/**
 * _Slab::getMDown:get Slab's down dimension, including margins
 * 
 * Adds Slab's down dimension with its down margin
 * 
 * Return Value: Slab's down dimension, including margin
 **/
int _Slab::getMDown ()
{
  return down + style->getMarginDown ();
}


/**
 * _Slab::getMWidth:get Slab's width, including margins
 * 
 * Return Value: Slab's width, including margins
 **/
int _Slab::getMWidth ()
{
  return getMLeft () + getMRight ();
}


/**
 * _Slab::getMHeight:get Slab's height, inlcuding margins
 * 
 * Return Value: Slab's height, including margins
 **/
int _Slab::getMHeight ()
{
  return getMUp () + getMDown ();
}


/**
 * _Slab::setLeft:set Slab's left dimension
 * @l: new dimension
 * 
 * @l doesn't include the slab's margin
 **/
void _Slab::setLeft (int l)
{
  left = l;
}


/**
 * _Slab::setUp:set Slab's up dimension
 * @u: new dimension
 * 
 * @u doesn't include the slab's margin
 **/
void _Slab::setUp (int u)
{
  up = u;
}


/**
 * _Slab::setRight:set Slab's right dimension
 * @r: new dimension
 * 
 * @u doesn't include the slab's margin
 **/
void _Slab::setRight (int r)
{
  right = r;
}


/**
 * _Slab::setDown:set Slab's down dimension
 * @d: new dimension
 * 
 * @d doesn't include the slab's margin
 **/
void _Slab::setDown (int d)
{
  down = d;
}


/**
 * _Slab::setMLeft:set Slab's left dimension, including margin
 * @l: new dimension
 * 
 * @l sets the Slab's total left dimension, including its margin.
 * Slab.left may be smaller than @l.
 **/
void _Slab::setMLeft (int l)
{
  left = l - style->getMarginLeft ();
}


/**
 * _Slab::setMUp:set Slab's up dimension, including margin
 * @u: new dimension
 * 
 * @u sets the Slab's total up dimension, including its margin.
 * Slab.up may be smaller than @u
 **/
void _Slab::setMUp (int u)
{
  up = u - style->getMarginUp ();
}


/**
 * _Slab::setMRight:set Slab's right dimension, including margin
 * @r: new dimension
 * 
 * @r sets the Slab's total right dimension, including its margin.
 * Slab.right may be smaller than @r.
 **/
void _Slab::setMRight (int r)
{
  right = r - style->getMarginRight ();
}


/**
 * _Slab::setMDown:set Slab's down dimension, including margin
 * @d: new dimension
 * 
 * @d sets the Slab's total down dimensino, including its margin.
 * Slab.down may be smaller than @d.
 **/
void _Slab::setMDown (int d)
{
  down = d - style->getMarginDown ();
}


/**
 * _Slab::getMajorSizePolicy:get Slab's major size policy
 * 
 * The major axis of a slab is its style's direction.  Every style has
 * and X and a Y size policy.  If this Slab's direction is left or right,
 * the X size policy of this Slab's style is returned.  Otherwise, the Y
 * size policy is returned.
 * 
 * Return Value: size policy along Slab's major axis.
 **/
WPSizePolicy _Slab::getMajorSizePolicy ()
{
  int direction = style->getDirection ();

  switch (direction)
    {
    case DIR_LEFT:
    case DIR_RIGHT:
      return style->getXSizePolicy ();
    case DIR_UP:
    case DIR_DOWN:
      return style->getYSizePolicy ();
    default:
      assert(0);
      return style->getXSizePolicy (); /* punt */
    }
}


/**
 * _Slab::getMajorLow: get low dimension along parent's major axis
 * @direction: direction of parent's major axis
 * 
 * getMajorLow returns the dimension of this Slab that is opposite of
 * @direction (not including margins)
 * 
 * Return Value: low dimension along parent's major axis
 **/
int _Slab::getMajorLow (int direction)
{
  switch (direction)
    {
    case DIR_LEFT: return getRight ();
    case DIR_UP: return getDown ();
    case DIR_RIGHT: return getLeft ();
    case DIR_DOWN: return getUp ();
    default:
      assert(0);
      return getRight (); /* punt */
    }
}


/**
 * _Slab::getMajorHigh: get high dimension along parent's major axis
 * @direction: direction of parent's major axis
 * 
 * getMajorHigh returns the dimension of this Slab that is along
 * @direction (not including margins)
 * 
 * Return Value: high dimension along parent's major axis
 **/
int _Slab::getMajorHigh (int direction)
{
  switch (direction)
    {
    case DIR_LEFT: return getLeft ();
    case DIR_UP: return getUp ();
    case DIR_RIGHT: return getRight ();
    case DIR_DOWN: return getDown ();
    default:
      assert(0);
      return getLeft (); /* punt */
    }
}



/**
 * _Slab::getMajorLow: get low dimension along parent's major axis
 * @direction: direction of parent's major axis
 * 
 * getMajorLow returns the dimension of this Slab that is opposite of
 * @direction (including margins)
 * 
 * Return Value: low dimension along parent's major axis
 **/
int _Slab::getMMajorLow (int direction)
{
  switch (direction)
    {
    case DIR_LEFT: return getMRight ();
    case DIR_UP: return getMDown ();
    case DIR_RIGHT: return getMLeft ();
    case DIR_DOWN: return getMUp ();
    default:
      assert(0);
      return getRight (); /* punt */
    }
}


/**
 * _Slab::getMajorHigh: get high dimension along parent's major axis
 * @direction: direction of parent's major axis
 * 
 * getMajorHigh returns the dimension of this Slab that is along
 * @direction (including margins)
 * 
 * Return Value: high dimension along parent's major axis
 **/
int _Slab::getMMajorHigh (int direction)
{
  switch (direction)
    {
    case DIR_LEFT: return getMLeft ();
    case DIR_UP: return getMUp ();
    case DIR_RIGHT: return getMRight ();
    case DIR_DOWN: return getMDown ();
    default:
      assert(0);
      return getLeft (); /* punt */
    }
}


/**
 * _Slab::topOfTree:find root of slab tree
 * 
 * This moved up to slab tree toward the root until
 * it finds a slab with no parent
 * 
 * Return Value: root of slab tree
 **/
Slab _Slab::topOfTree ()
{
  if (parent == NULL)
    return this;
    
  return parent->topOfTree ();
}


/**
 * _Slab::getDepth:get Slab's distance from tree's root
 * 
 * if Slab is the root of the tree, 1 is returned.  if Slab is
 * a child of the root, 2 is returned, etc.
 * 
 * Return Value: depth of Slab
 **/
int _Slab::getDepth ()
{
  if (parent == NULL)
    return 1;

  return parent->getDepth () + 1;
}


/**
 * _Slab::setParent:set Slab's parent node
 * @p: new parent node
 * 
 * 
 **/
void _Slab::setParent (Slab p)
{
  parent = p;
}


/**
 * _Slab::getParent:get Slab's parent node
 * 
 * Return Value: Slab's parent node
 **/
Slab _Slab::getParent ()
{
  return parent;
}


/**
 * _Slab::getXSizePolicy:get Slab's style's X size policy
 * 
 * Return Value: Slab's X size policy
 **/
WPSizePolicy _Slab::getXSizePolicy ()
{
  return style->getXSizePolicy ();
}


/**
 * _Slab::getYSizePolicy:get Slab's style's Y size policy
 * 
 * Return Value: Slab's Y size policy
 **/
WPSizePolicy _Slab::getYSizePolicy ()
{
  return style->getYSizePolicy ();
}


/**
 * _Slab::getSpotOfPoint:get the Spot nearest point
 * @p: point on display
 * 
 * Given an offset from the upper left corner of the document,
 * return a Spot in a leaf slab.  This is used mainly for
 * when a user clicks on a document with the mouse
 * 
 * Return Value: Spot within Slab nearest point
 **/
Spot _Slab::getSpotOfPoint (WpPoint p)
{
  return getSpotOfPoint (p.x, p.y);
}


/**
 * _Slab::isLast:does Slab prefer to be last?
 * 
 * Returns 1 if this Slab would like to be the last slab in
 * its parent's list of children.
 * 
 * Return Value: 1 if Slab prefers to be last, 0 otherwise
 **/
int _Slab::isLast ()
{
  return 0;
}


/**
 * _Slab::isFirst:does Slab prefer to be first
 * 
 * Returns 1 if this Slab would like to be the first slab in
 * its parent's list of children
 * 
 * Return Value: 1 if Slab prefers to be first, 0 otherwise
 **/
int _Slab::isFirst ()
{
  return 0;
}


/**
 * _Slab::styleChangeNotify:Adjust Slab after a style modification
 * @s: style that was changed
 * @spots: list of spots in the document
 * @views: list of views displaying the document
 * @debug: control's debugging spew (0 for lots, -1 for none).
 * 
 * call this when something in a style has changed (not to be
 * confused with switching a Slab to a different style).
 * 
 * 
 * Return Value: 1 if anything changed, 0 otherwise
 **/
int _Slab::styleChangeNotify (WPStyle *s, GList *spots,
			      GList *views, int debug)
{
  if (style == s && parent != NULL)
    {
      int ret;
      WpPoint p;
      int sleft, sup, sright, sdown;
      int dleft, dup, dright, ddown;

      sleft = left;
      sup = up;
      sright = right;
      sdown = down;;

      calculateSize ((debug >= 0)?debug+1:-1);

      dleft = left - sleft;
      dup = up - sup;
      dright = right - sright;
      ddown = down - sdown;

      left = sleft;
      up = sup;
      right = sright;
      down = sdown;;


      ret = parent->childSizeChange (this, spots,
				     dleft, dup, dright, ddown,
				     views, (debug >= 0)?debug+1:-1);

      if (ret == false)
	{
	  /* XXX what to do here?! */
	  fprintf (stderr, "resize during styleChangeNotify failed\n");
	}

      p = parent->getPointOfChild (this);

      erase (p.x, p.y, views);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
      return 1;
    }

  return 0;
}


/**
 * _Slab::slabIsInside:is Slab contained withing this one?
 * @s: Child Slab
 * 
 * return 1 if @s is 'this' or a child of
 * this or a child of a child of this, etc
 * 
 * Return Value: 1 if @s is within this, 0 otherwise
 **/
int _Slab::slabIsInside (Slab s)
{
  while (s != NULL)
    {
      if (s == this)
	return 1;

      s = s->getParent ();
    }

  return 0;
}


/**
 * _Slab::getPrevLeaf:get leaf before this Slab
 * 
 * Moves back in the tree and then drops to a leaf
 * 
 * Return Value: Previous Leaf Slab or NULL if none are found.
 **/
Slab _Slab::getPrevLeaf ()
{
  Slab s = this;
  Slab prev = NULL;

  do
    {
      if (s == NULL)
	break;
      prev = s->getPrevNode ();
      s = s->getParent ();
    }  while (prev == NULL);

  if (prev == NULL)
    return NULL;

  while (prev->classID () == ContSlabID &&
	 prev->getNumberOfChildren () > 0)
    {
      prev = (Slab) prev->getChildAt (prev->getNumberOfChildren () - 1);
    }

  return prev;
}


/**
 * _Slab::getNextLeaf:get left after this Slab
 * 
 * Moves forward in the tree and then drops to a leaf
 * 
 * Return Value: Next Leaf Slab or NULL if none are found
 **/
Slab _Slab::getNextLeaf ()
{
  Slab s = this;
  Slab next = NULL;

  do
    {
      if (s == NULL)
	break;
      next = s->getNextNode ();
      s = s->getParent ();
    }  while (next == NULL);

  if (next == NULL)
    return NULL;

  while (next->classID () == ContSlabID &&
	 next->getNumberOfChildren () > 0)
    {
      next = (Slab) next->getChildAt (0);
    }

  return next;
}


/**
 * _Slab::slab_overlaps_view:is any part of slab visible on view?
 * @x: Slab's X position from corner of document
 * @y: Slab's Y position from corner of document
 * @g: the view being tested
 * 
 * Tests the bounds of Slab (after scaling) against the bounds
 * of @g.
 * 
 * Return Value: 1 if Slab is visable, 0 otherwise
 **/
int _Slab::slab_overlaps_view (int x, int y, WPView *g)
{
  int vx = wp_view_get_x (g);
  int vy = wp_view_get_y (g);
  int vw = wp_view_get_width (g);
  int vh = wp_view_get_height (g);

  if (x + right < vx ||
      x - left > vx+vw ||
      y + down < vy ||
      y - up > vy+vh)
    return 0; /* out of bounds */

  return 1;
}


boolean _Slab::trim_slab (GList *spots, GList *views, int debug)
{
  /* if this container is now empty, remove 'this' */
  /*if (g_list_length (children) == 0 && parent != NULL)*/
  if (getNumberOfChildren () == 0 && parent != NULL)
    {
      if (debug >= 0)
	printf ("%sContSlab->deleteChild -- cleaning up a node...",
		tabber (debug));

      parent->deleteChild (parent->getIndexOfChild ((Object) this),
			   spots, views, (debug >= 0)?debug+1:-1, false);
      return true;
    }

  return false;
}
