/*
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include "wp_spot.h"
#include "wp_slab.h"

_Spot::_Spot ()
{
  slab = NULL;
  position = 0;
}


_Spot::_Spot (Slab slab, int position)
{
  this->slab = slab;
  this->position = position;
}


void _Spot::rise ()
{
  Slab parent = slab->getParent ();
  if (parent == NULL) return;

  int myPos = parent->getIndexOfChild (slab);

  slab = parent;
  position = myPos;
  /*return new _Spot (parent, myPos);*/
}


/* return a spot that is a level deeper */

void _Spot::fall ()
{
  if (classID () != ContSlabID)
    return;

  if (position == slab->getNumberOfChildren ())
    {
      Slab child = (Slab) slab->getChildAt (slab->getNumberOfChildren ());
      /*return new _Spot (child, child->getNumberOfChildren ());*/
      slab = child;
      position = child->getNumberOfChildren ();
    }
  else
    {
      Slab child = (Slab) slab->getChildAt (0);
      /*return new _Spot (child, 0);*/
      slab = child;
      position = 0;
    }
}


/* move this spot as deep as it will go */

void _Spot::sink ()
{
  while (slab->classID () == ContSlabID &&
	 slab->getNumberOfChildren () > 0)
    {
      if (position == slab->getNumberOfChildren ())
	{
	  slab = (Slab) slab->getChildAt (slab->getNumberOfChildren ()-1);
	  position = slab->getNumberOfChildren ();
	}
      else
	{
	  slab = (Slab) slab->getChildAt (0);
	  position = 0;
	}
    }
}


void _Spot::hopBack ()
{
  Slab prev;

  for (;;)
    {
      if (position > 0) return;

      prev = slab->getPrevNode ();
      if (prev == NULL) return;

      slab = prev;
      position = slab->getNumberOfChildren ();
    }
}


void _Spot::hopForward ()
{
  Slab next;

  for (;;)
    {
      if (position < slab->getNumberOfChildren ()) return;

      next = slab->getNextNode ();
      if (next == NULL) return;

      slab = next;
      position = 0;
    }
}



void _Spot::hopBackLeaf ()
{
  Slab prev;

  for (;;)
    {
      if (position > 0) return;

      prev = slab->getPrevLeaf ();
      if (prev == NULL) return;

      slab = prev;
      position = slab->getNumberOfChildren ();
    }
}


void _Spot::hopForwardLeaf ()
{
  Slab next;

  for (;;)
    {
      if (position < slab->getNumberOfChildren ()) return;

      next = slab->getNextLeaf ();
      if (next == NULL) return;

      slab = next;
      position = 0;
    }
}



Slab _Spot::getSlab ()
{
  return slab;
}


void _Spot::setSlab (Slab s)
{
  slab = s;
}


int _Spot::getPosition ()
{
  return position;
}


void _Spot::setPosition (int p)
{
  position = p;
}

