/*
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include "wp_st_slab.h"
#include "wp_utils.h"
#include "wp_prefs.h"


_STSlab::_STSlab (WPStyle *set_style) : _Slab (set_style)
{
  this->style = set_style;
  calculateSize (-1);
}


Slab _STSlab::newOfThisType ()
{
  STSlab s = new _STSlab (style);
  s->setPluginData (getPluginData ());
  /*
  s->setXSizePolicy (xSizePolicy);
  s->setYSizePolicy (ySizePolicy);
  */
  return s;
}


void _STSlab::calculateSize (int debug)
{
  GdkFont *f = style->getFont (1); /* 1 so we get points */

  if (debug >= 0)
    {
      printf ("%sSTSlab->calculateSize this=", tabber (debug));
      dd ();
      printf ("(l,u,r,d):%d %d %d %d\n", left, up, right, down);
    }

  /*left = 0;*/
  /*right = fm->stringWidth (" ");*/
  /*up = fm->getAscent ();*/
  /*down = fm->getDescent ();*/

  if (f)
    {
      left = 0;
      right = gdk_string_width (f, " ");
      up = f->ascent;
      down = f->descent;
    }
  else
    {
      left = 0;
      right = 0;
      up = 0;
      down = 0;
    }
}


void _STSlab::draw (int x, int y, GList *views, int /*debug*/)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;

      /*System->out->println ("STSlab drawing at " + x + ", " + y);*/

      GdkColor *bgc = style->getBackground ();

      /* draw background color (if different than parents */
      if (parent &&
	  colorcmp (parent->getStyle ()->getBackground (), bgc) == 0)
	{
	  wp_view_set_color (g, bgc);
	  wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
	}

      if (wp_view_hl_is_on (g) &&
	  wp_slab_is_between (wp_view_get_point (g)->getSlab (),
			      this,
			      wp_view_get_mark (g)->getSlab ()))
	{
	  Spot point = wp_view_get_point (g);
	  Spot mark = wp_view_get_mark (g);

	  if (! wp_spot_order (point, mark))
	    {
	      Spot t = point; point = mark; mark = t;
	    }

	  Slab ps = point->getSlab ();
	  Slab ms = mark->getSlab ();

	  wp_view_set_color (g, wp_view_get_cursor_color (g));

	  if (this != ps && this != ms)
	    wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
	  else if (this == ps && this == ms)
	    {
	      if (point->getPosition () != mark->getPosition ())
		wp_view_fill_rect (g, x-left, y-up,
				   left+right, up+down);
	    }
	  else if (this == ps)
	    {
	      if (point->getPosition () == 0)
		wp_view_fill_rect (g, x-left, y-up,
				   left+right, up+down);

	    }
	  else if (this == ms)
	    {
	      if (mark->getPosition () == 1)
		wp_view_fill_rect (g, x-left, y-up,
				   left+right, up+down);
	      
	    }
	  return;
	}

      if (! preferences->draw_fast)
	return;

      wp_view_set_color (g, style->getBackground ());
      wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
    }
}


void _STSlab::erase (int x, int y, GList *views)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;
      wp_view_set_color (g, style->getBackground ());
      wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
    }
}


boolean _STSlab::insertChild (int /*i*/,
			      Object /*what*/,
			      GList */*spots*/,
			      GList */*views*/,
			      int /*debug*/)
{
  printf ("error -- STSlab->insertChild called\n");
  return false;
}


Object _STSlab::deleteChild (int i,
			     GList *spots,
			     GList *views,
			     int debug,
			     boolean ignore_lost_spots)
{
  if (debug >= 0)
    {
      printf ("%sdeleteChild this=", tabber (debug));
      dd ();
      printf ("i=%d\n", i);
    }

  if (i != 0)
    {
      printf ("error -- STSlab->deleteChild, i=%d", i);
      return NULL;
    }

  return parent->deleteChild (parent->getIndexOfChild (this),
			      spots, views, debug, ignore_lost_spots);
}


boolean _STSlab::childSizeChange (Slab /*child*/, GList */*spots*/,
				  int /*dLeft*/, int /*dUp*/,
				  int /*dRight*/, int /*dDown*/,
				  GList */*views*/,
				  int /*debug*/)
{
  printf ("error -- STSlab::childSizeChange called\n");
  return false;
}


boolean _STSlab::ensureBorder (int i, GList */*spots*/,
			       GList */*views*/, int /*debug*/)
{
  if (i<0 || i>1)
    {
      printf ("error  -- STSlab->ensureBorder, i=%d\n", i);
      return false;
    }

  return true;
}


int _STSlab::getIndexOfChild (Object /*what*/)
{
  printf ("error -- STSlab->getIndexOfChild called\n");
  return -1;
}


Slab _STSlab::getChildAt (int /*i*/)
{
  printf ("error -- STSlab->getIndexOfChild called\n");
  return NULL;
}


void _STSlab::insertChildAt (Object /*what*/, int /*i*/)
{
  printf ("error -- STSlab->insertChildAt called\n");
}


void _STSlab::dumpToScreen (int tab)
{
  for (int t=0; t<tab; t++) printf ("+-");
  printf ("Space(%d,%d,%d,%d) pp:%p <this=%p><parent=%p>\n",
	  left, up, right, down, plugin_data, this, parent);
}


void _STSlab::dd ()
{
  printf ("[<sp>]");
}


int _STSlab::isLast ()
{
  return 0;
}


int _STSlab::isFirst ()
{
  return 0;
}


Slab _STSlab::firstLowestContainer () {return this;}
Slab _STSlab::lastLowestContainer () {return this;}


Slab _STSlab::getNextNode ()
{
  if (parent == NULL)
    return NULL;
  return parent->getNextNodeUp (this, 1);
}
Slab _STSlab::getNextNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _STSlab::getNextNodeDown (int depth)
{
  if (depth == 0)
    return this;
  return NULL;
}


Slab _STSlab::getPrevNode ()
{
  return parent->getPrevNodeUp (this, 1);
}
Slab _STSlab::getPrevNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _STSlab::getPrevNodeDown (int /*depth*/)
{
  return this;
}


int _STSlab::getNumberOfChildren ()
{
  return 1;
}


WpPoint _STSlab::getPointOfChild (Slab child)
{
  return getPointOfChild (getIndexOfChild (child));
}


WpPoint _STSlab::getPointOfChild (int childNum)
{
  int x=0, y=0;

  if (parent != NULL)
    {
      WpPoint my_point = parent->getPointOfChild (this);
      x = my_point.x;
      y = my_point.y;
    }

  if (childNum == 0)
    return newWpPoint (x, y);

  if (childNum == 1)
    return newWpPoint (x+right, y);

  printf ("error -- STSlab->getPointOfChild: childNum != 0\n");

  return newWpPoint (0, 0);
}


Spot _STSlab::getSpotOfPoint (int px, int /*py*/)
{
  if (px < left + right)
    return new _Spot (this, 0);

  return new _Spot (this, 1);
}

