/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "wp_style.h"
#include <string.h>
/*#include "mx_font.h"*/
#include "wp_font.h"


void *double_to_key (double d)
{
  return (void *)((int)(d*100.0));
}


WPStyle::WPStyle (char *style_name)
{
  /*f = NULL;*/
  /*
  for (int i=0; i<10; i++)
    f[ i ] = NULL;
  */

  scaled_fonts = g_hash_table_new (g_direct_hash, g_direct_equal);

  font_name = NULL;
  fg = NULL;
  bg = NULL;
  name = strdup (style_name);
  alignment = ALIGN_LOW;
  direction = DIR_RIGHT;
  xsp = SIZE_POLICY_BY_CHILDREN;
  ysp = SIZE_POLICY_BY_CHILDREN;

  fixed_left = 0;
  fixed_up = 0;
  fixed_right = 0;
  fixed_down = 0;

  margin_left = 0;
  margin_up = 0;
  margin_right = 0;
  margin_down = 0;

  margin_rendering = MARGIN_RENDERING_POLICY_BLANK;
  when_empty = WHEN_EMPTY_REMOVE;

  max_children = 0;

  underline = 0;
}

WPStyle::~WPStyle()
{
  free(name);
}


void WPStyle::copy (WPStyle *old)
{
  if (!old) return;

  /*f = old->f;*/
  scaled_fonts = old->scaled_fonts; /* XXX ? */

  if (font_name)
   free(font_name);
  font_name = old->font_name ? strdup(old->font_name) : (char *) NULL;
  fg = old->fg;
  bg = old->bg;
  alignment = old->alignment;

  fixed_left = old->fixed_left;
  fixed_up = old->fixed_up;
  fixed_right = old->fixed_right;
  fixed_down = old->fixed_down;

  margin_left = old->margin_left;
  margin_up = old->margin_up;
  margin_right = old->margin_right;
  margin_down = old->margin_down;

  max_children = old->max_children;
  
  direction = old->direction;
  xsp = old->xsp;
  ysp = old->ysp;

  margin_rendering = MARGIN_RENDERING_POLICY_BLANK;
  when_empty = WHEN_EMPTY_REMOVE;
  underline = old->underline;
}


GdkFont *WPStyle::getFont (double scale)
{
  if (font_name == NULL || strlen (font_name) == 0)
    return NULL;

  if (scale > 10.0) scale = 10.0;
  if (scale < 0.0) scale = 1.0;

  GdkFont *f =
    (GdkFont *) g_hash_table_lookup (scaled_fonts, double_to_key (scale));

  if (! f)
    f = load_font (scale);

  return f;
}

char *WPStyle::getFontName ()
{
  if (font_name == NULL)
    return "";
  return font_name;
}


GdkFont *WPStyle::load_font (double scale)
{
  if (font_name == NULL || strcmp (font_name, "") == 0)
    return NULL;

  char *foundry = getFontFoundry (font_name);
  char *family = getFontFamily (font_name);
  char *foundry_family =
    (char *) malloc (strlen (foundry) + strlen (family) + 5);
  float size = getFontPointSize (font_name) / 10.0;
  if (size < 0.0005)
    size = getFontPixelSize (font_name);

  sprintf (foundry_family, "%s-%s", foundry, family);

  free (foundry);
  free (family);
  /* XXX free foundry_family? */

  if (scale > 10.0) scale = 10.0;
  if (scale < 0.0) scale = 1.0;

  char *adjusted_font_name = /* XXX why 1.1?  better PS output? */
    setFontPointSize (font_name, (int) (size * 10.0 * 1.1 * scale));

  printf ("Loading Font '%s'\n", adjusted_font_name);
  GdkFont *f = gdk_font_load (adjusted_font_name);
  free (adjusted_font_name);

  if (f == NULL)
    f = gdk_font_load ("fixed"); /* XXX */
  if (f == NULL)
    {
      fprintf (stderr, "Couldn't even load 'fixed' font!\n");
      return NULL;
    }

  g_hash_table_insert (scaled_fonts, double_to_key (scale), f);

  return f;
}

void WPStyle::setFontName (char *set_font_name)
{
  /*char *old_font_name = font_name;*/

  if (set_font_name == NULL)
    return;

  font_name = strdup (set_font_name);

# if 0
  if (load_font () < 0)
    {
      if (set_font_name != NULL && strlen (set_font_name) > 0)
	fprintf (stderr, "font load failed for '%s'\n", set_font_name);
      free (font_name);
      font_name = old_font_name; 
      if (load_font () < 0)
	{
	  font_name =
	    strdup ("-*-helvetica-medium-r-normal-*-*-100-*-*-*-*-*-*");
	  if (load_font () < 0)
	    {
	      printf ("can't get default font loaded\n");
	      free (font_name);
	      font_name = strdup ("");
	      f = NULL;
	    }
	}
    }
  else
    free (old_font_name);
# endif /* 0 */
}


void WPStyle::setForeground(GdkColor *set_fg)
{
  /* XXX free the old one? */
  fg = set_fg;
}

void WPStyle::setBackground(GdkColor *set_bg)
{
  bg = set_bg;
}

GdkColor* WPStyle::getForeground()
{
  return fg;
}

GdkColor* WPStyle::getBackground()
{
  return bg;
}

void WPStyle::setAlignment (WPAlignment set_a)
{
  alignment = set_a;
}


WPAlignment WPStyle::getAlignment ()
{
  return alignment;
}


void WPStyle::setFixedLeft (unsigned long int v)
{
  fixed_left = v;
}

void WPStyle::setFixedUp (unsigned long int v)
{
  fixed_up = v;
}

void WPStyle::setFixedRight (unsigned long int v)
{
  fixed_right = v;
}

void WPStyle::setFixedDown (unsigned long int v)
{
  fixed_down = v;
}


unsigned long int WPStyle::getFixedLeft ()
{
  return fixed_left;
}

unsigned long int WPStyle::getFixedUp ()
{
  return fixed_up;
}

unsigned long int WPStyle::getFixedRight ()
{
  return fixed_right;
}

unsigned long int WPStyle::getFixedDown ()
{
  return fixed_down;
}


void WPStyle::setMarginLeft (unsigned long int v)
{
  margin_left = v;
}

void WPStyle::setMarginUp (unsigned long int v)
{
  margin_up = v;
}

void WPStyle::setMarginRight (unsigned long int v)
{
  margin_right = v;
}

void WPStyle::setMarginDown (unsigned long int v)
{
  margin_down = v;
}
unsigned long int WPStyle::getMarginLeft ()
{
  return margin_left;
}

unsigned long int WPStyle::getMarginUp ()
{
  return margin_up;
}

unsigned long int WPStyle::getMarginRight ()
{
  return margin_right;
}

unsigned long int WPStyle::getMarginDown ()
{
  return margin_down;
}


void WPStyle::setMarginRendering (WPMarginRenderingPolicy set_mrp)
{
  margin_rendering = set_mrp;
}


WPMarginRenderingPolicy WPStyle::getMarginRendering ()
{
  return margin_rendering;
}


void WPStyle::setWhenEmpty (WPWhenEmptyPolicy set_we)
{
  when_empty = set_we;
}


WPWhenEmptyPolicy WPStyle::getWhenEmpty ()
{
  return when_empty;
}


/* max subparts a slab can have.  0 = unlimited */
void WPStyle::setMaxChildren (unsigned long int m)
{
  max_children = m;
}

unsigned long int WPStyle::getMaxChildren ()
{
  return max_children;
}




void WPStyle::setUnderline (int u)
{
  underline = u;
}


int WPStyle::getUnderline ()
{
  return underline;
}


char *style_alignment_to_string (WPAlignment a)
{
  switch (a)
    {
    case ALIGN_LOW: return "low";
    case ALIGN_HIGH: return "high";
    case ALIGN_CENTER: return "center";
    default: return "unknown";
    }
}
WPAlignment style_string_to_alignment (unsigned const char *s)
{
  if (!s) return ALIGN_LOW;
  if (strcasecmp ((char *) s, "low") == 0) return ALIGN_LOW;
  if (strcasecmp ((char *) s, "high") == 0) return ALIGN_HIGH;
  if (strcasecmp ((char *) s, "center") == 0) return ALIGN_CENTER;
  return ALIGN_LOW;
}


char *style_direction_to_string (WPDirection a)
{
  switch (a)
    {
    case DIR_LEFT: return "left";
    case DIR_UP: return "up";
    case DIR_RIGHT: return "right";
    case DIR_DOWN: return "down";
    default: return "unknown";
    }
}
WPDirection style_string_to_direction (unsigned const char *s)
{
  if (!s) return DIR_RIGHT;
  if (strcasecmp ((char *) s, "left") == 0) return DIR_LEFT;
  if (strcasecmp ((char *) s, "up") == 0) return DIR_UP;
  if (strcasecmp ((char *) s, "right") == 0) return DIR_RIGHT;
  if (strcasecmp ((char *) s, "down") == 0) return DIR_DOWN;
  return DIR_RIGHT;
}


char *style_size_policy_to_string (WPSizePolicy a)
{
  switch (a)
    {
    case SIZE_POLICY_FIXED: return "fixed";
    case SIZE_POLICY_BY_CHILDREN: return "children";
    case SIZE_POLICY_BY_PARENT: return "parent";
    default: return "unknown";
    }
}
WPSizePolicy style_string_to_size_policy (unsigned const char *s)
{
  if (!s) return SIZE_POLICY_BY_CHILDREN;
  if (strcasecmp ((char *) s, "fixed") == 0) return SIZE_POLICY_FIXED;
  if (strcasecmp ((char *) s, "children") == 0) return SIZE_POLICY_BY_CHILDREN;
  if (strcasecmp ((char *) s, "parent") == 0) return SIZE_POLICY_BY_PARENT;
  return SIZE_POLICY_BY_CHILDREN;
}


char *style_margin_rendering_policy_to_string (WPMarginRenderingPolicy a)
{
  switch (a)
    {
    case MARGIN_RENDERING_POLICY_BLANK: return "blank";
    case MARGIN_RENDERING_POLICY_FILL: return "fill";
    case MARGIN_RENDERING_POLICY_OUTLINE: return "outline";
    default: return "unknown";
    }
}
WPMarginRenderingPolicy
style_string_to_margin_rendering_policy (unsigned const char *s)
{
  if (!s) return MARGIN_RENDERING_POLICY_BLANK;
  if (strcasecmp ((char *) s, "blank") == 0)
    return MARGIN_RENDERING_POLICY_BLANK;
  if (strcasecmp ((char *) s, "3d") == 0)
    return MARGIN_RENDERING_POLICY_FILL;
  if (strcasecmp ((char *) s, "fill") == 0)
    return MARGIN_RENDERING_POLICY_FILL;
  if (strcasecmp ((char *) s, "outline") == 0)
    return MARGIN_RENDERING_POLICY_OUTLINE;
  return MARGIN_RENDERING_POLICY_BLANK;
}


char *style_when_empty_policy_to_string (WPWhenEmptyPolicy a)
{
  switch (a)
    {
    case WHEN_EMPTY_REMOVE: return "remove";
    case WHEN_EMPTY_KEEP: return "keep";
    default: return "unknown";
    }
}
WPWhenEmptyPolicy style_string_to_when_empty_policy (unsigned const char *s)
{
  if (!s) return WHEN_EMPTY_REMOVE;
  if (strcasecmp ((char *) s, "remove") == 0) return WHEN_EMPTY_REMOVE;
  if (strcasecmp ((char *) s, "keep") == 0) return WHEN_EMPTY_KEEP;
  return WHEN_EMPTY_REMOVE;
}
