/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "wp_style_manager.h"
#include "wp_style.h"
#include "wp_font.h"


WPStyleManager::WPStyleManager ()
{
  srand (time (NULL));
  style_groups = NULL;
}


/*
char *WPStyleManager::getSelectedStyleGroupName ()
{
  GtkNotebook *nb = GTK_NOTEBOOK (notebook);
  GtkNotebookPage *cur_page = nb->cur_page;
  GtkWidget *tab_label = cur_page->tab_label;
  GtkLabel *lbl = GTK_LABEL (tab_label);
  return lbl->label;
}
*/



style_group *WPStyleManager::newStyleGroup (char *name)
{
  GList *i;

  for (i = style_groups; i != NULL; i=i->next)
    {
      style_group *sg = (style_group *) i->data;
      if (strcmp (sg->name, name) == 0)
	/* already have one by that name */
	return sg;
    }

  style_group *new_group;

  new_group = (style_group *) malloc (sizeof (style_group));
  new_group->name = (char *) malloc (strlen (name) + 1);
  strcpy (new_group->name, name);
  new_group->styles = NULL;
  /*new_group->current = NULL;*/
  /*new_group->default_style = NULL;*/

  style_groups = g_list_append (style_groups, new_group);

  return new_group;
}


style_group *WPStyleManager::getStyleGroupByName (char *name)
{
  GList *i;

  for (i = style_groups; i != NULL; i=i->next)
    {
      style_group *sg = (style_group *) i->data;
      if (strcmp (sg->name, name) == 0)
	return sg;
    }

  return newStyleGroup (name);
}


WPStyle *WPStyleManager::newStyle (char *group_name, char *style_name)
{
  return newStyle (getStyleGroupByName (group_name), style_name);
}


WPStyle *WPStyleManager::newStyle (style_group *sg, char *style_name)
{
  WPStyle *s;

  if ((s = getStyleByName (sg, style_name)) != NULL)
    return s; /* already have one by that name. */

  s = new WPStyle (style_name);

  sg->styles = g_list_append (sg->styles, s);

  /*
  if (sg->current == NULL)
    sg->current = sg->styles;
  */

  return s;
}


WPStyle *WPStyleManager::copyStyle (char *group_name,
				    char *style_name,
				    WPStyle *style)
{
  newStyleGroup(group_name);
  WPStyle *s=newStyle(group_name,style_name);
  s->copy(style);

  return s;
}


WPStyle *WPStyleManager::getStyleByName (char *group_name, char *style_name)
{
  return getStyleByName (getStyleGroupByName (group_name), style_name);
}


WPStyle *WPStyleManager::getStyleByName (style_group *sg, char *style_name)
{
  GList *i;

  if (sg == NULL)
    return NULL;

  for (i = sg->styles; i != NULL; i=i->next)
    {
      WPStyle *s = (WPStyle *) i->data;
      if (strcmp (s->getName (), style_name) == 0)
	return s;
    }

  return NULL;
}


/* free what this returns */
char *WPStyleManager::unique_style_name (style_group *sg)
{
  int i;
  char new_name[ 20 ];

  for (i=0; i<19; i++)
    {
      char c = 'a' + (rand () % ('z' - 'a'));
      new_name[ i ] = c;
    }

  new_name[ i ] = '\0';

  if (getStyleByName (sg, new_name) != NULL)
    /* try again */
    return unique_style_name (sg);

  return strdup (new_name);
}


/* compare colors */
int colorcmp (GdkColor *a, GdkColor *b)
{
  if (!a && !b) return 1;
  if (!a) return 0;
  if (!b) return 0;

  if (a->red == b->red &&
      a->green == b->green &&
      a->blue == b->blue)
    return 1;
  return 0;
}


/* scan the styles in the named group searching for one that
   matches the given spec.  if found return it.  if one isn't
   found, create a new one and return that. */

WPStyle *
WPStyleManager::
findOrCreateStyle (style_group *sg,
		   char *font_name, /* font name */
		   GdkColor *fg, /* foreground */
		   GdkColor *bg, /* background */
		   WPAlignment algn, /* alignment */
		   WPDirection dir, /* direction */
		   WPSizePolicy xsp, /* x size policy */
		   WPSizePolicy ysp, /* x size policy */
		   unsigned long int fl, /* fixed left */
		   unsigned long int fu, /* fixed up */
		   unsigned long int fr, /* fixed right */
		   unsigned long int fd, /* fixed down */
		   unsigned long int fml, /* fixed margin left */
		   unsigned long int fmu, /* fixed margin up */
		   unsigned long int fmr, /* fixed margin right */
		   unsigned long int fmd, /* fixed margin down */
		   unsigned long int mc, /* max children */
		   WPMarginRenderingPolicy mrp, /* margin rendering */
		   WPWhenEmptyPolicy we,
		   int underline) /* delete when empty? */
{
  GList *i;

  if (sg == NULL)
    return NULL;

  for (i = sg->styles; i != NULL; i=i->next)
    {
      WPStyle *s = (WPStyle *) i->data;

      if (strcmp (font_name, s->getFontName ()) == 0 &&
	  colorcmp (fg, s->getForeground ()) &&
	  colorcmp (bg, s->getBackground ()) &&
	  algn == s->getAlignment () &&
	  dir == s->getDirection () &&
	  xsp == s->getXSizePolicy () &&
	  ysp == s->getYSizePolicy () &&
	  fl == s->getFixedLeft () &&
	  fu == s->getFixedUp () &&
	  fr == s->getFixedRight () &&
	  fd == s->getFixedDown () &&
	  fml == s->getMarginLeft () &&
	  fmu == s->getMarginUp () &&
	  fmr == s->getMarginRight () &&
	  fmd == s->getMarginDown () &&
	  mc == s->getMaxChildren () &&
	  mrp == s->getMarginRendering () &&
	  we == s->getWhenEmpty () &&
	  underline == s->getUnderline ())
	return s;
    }

  return createStyle (sg, font_name, fg, bg, algn, dir, xsp, ysp,
		      fl, fu, fr, fd, fml, fmu, fmr, fmd, mc, mrp, we,
		      underline);
}


WPStyle *
WPStyleManager::
createStyle (style_group *sg,
	     char *font_name, /* font name */
	     GdkColor *fg, /* foreground */
	     GdkColor *bg, /* background */
	     WPAlignment algn, /* alignment */
	     WPDirection dir, /* direction */
	     WPSizePolicy xsp, /* x size policy */
	     WPSizePolicy ysp, /* x size policy */
	     unsigned long int fl, /* fixed left */
	     unsigned long int fu, /* fixed up */
	     unsigned long int fr, /* fixed right */
	     unsigned long int fd, /* fixed down */
	     unsigned long int fml, /* fixed margin left */
	     unsigned long int fmu, /* fixed margin up */
	     unsigned long int fmr, /* fixed margin right */
	     unsigned long int fmd, /* fixed margin down */
	     unsigned long int mc, /* max children */
	     WPMarginRenderingPolicy mrp, /* margin rendering */
	     WPWhenEmptyPolicy we, /* delete when empty? */
	     int underline) 
{
  if (sg == NULL)
    return NULL;

  char *name_if_new; /* what to name it if we create it */
  name_if_new = unique_style_name (sg);
  WPStyle *n = newStyle (sg, name_if_new);
  n->setFontName (font_name);
  n->setForeground (fg);
  n->setBackground (bg);
  n->setAlignment (algn);
  n->setDirection (dir);
  n->setXSizePolicy (xsp);
  n->setYSizePolicy (ysp);
  n->setFixedLeft (fl);
  n->setFixedUp (fu);
  n->setFixedRight (fr);
  n->setFixedDown (fd);
  n->setMarginLeft (fml);
  n->setMarginUp (fmu);
  n->setMarginRight (fmr);
  n->setMarginDown (fmd);
  n->setMaxChildren (mc);
  n->setMarginRendering (mrp);
  n->setWhenEmpty (we);
  n->setUnderline (underline);

  free (name_if_new);

  return n;
}
