
#include "config.h"
#include <gnome.h>
#include <libgnomeui/libgnomeui.h>

#include "wp_commands.h"
#include "wp_file_ops.h"
#include "wp_about.h"
#include "pixmaps/foot.xpm"
#include "wp_prefs.h"


static GnomeUIInfo foot_menu_info[] = {

  { GNOME_APP_UI_ITEM, N_("About..."), NULL,
    about_box_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_ABOUT },

  { GNOME_APP_UI_ITEM, "Preferences...", NULL,
    preferences_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PROP },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ITEM, "Quit", NULL,
    file_quit_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_QUIT },

  { GNOME_APP_UI_ENDOFINFO }
};


static GnomeUIInfo file_menu_info[] = {
  { GNOME_APP_UI_ITEM, "New View", NULL,
    file_new_view_cmd_callback, NULL, NULL },
  
  { GNOME_APP_UI_ITEM, "New Document", NULL,
    file_new_doc_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_NEW },

  { GNOME_APP_UI_ITEM, "Open...", NULL,
    file_open_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_OPEN },

  { GNOME_APP_UI_ITEM, "Insert...", NULL,
    file_insert_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_OPEN },
  
  { GNOME_APP_UI_SEPARATOR },
  
  { GNOME_APP_UI_ITEM, "Save", NULL,
    file_saveas_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE },

  { GNOME_APP_UI_ITEM, "Save As", NULL,
    file_saveas_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE_AS },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ITEM, "Print Preview...", NULL,
    file_printpreview_cmd_callback, NULL, NULL },

  { GNOME_APP_UI_ITEM, "Print", NULL,
    file_print_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PRINT },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ITEM, "Close", NULL,
    file_close_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_CLOSE },

  { GNOME_APP_UI_ENDOFINFO }
};

static GnomeUIInfo edit_menu_info[] = {
  /*
  { GNOME_APP_UI_ITEM, "Undo", NULL,
    edit_undo_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_UNDO  },

  { GNOME_APP_UI_ITEM, "Redo", NULL,
    edit_redo_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_REDO  },
  */

  { GNOME_APP_UI_ITEM, "Cut", NULL,
    edit_cut_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_CUT },

  { GNOME_APP_UI_ITEM, "Copy", NULL,
    edit_copy_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_COPY },

  { GNOME_APP_UI_ITEM, "Paste", NULL,
    edit_paste_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PASTE },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ITEM, "Clear", NULL,
    edit_clear_cmd_callback, NULL, NULL },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ENDOFINFO }
};

static GnomeUIInfo view_menu_info[] = {
# if BROKEN
  { GNOME_APP_UI_ITEM, "Show Rulers", NULL,
    view_showrulers_cmd_callback },
# endif /* BROKEN */
  /*  { GNOME_APP_UI_ITEM, "Split Vertically", NULL,
      view_vsplit_cmd_callback },*/
# if BROKEN
  { GNOME_APP_UI_ITEM, "Split Horizontally", NULL,
    view_hsplit_cmd_callback },
# endif /* BROKEN */
  /*  { GNOME_APP_UI_ITEM, "Unsplit", NULL,
      view_unsplit_cmd_callback },*/
  { GNOME_APP_UI_SEPARATOR },
  { GNOME_APP_UI_ITEM, "Refresh", NULL,
    view_refresh_cmd_callback },
  { GNOME_APP_UI_SEPARATOR },
  { GNOME_APP_UI_ENDOFINFO }
};

static GnomeUIInfo tools_menu_info[] = {
  /*{ GNOME_APP_UI_ITEM, "Refresh Plugin List" },*/
  { GNOME_APP_UI_SEPARATOR },
  { GNOME_APP_UI_ENDOFINFO }
};

static GnomeUIInfo plugins_menu_info[] = {
  /*{ GNOME_APP_UI_ITEM, "Refresh Plugin List" },*/
  { GNOME_APP_UI_SEPARATOR },
  { GNOME_APP_UI_ENDOFINFO }
};

/*
static GnomeUIInfo insert_menu_info[] = {
  { GNOME_APP_UI_ENDOFINFO }
};
*/

static GnomeUIInfo documents_menu_info[] = {
  { GNOME_APP_UI_ENDOFINFO }
};

static GnomeUIInfo help_menu_info[] = {
  /*  { GNOME_APP_UI_HELP, "HelpStuff", NULL, "Gwp" },*/
  { GNOME_APP_UI_ENDOFINFO }
};

#ifdef DEBUG
static GnomeUIInfo debug_menu_info[] = {
  { GNOME_APP_UI_ITEM, "Dump Document Structure", NULL,
    dump_doc_structure_callback, NULL, NULL },

  { GNOME_APP_UI_ENDOFINFO }
};
#endif

/* main menu bar */
GnomeUIInfo menubar_info[] = {
  { GNOME_APP_UI_SUBTREE, " ", NULL,
    foot_menu_info, NULL, NULL,
    GNOME_APP_PIXMAP_DATA, foot_xpm, 0, (GdkModifierType) 0, NULL },
  { GNOME_APP_UI_SUBTREE, "File", NULL, file_menu_info },
  { GNOME_APP_UI_SUBTREE, "Edit", NULL, edit_menu_info },
  { GNOME_APP_UI_SUBTREE, "View", NULL, view_menu_info },
  { GNOME_APP_UI_SUBTREE, "Tools", NULL, tools_menu_info },
  /*{ GNOME_APP_UI_SUBTREE, "Insert", NULL, insert_menu_info },*/
  { GNOME_APP_UI_SUBTREE, "Plugins", NULL, plugins_menu_info },
  { GNOME_APP_UI_SUBTREE, "Documents", NULL, documents_menu_info },
  { GNOME_APP_UI_SUBTREE, "Help", NULL, help_menu_info },
# ifdef DEBUG
  { GNOME_APP_UI_SUBTREE, "Debug", NULL, debug_menu_info },
# endif
  { GNOME_APP_UI_ENDOFINFO }
};

GnomeUIInfo toolbar_info[] = {
  { GNOME_APP_UI_ITEM,
    "New", "Create New Document",
    file_new_doc_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_NEW },

  { GNOME_APP_UI_ITEM,
    "Open", "Open Existing Document",
    file_open_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_OPEN },

  { GNOME_APP_UI_ITEM,
    "Save", "Save Current Document",
    file_saveas_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_SAVE },

  { GNOME_APP_UI_SEPARATOR },

  { GNOME_APP_UI_ITEM,
    "Cut", "Cut selected text",
    edit_cut_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_CUT },

  { GNOME_APP_UI_ITEM,
    "Copy", "Copy selected text",
    edit_copy_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_COPY },

  { GNOME_APP_UI_ITEM,
    "Paste", "Paste text",
    edit_paste_cmd_callback, NULL, NULL,
    GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_PIXMAP_PASTE },

  { GNOME_APP_UI_ENDOFINFO }
};
