/* BLAH BLAH!  BLAHBLAHBLAH!
 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include <gnome.h>
#include "wp_view.h"
#include "wp_commands.h"
#include "wp_clipboard.h"
#include "wp_plugin_manager.h"
#include "wp_plugin.h"
#include "wp_prefs.h"
#include "wp.h"
#include "wp_mdi_document.h"

/* Go plugin stuff */
#include "plugin.h"
#include "gwp_plugin_api.h"


static inline int doc_coords_to_view (WPView *view, long int v);
static inline int doc_xcoord_to_view (WPView *view, long int v);
static inline int doc_ycoord_to_view (WPView *view, int v);
static inline int view_coords_to_doc (WPView *view, int v);
static inline int view_xcoord_to_doc (WPView *view, int v);
static inline int view_ycoord_to_doc (WPView *view, int v);


static gint wp_view_scroll_event (GtkWidget *widget,
				  GdkEvent *event, void *client_data);
static gint wp_view_drawing_area_event (GtkWidget *widget,
					GdkEvent *event, void *client_data);


void view_selection_get (GtkWidget * widget,
			 GtkSelectionData *selection_data,
			 guint info,
			 guint time,
			 gpointer data);
void view_selection_received (GtkWidget *widget,
			      GtkSelectionData *selection_data,
			      guint time,
			      gpointer data);


static void wp_view_class_init     (WPViewClass     *klass);
static void wp_view_init           (WPView          *view);
static void wp_view_map            (GtkWidget        *widget);
static void wp_view_paint          (GtkWidget        *widget,
				    GdkRectangle     *area);
static void wp_view_draw           (GtkWidget        *widget,
				    GdkRectangle     *area);
static gint wp_view_expose         (GtkWidget        *widget,
				    GdkEventExpose   *event);
static gint wp_view_button_press   (GtkWidget        *widget,
 				    GdkEventButton   *event);
static gint wp_view_button_release (GtkWidget        *widget,
				    GdkEventButton   *event);
static gint wp_view_motion_notify  (GtkWidget         *widget,
				    GdkEventMotion    *event);
static gint wp_view_enter_notify   (GtkWidget        *widget,
				    GdkEventCrossing *event);
static gint wp_view_leave_notify   (GtkWidget        *widget,
				    GdkEventCrossing *event);
static gint wp_view_focus_in       (GtkWidget        *widget,
				    GdkEventFocus    *event);
static gint wp_view_focus_out      (GtkWidget        *widget,
				    GdkEventFocus    *event);
static gint wp_view_configure      (GtkWidget        *widget,
				    GdkEventConfigure *event);
static gint wp_view_key_press      (GtkWidget        *widget,
				    GdkEventKey      *event);


static GtkHBoxClass *parent_class;

guint wp_view_get_type ()
{
  static guint wp_type = 0;

  if (!wp_type)
    {
      GtkTypeInfo wp_view_info =
      {
	"WPView",
	sizeof (WPView),
	sizeof (WPViewClass),
	 (GtkClassInitFunc) wp_view_class_init,
	 (GtkObjectInitFunc) wp_view_init,
	 (GtkArgSetFunc) NULL,
	 (GtkArgGetFunc) NULL,
      };

      wp_type = gtk_type_unique (gtk_hbox_get_type (), &wp_view_info);
    }

  return wp_type;
}


static void
wp_view_class_init (WPViewClass *klass)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkHBoxClass *hbox_class;

  object_class = (GtkObjectClass *) klass;
  widget_class = (GtkWidgetClass *) klass;
  hbox_class = (GtkHBoxClass *) klass;

  parent_class = (GtkHBoxClass *) gtk_type_class (gtk_hbox_get_type ());

  widget_class->map = wp_view_map;
  widget_class->draw = wp_view_draw;
  widget_class->expose_event = wp_view_expose;
  /*
  widget_class->motion_notify_event = wp_view_motion_notify;
  widget_class->button_press_event = wp_view_button_press;
  widget_class->button_release_event = wp_view_button_release;
  widget_class->configure_event = wp_view_configure;
  */
  widget_class->enter_notify_event = wp_view_enter_notify;
  widget_class->leave_notify_event = wp_view_leave_notify;
  widget_class->focus_in_event = wp_view_focus_in;
  widget_class->focus_out_event = wp_view_focus_out;
  widget_class->key_press_event = wp_view_key_press;
}


static void
wp_view_init (WPView *view)
{
  GTK_WIDGET_SET_FLAGS (view, GTK_CAN_FOCUS);

  view->gc = NULL;
  view->point = NULL;
  view->mark = NULL;

  view->fg = NULL;
  view->bg = NULL;
  view->cursor_color = NULL;

  view->hl_on = false;
  view->have_selection = false;

  view->point_save_under = NULL;
  view->mark_save_under = NULL;

  view->user_set_cursor_pos_set = false;
  view->user_set_cursor_pos_x = 0;
  view->user_set_cursor_pos_y = 0;

  view->plugin_scratchspace = NULL;

  /* XXX */
  /*
  int width = GTK_WIDGET (view)->allocation.width;
  int height = GTK_WIDGET (view)->allocation.width;
  */
  int width = 654;
  int height = 300;

  view->vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (view->vbox);

  view->plugin_toolbar = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (view->plugin_toolbar);

  GtkWidget *hb = gtk_handle_box_new ();
  gtk_widget_show (hb);
  gtk_container_add (GTK_CONTAINER (hb), view->plugin_toolbar);

  gtk_box_pack_start (GTK_BOX (view->vbox), hb, FALSE, FALSE, 0);

  view->drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_usize (view->drawing_area, width, height);
  gtk_widget_show (view->drawing_area);

  view->back_pixmap = gdk_pixmap_new (view->drawing_area->window,
				      width, height,
				      gdk_visual_get_best_depth ());

  gtk_widget_set_usize (view->drawing_area, width, height);

  view->vadj = gtk_adjustment_new (0, 0, height, 10, height, height);
  view->hadj = gtk_adjustment_new (0, 0, width, 10, width, width);

  view->hscrollbar = gtk_hscrollbar_new (GTK_ADJUSTMENT (view->hadj));
  gtk_widget_show (view->hscrollbar);
  view->vscrollbar = gtk_vscrollbar_new (GTK_ADJUSTMENT (view->vadj));
  gtk_widget_show (view->vscrollbar);

  gtk_range_set_update_policy (GTK_RANGE (view->vscrollbar),
			       GTK_UPDATE_CONTINUOUS);
  gtk_range_set_update_policy (GTK_RANGE (view->hscrollbar),
			       GTK_UPDATE_CONTINUOUS);

  gtk_object_set_user_data (GTK_OBJECT (view->vadj), view);
  gtk_object_set_user_data (GTK_OBJECT (view->hadj), view);

  gtk_signal_connect (view->hadj, "value_changed",
		      (GtkSignalFunc) wp_view_scroll_event, NULL);
  gtk_signal_connect (view->vadj, "value_changed",
		      (GtkSignalFunc) wp_view_scroll_event, NULL);


  gtk_box_pack_start (GTK_BOX (view->vbox), view->drawing_area, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (view->vbox), view->hscrollbar, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (view), view->vbox, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (view), view->vscrollbar, FALSE, FALSE, 0);

  gtk_widget_set_events (view->drawing_area,
			 gtk_widget_get_events (view->drawing_area) |
			 (
			  /*
			  GDK_KEY_PRESS_MASK |
			  GDK_KEY_RELEASE_MASK |
			  GDK_ENTER_NOTIFY_MASK |
			  GDK_LEAVE_NOTIFY_MASK |
			  */
			  GDK_EXPOSURE_MASK |
			  GDK_BUTTON_PRESS_MASK |
			  GDK_BUTTON_RELEASE_MASK |
			  GDK_BUTTON1_MOTION_MASK |
			  GDK_STRUCTURE_MASK)
			 );

  gtk_signal_connect (GTK_OBJECT(view->drawing_area), "event",
		      (GtkSignalFunc) wp_view_drawing_area_event, view);

  gtk_selection_add_target (view->drawing_area,
			    GDK_SELECTION_PRIMARY,
			    GDK_SELECTION_TYPE_STRING,
			    0);

  gtk_signal_connect (GTK_OBJECT (view->drawing_area), "selection_get",
		      (GtkSignalFunc) view_selection_get, view);
  gtk_signal_connect (GTK_OBJECT (view->drawing_area), "selection_received",
		      (GtkSignalFunc) view_selection_received, view);
  
  /* these are used to keep the cursor where it should
     be during previous and next line commands. */
  view->user_set_cursor_pos_set = 0;
  view->user_set_cursor_pos_x = 0;
  view->user_set_cursor_pos_y = 0;

  view->hl_on = 0;
  view->have_selection = 0;

  view->scale = 1.0;

  int depth;
  gdk_window_get_geometry (view->drawing_area->window,
			   NULL, NULL, NULL, NULL, &depth);

  view->point_save_under =
    gdk_pixmap_new (view->drawing_area->window,
		    3, 200, depth); /* XXX what size? */
  view->mark_save_under =
    gdk_pixmap_new (view->drawing_area->window,
		    3, 200, depth); /* XXX what size? */

  view->plugin_scratchspace = g_hash_table_new (g_str_hash, g_str_equal);

  wp_view_hl_off (view);

  view->next_key_is_command = false;
  view->cursor_on = false;
}


GtkWidget *wp_view_new ()
{
  WPView *view;
  view = (WPView *) gtk_type_new (wp_view_get_type ());
  return GTK_WIDGET (view);
}


void
wp_view_activate_plugin_callback (GtkWidget */*widget*/,
				  gpointer client_data)
{
  g_return_if_fail (client_data);
  plugin_info *info = (plugin_info *) client_data;

  printf ("    Go plugin:%s activated\n", info->plugin_name);

  plugin_callback_struct callbacks;
  plugin *plug = plugin_new (info->plugin_name);
  gwp_setup_plugin_callbacks (&callbacks);
  int context = GPOINTER_TO_INT (WP_VIEW (mdi->active_view));
  plugin_register (plug, &callbacks, context);
}


static void
wp_view_add_to_plugins_menu (WPView *view, plugin_info *info)
{
  char buf[200];
  GnomeUIInfo entry[2];

  g_snprintf (buf, sizeof(buf), "%s", info->menu_location);

  /*
  printf ("adding Go plugin menu item: location='%s'  acc='%s'  name='%s'\n",
	  info->menu_location ? info->menu_location : "",
	  info->suggested_accelerator ? info->suggested_accelerator : "",
	  info->plugin_name ? info->plugin_name : "");
  */

      /*
	info->type
	char * info->menu_location
	char * info->suggested_accelerator
	char * info->plugin_name
      */

  entry[0].type = GNOME_APP_UI_ITEM;
  entry[0].label = buf;
  entry[0].hint = NULL;
  entry[0].moreinfo = wp_view_activate_plugin_callback;
  entry[0].user_data = info;
  entry[0].unused_data = NULL;
  entry[0].pixmap_type = GNOME_APP_PIXMAP_NONE;
  entry[0].pixmap_info = NULL;
  entry[0].accelerator_key = 0;
  entry[1].type = GNOME_APP_UI_ENDOFINFO;

  GnomeApp *app = gnome_mdi_get_app_from_view (GTK_WIDGET (view));

  int menu_pos;
  char menu_item_path[ strlen (entry[0].label) + 20 ];
  sprintf (menu_item_path, "Plugins/%s", entry[0].label);
  if (gnome_app_find_menu_pos (app->menubar, menu_item_path, &menu_pos))
    return; /* already did this one */

  gnome_app_insert_menus_with_data (app, "Plugins/", entry, info);
}


/* ugh */
static void
wp_view_box_map (GtkWidget *widget)
{
  GtkBox *box;
  GtkBoxChild *child;
  GList *children;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_BOX (widget));

  box = GTK_BOX (widget);
  GTK_WIDGET_SET_FLAGS (box, GTK_MAPPED);

  children = box->children;
  while (children)
    {
      child = (GtkBoxChild *) children->data;
      children = children->next;

      if (GTK_WIDGET_VISIBLE (child->widget) &&
	  !GTK_WIDGET_MAPPED (child->widget))
	gtk_widget_map (child->widget);
    }
}


static void wp_view_map (GtkWidget *widget)
{
  wp_view_box_map (widget);

  g_return_if_fail (widget != NULL);
  g_return_if_fail (IS_WP_VIEW (widget));

  WPView *view = WP_VIEW (widget);

  /* tell each plugin to add their stuff to this view */
  for (GList *cur = WPPluginManager::plugins;
       cur != NULL;
       cur = cur->next)
    {
      WPPlugin *plugin = (WPPlugin *) cur->data;
      /*
      printf ("calling %s->newUI on view '%p'\n", plugin->getUIName (), view);
      */
      plugin->newUI (view);
    }

  /* add a menu entry for each Go plugin */
  for (GList *go_plugin = go_plugins; go_plugin; go_plugin=go_plugin->next)
    {
      plugin_info *info = (plugin_info *) (go_plugin->data);
      wp_view_add_to_plugins_menu (view, info);
    }


  GdkColormap *colormap = gtk_widget_get_colormap (view->drawing_area);
  
  view->bg = (GdkColor *) malloc (sizeof (GdkColor));
  view->bg->red = 0xA5A5;
  view->bg->green = 0xA5A5;
  view->bg->blue = 0xA5A5;
  gdk_color_alloc (colormap, view->bg);
  
  view->fg = (GdkColor *) malloc (sizeof (GdkColor));
  view->fg->red = 0x0000;
  view->fg->green = 0x0000;
  view->fg->blue = 0x0000;
  gdk_color_alloc (colormap, view->fg);

  view->cursor_color = (GdkColor *) malloc (sizeof (GdkColor));
  view->cursor_color->red = 0;
  view->cursor_color->green = 0x8000;
  view->cursor_color->blue = 0x4000;
  gdk_color_alloc (colormap, view->cursor_color);

  gdk_window_set_background (view->drawing_area->window, view->bg);

  view->gc = gdk_gc_new (view->drawing_area->window);

  /* XXX this doesn't belong here */
  WPDocument *doc = wp_view_get_document (view);
  view->point = doc->newDocumentSpot ();
  view->mark = doc->newDocumentSpot ();
}


static void wp_view_paint (GtkWidget *widget, GdkRectangle *area)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (IS_WP_VIEW (widget));

  WPView *view = WP_VIEW (widget);

  gdk_gc_set_foreground (view->gc, view->bg);
  wp_view_fill_rect (view, area->x, area->y, area->width, area->height);

  WPDocument *doc = wp_view_get_document (view);

  if (preferences->draw_fast)
    {
      /* convert screen coords to document coords */
      GdkRectangle a;
      
      a.x = view_xcoord_to_doc (view, area->x);
      a.y = view_ycoord_to_doc (view, area->y);
      a.width = view_coords_to_doc (view, area->width);
      a.height = view_coords_to_doc (view, area->height);
      doc->paint_slab (view, NULL, &a);
    }
  else
    doc->refresh ();

  if (view->cursor_on)
    wp_view_draw_cursor (view);
}


static void wp_view_draw (GtkWidget *widget, GdkRectangle *area)
{
  WPView *view;
  GdkRectangle child_area;

  /*printf ("wp_view_draw called:%p\n", widget);*/

  g_return_if_fail (widget != NULL);
  g_return_if_fail (IS_WP_VIEW (widget));

  if (GTK_WIDGET_VISIBLE (widget) && GTK_WIDGET_MAPPED (widget))
    {
      view = WP_VIEW (widget);

      
      if (GTK_WIDGET_VISIBLE (view->vbox) &&
	  gtk_widget_intersect (view->vbox, area, &child_area))
	gtk_widget_draw (view->vbox, &child_area);
  

      if (GTK_WIDGET_VISIBLE (view->vscrollbar) &&
	  gtk_widget_intersect (view->vscrollbar, area, &child_area))
	gtk_widget_draw (view->vscrollbar, &child_area);

      wp_view_paint (widget, area);
    }
}


static gint wp_view_expose (GtkWidget *widget, GdkEventExpose *event)
{
  GdkEventExpose child_event;

  /*printf ("wp_view_expose called:%p\n", widget);*/

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  WPView *view = WP_VIEW (widget);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      child_event = *event;

      if (GTK_WIDGET_NO_WINDOW (view->vbox) &&
	  gtk_widget_intersect (view->vbox, &event->area, &child_event.area))
	gtk_widget_event (view->vbox, (GdkEvent*) &child_event);

      if (GTK_WIDGET_NO_WINDOW (view->vscrollbar) &&
	  gtk_widget_intersect (view->vscrollbar,
				&event->area, &child_event.area))
	gtk_widget_event (view->vscrollbar, (GdkEvent*) &child_event);

      wp_view_paint (widget, &event->area);
    }

  wp_view_sync_scrollbars (WP_VIEW (widget));

  return FALSE;
}


static gint
wp_view_button_press (GtkWidget *widget, GdkEventButton *event)
{
  WPView *view;
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type == GDK_BUTTON_PRESS)
    {
      view = WP_VIEW (widget);

      if (GTK_WIDGET_CAN_DEFAULT (widget) && (event->button == 1))
	gtk_widget_grab_default (widget);
      if (!GTK_WIDGET_HAS_FOCUS (widget))
	gtk_widget_grab_focus (widget);

      WPDocument *doc = wp_view_get_document (view);

      int gebx = (int) event->x;
      int geby = (int) event->y;

      if (event->button == 1)
	{
	    int hl_was_on = wp_view_hl_is_on (view);
	    wp_view_hl_off (view);

	    if (view == WP_VIEW (mdi->active_view))
	      {
	        wp_view_cursor_off (view);

	        doc->setSpotToPoint (view->point, view,
				     view_xcoord_to_doc (view, gebx),
				     view_ycoord_to_doc (view, geby));
    
	        doc->setSpotToPoint (view->mark, view,
				     view_xcoord_to_doc (view, gebx),
				     view_ycoord_to_doc (view, geby));

	        wp_view_cursor_on (view);

	        if (hl_was_on)
	          {
		    GdkRectangle a;
		    a.x = wp_view_get_x (view);
		    a.y = wp_view_get_y (view);
		    a.width = wp_view_get_pixel_width (view);
		    a.height = wp_view_get_pixel_height (view);

		    gdk_gc_set_foreground (view->gc, view->bg);
		    wp_view_fill_rect (view, a.x, a.y, a.width, a.height);
		    doc->paint_slab (view, NULL, &a);
	          }
	      }
	    else
	      {
		printf ("need to set focus view here\n");
	      }

	    wp_view_clear_user_set_cursor_pos (view);

	    if (! preferences->draw_fast)
	      wp_view_copy_back_pixmap (view);
	    
	    doc->getPlugin ()->updateUI (view);
	  }
	else if (event->button == 2)
	  {
	    doc->setSpotToPoint (view->point, view,
				 view_xcoord_to_doc (view, gebx),
				 view_ycoord_to_doc (view, geby));

	    gtk_selection_convert (view->drawing_area, GDK_SELECTION_PRIMARY,
				   GDK_SELECTION_TYPE_STRING, event->time);
	  }
	else
	  {
	    doc->getPlugin ()->buttonEvent (view, event);
	  }

	return TRUE;
    }
  
  return TRUE;
}


static gint wp_view_button_release (GtkWidget *widget, GdkEventButton *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  return TRUE;
}


static gint wp_view_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{
  WPView *view;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  view = WP_VIEW (widget);

  WPDocument *doc = wp_view_get_document (view);

  Slab a = view->point->getSlab ();

  int gemx = (int) event->x;
  int gemy = (int) event->y;

  wp_view_cursor_off (view);
  doc->setSpotToPoint (view->point, view,
		       view_xcoord_to_doc (view, gemx),
		       view_ycoord_to_doc (view, gemy));

  if (view->point->getSlab () != view->mark->getSlab () ||
      view->point->getPosition () != view->mark->getPosition ())
    wp_view_hl_on (view);
  else
    wp_view_hl_off (view);

  wp_view_cursor_on (view);

  Slab b = view->point->getSlab ();

  if (wp_slab_order (b, a))
    {
      Slab c = a; a = b; b = c;
    }

  GdkRectangle r;
  r.x = wp_view_get_x (view);
  r.y = wp_view_get_y (view);
  r.width = wp_view_get_width (view);
  r.height = wp_view_get_height (view);

  Slab n;
  for (n=a;
       n != NULL && wp_slab_order (n, b);
       n = n->getNextLeaf ())
    {
      doc->erase_slab (view, n, &r);
      doc->paint_slab (view, n, &r);
    }

  if (! preferences->draw_fast)
    wp_view_copy_back_pixmap (view);

  wp_view_scroll_point_into_view (view);
  wp_view_clear_user_set_cursor_pos (view);
  return TRUE;

  /*return FALSE;*/
}


static gint wp_view_enter_notify (GtkWidget *widget, GdkEventCrossing *event)
{
  WPView *view;
  GtkWidget *event_widget;

  /*printf ("wp_view_enter_notify called:%p\n", widget);*/

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  view = WP_VIEW (widget);
  event_widget = gtk_get_event_widget ( (GdkEvent*) event);

  if ( (event_widget == widget) &&
      (event->detail != GDK_NOTIFY_INFERIOR))
    {
      printf ("wp_view enter notify\n");
    }

  return FALSE;
}


static gint wp_view_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
  WPView *view;
  GtkWidget *event_widget;

  /*printf ("wp_view_leave_notify called:%p\n", widget);*/

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  view = WP_VIEW (widget);
  event_widget = gtk_get_event_widget ( (GdkEvent*) event);

  if ( (event_widget == widget) &&
      (event->detail != GDK_NOTIFY_INFERIOR))
    {
      printf ("wp_view leave notify\n");
    }

  return FALSE;
}


static gint wp_view_focus_in (GtkWidget *widget, GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);

  return FALSE;
}


static gint wp_view_focus_out (GtkWidget *widget, GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);

  return FALSE;
}


static gint
wp_view_configure (GtkWidget *widget, GdkEventConfigure *event)
{
  /*printf ("wp_view_configure called:%p\n", widget);*/

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type == GDK_CONFIGURE)
    {
      WPView *view = WP_VIEW (widget);

      wp_view_sync_scrollbars (WP_VIEW (widget));
      
      if (view->back_pixmap)
	{
	  gdk_pixmap_unref (view->back_pixmap);
	  view->back_pixmap =
	    gdk_pixmap_new (view->drawing_area->window,
			    GTK_WIDGET (view)->allocation.width,
			    GTK_WIDGET (view)->allocation.height,
			    gdk_visual_get_best_depth ());

	  WPDocument *doc = wp_view_get_document (view);
	  doc->refresh ();
	}
      else
	{
	  printf ("  --> no back pixmap yet\n");
	}

	return FALSE;
    }

  return FALSE;
}



static gint
wp_view_key_press (GtkWidget *widget, GdkEventKey *event)
{
  WPView *view;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (IS_WP_VIEW (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->type == GDK_KEY_PRESS)
    {
      view = WP_VIEW (widget);
      wp_view_cursor_off (view);

      if (event->state & GDK_CONTROL_MASK)
	{
	  switch (event->keyval)
	    {
	    case GDK_a:
	    case GDK_Home:
	    case GDK_KP_Home:
	      wp_view_move_to_start_of_line (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_b:
	      wp_view_move_to_previous_character (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_Left:
	    case GDK_KP_Left:
	      wp_view_move_to_previous_word (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_d:
	      wp_view_delete_char (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_e:
	    case GDK_End:
	    case GDK_KP_End:
	      wp_view_move_to_end_of_line (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_f:
	      wp_view_move_to_next_character (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_Right:
	    case GDK_KP_Right:
	      wp_view_move_to_next_word (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_g:
	      gdk_beep();
	      view->next_key_is_command = false;
	      break;
	    case GDK_n:
	    case GDK_Down:
	    case GDK_KP_Down:
	      wp_view_move_to_next_line (view);
	      break;
	    case GDK_p:
	    case GDK_Up:
	    case GDK_KP_Up:
	      wp_view_move_to_previous_line (view);
	      break;
	    case GDK_x:
	      view->next_key_is_command = true;
	      break;
	    case GDK_l:
	      /*view_refresh_cmd_callback (NULL, frame);*/
	      wp_view_get_document (view)->refresh ();
	      break;
	    default:
	      break;
	    }
	}
      else if (event->state & GDK_MOD1_MASK)
	{
	  switch (event->keyval)
	    {
	    case GDK_f:
	      wp_view_move_to_next_word (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_b:
	      wp_view_move_to_previous_word (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    }
	}
      else if (view->next_key_is_command)
	{
	  switch (event->keyval)
	    {
	    case GDK_o:
	      {
		/* switch to other view. XXX */
		break;
	      }
	    default:
	      gdk_beep();
	      break;
	    }

	  view->next_key_is_command = false;
	}
      else
	{
	  switch (event->keyval)
	    {
	    case GDK_Home:
	    case GDK_KP_Home:
	      wp_view_move_to_start_of_line (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_Left:
	    case GDK_KP_Left:
	      wp_view_move_to_previous_character (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_End:
	    case GDK_KP_End:
	      wp_view_move_to_end_of_line (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_Right:
	    case GDK_KP_Right:
	      wp_view_move_to_next_character (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    case GDK_Down:
	    case GDK_KP_Down:
	      wp_view_move_to_next_line (view);
	      break;
	    case GDK_Up:
	    case GDK_KP_Up:
	      wp_view_move_to_previous_line (view);
	      break;

	    case GDK_Shift_L:
	    case GDK_Shift_R:
	    case GDK_Control_L:
	    case GDK_Control_R:
	    case GDK_Caps_Lock:
	    case GDK_Shift_Lock:
	    case GDK_Meta_L:
	    case GDK_Meta_R:
	    case GDK_Alt_L:
	    case GDK_Alt_R:
	    case GDK_Super_L:
	    case GDK_Super_R:
	    case GDK_Hyper_L:
	    case GDK_Hyper_R:
	      break;
	    case GDK_Escape:
	      wp_view_dumpToScreen (view);
	      break;
	      /*
		case GDK_F1:
		wp_view_insert_image (view, frame->window, 5, 5);
		break;
		case GDK_F2:
		wp_view_insert_image (view, frame->window, 15, 15);
		break;
		case GDK_F3:
		wp_view_insert_image (view, frame->window, 45, 45);
		break;
	      */
	    case GDK_BackSpace:
	    case GDK_Delete:
	      if (wp_view_move_to_previous_character (view))
		wp_view_delete_char (view);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    default:
	      wp_view_insert_char(view, event->keyval);
	      wp_view_clear_user_set_cursor_pos (view);
	      break;
	    }
	}

      wp_view_cursor_on (view);

      return TRUE;
    }

  return FALSE;
}


static inline int doc_coords_to_view (WPView *view, long int v)
{
  if (view->scale != 1)
    v = (int) ( (double) v * view->scale);
  return v;
}


static inline int doc_xcoord_to_view (WPView *view, long int v)
{
  if (view->scale != 1)
    v = (int) ( (double) v * view->scale);
  v += wp_view_get_scroll_x (view);
  return v;
}


static inline int doc_ycoord_to_view (WPView *view, int v)
{
  if (view->scale != 1)
    v = (int) ( (double) v * view->scale);
  v += wp_view_get_scroll_y (view);
  return v;
}


static inline int view_coords_to_doc (WPView *view, int v)
{
  if (view->scale != 1)
    v = (int) ( (double) v / view->scale);
  return v;
}


static inline int view_xcoord_to_doc (WPView *view, int v)
{
  v -= wp_view_get_scroll_x (view);
  if (view->scale != 1)
    v = (int) ( (double) v / view->scale);
  return v;
}


static inline int view_ycoord_to_doc (WPView *view, int v)
{
  v -= wp_view_get_scroll_y (view);
  if (view->scale != 1)
    v = (int) ( (double) v / view->scale);
  return v;
}


/**
 * wp_view_sync_scrollbars:set scrollbar sizes
 * @view: the view
 * 
 * Sets the range and size of @view's scrollbars so that
 * the entire document can be scrolled over.
 **/
void wp_view_sync_scrollbars (WPView *view)
{
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  GtkAdjustment *adj;

  adj = gtk_range_get_adjustment (GTK_RANGE (view->vscrollbar));
  adj->upper = doc->getDocumentHeight () * view->scale;
  gtk_signal_emit_by_name (GTK_OBJECT (adj), "changed", view);

  adj = gtk_range_get_adjustment (GTK_RANGE (view->hscrollbar));
  adj->upper = doc->getDocumentWidth () * view->scale;
  gtk_signal_emit_by_name (GTK_OBJECT (adj), "changed", view);
}


/* callback for vertical scroll events */
static gint
wp_view_scroll_event (GtkWidget *widget,
		      GdkEvent */*event*/, void */*client_data*/)
{
  /*GtkAdjustment *ga = GTK_ADJUSTMENT (widget);*/
  WPView *view = WP_VIEW (gtk_object_get_user_data (GTK_OBJECT (widget)));
  if (! view->gc) return false;

  GdkRectangle a;
  a.x = wp_view_get_x (view);
  a.y = wp_view_get_y (view);
  a.width = wp_view_get_pixel_width (view);
  a.height = wp_view_get_pixel_height (view);

  gdk_gc_set_foreground (view->gc, view->bg);
  wp_view_fill_rect (view, a.x, a.y, a.width, a.height);

  WPDocument *doc = wp_view_get_document (view);

  if (preferences->draw_fast)
    {
      a.x = view_xcoord_to_doc (view, a.x);
      a.y = view_ycoord_to_doc (view, a.y);
      a.width = view_coords_to_doc (view, a.width);
      a.height = view_coords_to_doc (view, a.height);
      doc->paint_slab (view, NULL, &a);
    }
  else
    doc->refresh ();

  return TRUE;
}


void
view_selection_get (GtkWidget * /*widget*/,
		    GtkSelectionData *selection_data,
		    guint       /*info*/,
		    guint       /*time*/,
		    gpointer data)
{
  guchar *string;
  WPView *view = (WPView *)data;

  /* I have to do a lot of typecasting with char and unsigned char to get
     the compiler to shut up.  it should be looked into more */
  
  string = wp_clipboard_selection_copy (view, view->mark, view->point);
  
  gtk_selection_data_set (selection_data, GDK_SELECTION_TYPE_STRING,
                          8, string, strlen ( (char *) string));
  free (string);
}


void
view_selection_received (GtkWidget * /*widget*/,
			 GtkSelectionData *selection_data,
			 guint    /*time*/,
			 gpointer data)
{
  /* XXX combine some of this is wp_clipboard.C:wp_clipboard_paste */

  WPView *view = (WPView *) data;
  WPDocument *doc = wp_view_get_document (view);
  guchar *string;

  if (doc == NULL)
    return;

  doc->sanityCheckSpots ();

  /* XXX **** IMPORTANT **** Check to see if retrieval succeeded  */
  if (selection_data->length < 0)
    {
      g_print ("Selection retrieval failed\n");
      return;
    }
  /* Make sure we got the data in the expected form */
  if (selection_data->type != GDK_SELECTION_TYPE_STRING)
    {
      g_print ("Selection \"STRING\" was not returned as atoms!\n");
      return;
    }
        
  string = (guchar *)selection_data->data;
  
  /*printf ("The selection received is %s\n", string);*/

  int length = strlen ( (char *) string);

  doc->displayRefreshOff ();

  for (int i=0; i<length; i++)
    wp_view_insert_char (view, string[ i ]);

  doc->displayRefreshOn ();
  doc->refresh ();

  return;
}


gint
wp_view_drawing_area_event (GtkWidget *widget,
			    GdkEvent *event, void */*client_data*/)
{
  g_return_val_if_fail (widget, false);
  g_return_val_if_fail (widget->parent, false);
  g_return_val_if_fail (widget->parent->parent, false);
  g_return_val_if_fail (IS_WP_VIEW (widget->parent->parent), false);
  g_return_val_if_fail (event, false);

  GtkWidget *view = widget->parent->parent;

  switch (event->type)
    {
    case GDK_CONFIGURE:
      return wp_view_configure (view, (GdkEventConfigure *) event);
    case GDK_MOTION_NOTIFY:
      return wp_view_motion_notify (view, (GdkEventMotion *) event);
    case GDK_BUTTON_PRESS: 
      return wp_view_button_press (view, (GdkEventButton *) event);
    case GDK_BUTTON_RELEASE: 
      return wp_view_button_release (view, (GdkEventButton *) event);
    case GDK_EXPOSE:
      return wp_view_expose (view, (GdkEventExpose *) event);
    default:
      return FALSE;
    }
}



GtkWidget *wp_view_get_plugin_toolbar (WPView *view)
{
  return view->plugin_toolbar;
}



/**
 * wp_view_get_plugin_scratch_space:get a plugin's scratch area
 * @view: the view
 * @plugin_name: the name of the plugin that wants its scratch area
 * 
 * Used to retrieve a block of data that a plugin has stored
 * with the view.  The view does not try to interpret the
 * data that is stored this way.  use wp_view_set_plugin_scratch_space ()
 * to set the data.  This returns NULL if no data has been tied
 * to the name @plugin_name.
 * 
 * Return value: pointer to the scratch area or NULL
 **/
void *
wp_view_get_plugin_scratch_space (WPView *view, char *plugin_name)
{
  void *space = g_hash_table_lookup (view->plugin_scratchspace,
				     plugin_name);
  return space;
}


/**
 * wp_view_set_plugin_scratch_space:set a plugin's scratch area
 * @view: the view
 * @plugin_name: the name of the plugin that is saving data with @view
 * @space: pointer to the scratch area
 * 
 * Used by plugins to tie a block of data to a view.  The view doesn't
 * interpret the data -- it just holds on to it.
 * use wp_view_get_plugin_scratch_space () to retrieve the data.
 **/
void
wp_view_set_plugin_scratch_space (WPView *view, char *plugin_name, void *space)
{
  g_hash_table_insert (view->plugin_scratchspace, plugin_name, space);
}


/**
 * wp_view_destroy:delete a view
 * @view: the view to delete
 * 
 * unhooks @view from @view's document and frees @view
 **/
void
wp_view_destroy (WPView */*view*/)
{
  /*
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  */
}


/**
 * wp_view_get_drawing_area:get a view's drawing area
 * @view: the view
 * 
 * Used to get the drawing area from @view
 * 
 * Return value: @view's drawing area
 **/
GtkWidget *
wp_view_get_drawing_area (WPView *view)
{
  return view->drawing_area;
}


/**
 * wp_view_set_color:set the drawing color
 * @view: the view
 * @c: color to do future drawing in
 * 
 * all drawing done through the view's drawing api will
 * be in color @c until this is called again.
 **/
void
wp_view_set_color (WPView *view, GdkColor *c)
{
  gdk_gc_set_foreground (view->gc, c);
}


/**
 * wp_view_get_cursor_color:get the color of view's cursor
 * @view: the view
 * 
 * the view's cursor color is used for drawing the cursor and
 * the background of highlighted text
 * 
 * Return value: the color of view's cursor
 **/
GdkColor *
wp_view_get_cursor_color (WPView *view)
{
  return view->cursor_color;
}


/**
 * wp_view_draw_string:draw text in a view
 * @view: the view
 * @word: text to draw
 * @len: length of @word in bytes (rather than characters)
 * @font: font to draw @word in
 * @x: x position within view of left edge of text to draw
 * @y: x position within view of baseline of text to draw
 * @d: directin of the left
 * 
 * used to draw text in a view.  x and y are from the upper
 * left of the view, and are in pixels.
 **/
void
wp_view_draw_string (WPView *view,
		     char *word, int len, GdkFont *font,
		     long int x, long int y, WPDirection d)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  if (d == DIR_RIGHT)
    {
      if (preferences->draw_fast)
	gdk_draw_text (view->drawing_area->window, font, view->gc,
		       doc_xcoord_to_view (view, x),
		       doc_ycoord_to_view (view, y),
		       word, len);
      else
	gdk_draw_text (view->back_pixmap, font, view->gc,
		       doc_xcoord_to_view (view, x),
		       doc_ycoord_to_view (view, y),
		       word, len);
    }
  else /* assume DIR_LEFT */
    {
      /* ickky */
      int i;
      int wid;

      /* XXX assumes 1 byte/char */
      char *tmp = (char *) malloc (len + 1);
      for (i=0; i<len; i++)
	tmp[ i ] = word[ len-i-1 ];
      tmp[ len ] = '\0';

      wid = gdk_text_width (font, tmp, len);
      if (preferences->draw_fast)
	gdk_draw_text (view->drawing_area->window, font, view->gc,
		       doc_xcoord_to_view (view, x - wid),
		       doc_ycoord_to_view (view, y),
		       tmp, len);
      else
	gdk_draw_text (view->back_pixmap, font, view->gc,
		       doc_xcoord_to_view (view, x - wid),
		       doc_ycoord_to_view (view, y),
		       tmp, len);
      free (tmp);
    }
}


/**
 * wp_view_draw_arc:draw an arc on the view
 * @view: the view
 * @filled: should the arc be filled?
 * @x1: center x
 * @y1: center y
 * @x2: width
 * @y2: height
 * @angle1: start angle
 * @angle2: end angle
 * 
 * x and y are measured from the upper left of the view in pixels
 **/
void
wp_view_draw_arc (WPView *view,
		  int filled,
		  long int x1, long int y1, long int width, long int height,
		  int angle1, int angle2)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;


  if (preferences->draw_fast)
    gdk_draw_arc (view->drawing_area->window, view->gc,
		  filled,
		  doc_xcoord_to_view (view, x1),
		  doc_ycoord_to_view (view, y1),
		  doc_coords_to_view (view, width),
		  doc_coords_to_view (view, height),
		  angle1, angle2);
  else
    gdk_draw_arc (view->back_pixmap, view->gc,
		  filled,
		  doc_xcoord_to_view (view, x1),
		  doc_ycoord_to_view (view, y1),
		  doc_coords_to_view (view, width),
		  doc_coords_to_view (view, height),
		  angle1, angle2);
}


/**
 * wp_view_draw_line:draw a line on the view
 * @view: the view
 * @x1: x1
 * @y1: y1
 * @x2: x2
 * @y2: y2
 * 
 * draws a line on the view from (@x1, @y1) to (@x2, @y2)
 **/
void
wp_view_draw_line (WPView *view,
		   long int x1, long int y1, long int x2, long int y2)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  if (preferences->draw_fast)
    gdk_draw_line (view->drawing_area->window, view->gc,
		   doc_xcoord_to_view (view, x1),
		   doc_ycoord_to_view (view, y1),
		   doc_xcoord_to_view (view, x2),
		   doc_ycoord_to_view (view, y2));
  else
    gdk_draw_line (view->back_pixmap, view->gc,
		   doc_xcoord_to_view (view, x1),
		   doc_ycoord_to_view (view, y1),
		   doc_xcoord_to_view (view, x2),
		   doc_ycoord_to_view (view, y2));
}


void
wp_view_fill_rect (WPView *view,
		   long int x, long int y, long int width, long int height)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  if (preferences->draw_fast)
    gdk_draw_rectangle (view->drawing_area->window, view->gc, TRUE,
			doc_xcoord_to_view (view, x),
			doc_ycoord_to_view (view, y),
			doc_coords_to_view (view, width),
			doc_coords_to_view (view, height));
  else
    gdk_draw_rectangle (view->back_pixmap, view->gc, TRUE,
			doc_xcoord_to_view (view, x),
			doc_ycoord_to_view (view, y),
			doc_coords_to_view (view, width),
			doc_coords_to_view (view, height));
}


void
wp_view_draw_rect (WPView *view,
		   long int x, long int y, long int width, long int height)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;


  if (preferences->draw_fast)
    gdk_draw_rectangle (view->drawing_area->window, view->gc, FALSE,
			doc_xcoord_to_view (view, x),
			doc_ycoord_to_view (view, y),
			doc_coords_to_view (view, width),
			doc_coords_to_view (view, height));
  else
    gdk_draw_rectangle (view->back_pixmap, view->gc, FALSE,
			doc_xcoord_to_view (view, x),
			doc_ycoord_to_view (view, y),
			doc_coords_to_view (view, width),
			doc_coords_to_view (view, height));
}


void
wp_view_copy_area (WPView *view, 
		   int sx, int sy, int width, int height, int dx, int dy)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  dx = doc_coords_to_view (view, dx);
  dy = doc_coords_to_view (view, dy);
  sx = doc_xcoord_to_view (view, sx);
  sy = doc_ycoord_to_view (view, sy);
  width = doc_coords_to_view (view, width);
  height = doc_coords_to_view (view, height);

  if (preferences->draw_fast)
    {
      gdk_gc_set_exposures (view->gc, 0);
      gdk_window_copy_area (view->drawing_area->window,
			    view->gc,
			    sx+dx, sy+dy,
			    view->drawing_area->window,
			    sx, sy,
			    width, height);
    }
  else
    gdk_window_copy_area (view->back_pixmap,
			  view->gc,
			  sx+dx, sy+dy,
			  view->drawing_area->window,
			  sx, sy,
			  width, height);
}


void
wp_view_copy_from_pixmap (WPView *view,
			  GdkPixmap *pix,
			  int sx, int sy,
			  int width, int height,
			  int dx, int dy)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  if (preferences->draw_fast)
    {
      /* XXX scale? */
      gdk_gc_set_exposures (view->gc, 0);
      gdk_window_copy_area (view->drawing_area->window,
			    view->gc,
			    doc_xcoord_to_view (view, sx+dx),
			    doc_ycoord_to_view (view, sy+dy),
			    pix,
			    /*
			    doc_xcoord_to_view (view, sx),
			    doc_ycoord_to_view (view, sy),
			    doc_coords_to_view (view, width),
			    doc_coords_to_view (view, height)
			    */
			    sx, sy, width, height);
    }
  else
    gdk_window_copy_area (view->back_pixmap,
			  view->gc,
			  doc_xcoord_to_view (view, sx+dx),
			  doc_ycoord_to_view (view, sy+dy),
			  pix,
			  /*
			  doc_xcoord_to_view (view, sx),
			  doc_ycoord_to_view (view, sy),
			  doc_coords_to_view (view, width),
			  doc_coords_to_view (view, height)
			  */
			  sx, sy, width, height);
}


void wp_view_dumpToScreen (WPView *view)
{
  g_return_if_fail (view);

  printf ("point=[%p, %d], mark=[%p, %d]\n",
	  view->point->getSlab (), view->point->getPosition (),
	  view->mark->getSlab (), view->mark->getPosition ());

  WPDocument *doc = wp_view_get_document (view);
  doc->dumpToScreen ();
}


WPDocument *
wp_view_get_document (WPView *view)
{
  g_return_val_if_fail (view, NULL);

  WPMDIDocument *mdi_doc =
    WP_MDI_DOCUMENT (gnome_mdi_get_child_from_view (GTK_WIDGET (view)));
  WPDocument *doc = mdi_doc->doc;
  return doc;
}


void
wp_view_refresh (WPView *view)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  if (preferences->draw_fast)
    gdk_window_clear_area_e (view->drawing_area->window,
			     0, 0,
			     wp_view_get_pixel_width (view),
			     wp_view_get_pixel_height (view));
  else
    doc->refresh ();
}


int wp_view_get_scroll_x (WPView *view)
{
  GtkRange *range = GTK_RANGE (view->hscrollbar);
  GtkAdjustment *adj = gtk_range_get_adjustment (range);

  /*return (int) (view->document->getDocumentLeft () - adj->value);*/
  return (int) ( - adj->value);
  /*return view_coords_to_doc (view, (int) ( - adj->value));*/
}


int wp_view_get_scroll_y (WPView *view)
{
  GtkRange *range = GTK_RANGE (view->vscrollbar);
  GtkAdjustment *adj = gtk_range_get_adjustment (range);
  /*return (int) (view->document->getDocumentUp () - adj->value);*/
  return (int) ( - adj->value);
  /*return view_coords_to_doc (view, (int) ( - adj->value));*/
}


int wp_view_get_x (WPView *view)
{
  return view_xcoord_to_doc (view, 0);
}


int wp_view_get_y (WPView *view)
{
  return view_ycoord_to_doc (view, 0);
}


/**
 * wp_view_get_width:
 * @view: 
 * 
 * get view width in points
 * 
 * Return value: 
 **/
int wp_view_get_width (WPView *view)
{
  return view_coords_to_doc (view, view->drawing_area->allocation.width);
}


/**
 * wp_view_get_height:
 * @view: 
 * 
 * get view height in points
 * 
 * Return value: 
 **/
int wp_view_get_height (WPView *view)
{
  return view_coords_to_doc (view, view->drawing_area->allocation.height);
}


int wp_view_get_pixel_width (WPView *view)
{
  return view->drawing_area->allocation.width;
}


int wp_view_get_pixel_height (WPView *view)
{
  return view->drawing_area->allocation.width;
}




double wp_view_get_scale_factor (WPView *view)
{
  return view->scale;
}


void wp_view_set_scale_factor (WPView *view, double set_scale)
{
  g_return_if_fail (view);
  g_return_if_fail (set_scale > 0.0);

  if (view->scale == set_scale) return;
  view->scale = set_scale;

  WPDocument *doc = wp_view_get_document (view);
  doc->refresh ();
}


void
wp_view_draw_cursor (WPView *view)
{
  int x, y, height;

  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  doc->sanityCheckSpots ();

  doc->getPointFromSpot (view->point, view, &x, &y, &height);

  x = doc_xcoord_to_view (view, x);
  y = doc_ycoord_to_view (view, y);
  height = doc_coords_to_view (view, height);

  gdk_gc_set_foreground (view->gc, view->cursor_color);
  gdk_gc_set_exposures (view->gc, 0);
  gdk_window_copy_area (view->point_save_under,
			view->gc,
			0, 0,
			view->drawing_area->window,
			x, y-height,
			3, height+1);

  gdk_draw_line (view->drawing_area->window, view->gc,
		 x, y, x, y - height);
  gdk_draw_line (view->drawing_area->window, view->gc,
		 x+1, y, x+1, y - height);
  gdk_draw_line (view->drawing_area->window, view->gc,
		 x+2, y, x+2, y - height);

  if (wp_view_hl_is_on (view))
      {
	doc->getPointFromSpot (view->mark, view, &x, &y, &height);

	x = doc_xcoord_to_view (view, x);
	y = doc_ycoord_to_view (view, y);
	height = doc_coords_to_view (view, height);

	gdk_window_copy_area (view->mark_save_under,
			      view->gc,
			      0, 0,
			      view->drawing_area->window,
			      x, y-height,
			      3, height+1);

	gdk_draw_line (view->drawing_area->window, view->gc,
		       x, y, x, y - height);
	gdk_draw_line (view->drawing_area->window, view->gc,
		       x+1, y, x+1, y - height);
	gdk_draw_line (view->drawing_area->window, view->gc,
		       x+2, y, x+2, y - height);
      }
}


void
wp_view_erase_cursor (WPView *view)
{
  int x, y, height;

  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  doc->sanityCheckSpots ();

  doc->getPointFromSpot (view->point, view, &x, &y, &height);

  x = doc_xcoord_to_view (view, x);
  y = doc_ycoord_to_view (view, y);
  height = doc_coords_to_view (view, height);

  gdk_gc_set_exposures (view->gc, 0);
  gdk_window_copy_area (view->drawing_area->window,
			view->gc,
			x, y-height,
			view->point_save_under,
			0, 0,
			3, height+1);

  if (wp_view_hl_is_on (view))
      {
	doc->getPointFromSpot (view->mark, view, &x, &y, &height);

	x = doc_xcoord_to_view (view, x);
	y = doc_ycoord_to_view (view, y);
	height = doc_coords_to_view (view, height);

	gdk_window_copy_area (view->drawing_area->window,
			      view->gc,
			      x, y-height,
			      view->mark_save_under,
			      0, 0,
			      3, height+1);
      }
}


void wp_view_scroll_point_into_view (WPView *view)
{
  g_return_if_fail (view);
  g_return_if_fail (view->drawing_area);
  g_return_if_fail (view->drawing_area->window);

  WPDocument *doc = wp_view_get_document (view);

  if (! doc->displayRefreshIsOn ()) return;

  doc->sanityCheckSpots ();

  int cx, cy, ch; /* new cursor position */
  int redraw = 0;

  /* Try to keep the point in the window */
  doc->getPointFromSpot (view->point, view, &cx, &cy, &ch);

  cx = doc_xcoord_to_view (view, cx);
  cy = doc_ycoord_to_view (view, cy);
  ch = doc_coords_to_view (view, ch);

  if (cx < 0 || cx > view->drawing_area->allocation.width)
    {
      /* scroll in X direction XXX */
    }

  /*
  if (cy + view->current_insert_style->getFont (view->scale)->descent >
      view->drawing_area->allocation.height)
  */

  if (! (view->point->getSlab ()->getStyle ()->getFont (view->scale)))
    return;

  int current_descent =
    view->
    point->getSlab ()->
    getStyle ()->
    getFont (view->scale)->
    descent;

  current_descent = doc_coords_to_view (view, current_descent);

  if (cy + current_descent > view->drawing_area->allocation.height)
    {
      GtkAdjustment *va = GTK_ADJUSTMENT (view->vadj);

      /* scroll down */
      float t = va->value;
      va->value += va->step_increment;

      if (va->value > va->upper)
	va->value = va->upper;

      if (t != va->value)
	gtk_signal_emit_by_name (GTK_OBJECT (view->vadj), "changed", view);

      redraw = 1;
    }
  else if (cy - ch < 0)
    {
      GtkAdjustment *va = GTK_ADJUSTMENT (view->vadj);

      /* scroll up */
      float t = va->value;
      va->value -= va->step_increment;

      if (va->value < va->lower)
	va->value = va->lower;

      if (t != va->value)
	gtk_signal_emit_by_name (GTK_OBJECT (view->vadj), "changed", view);

      redraw = 1;
    }

  if (redraw == 1)
    {
      GdkRectangle a;

      a.x = wp_view_get_x (view);
      a.y = wp_view_get_y (view);
      a.width = wp_view_get_pixel_width (view);
      a.height = wp_view_get_pixel_height (view);

      gdk_gc_set_foreground (view->gc, view->bg);
      wp_view_fill_rect (view, a.x, a.y, a.width, a.height);
      doc->paint_slab (view, NULL, &a);
      if (! preferences->draw_fast)
	wp_view_copy_back_pixmap (view);
    }
}


/* delete are between point and mark, clear hl,
   don't mess with selection */
void
wp_view_delete_range (WPView *view)
{
  if (!view) return;
  WPDocument *doc = wp_view_get_document (view);
  if (!doc) return;

  /* XXX -- this isn't very efficient. */
  Spot spt1, spt2;
  spt1 = wp_view_get_point (view);
  spt2 = wp_view_get_mark (view);

  if (! wp_spot_order (spt1, spt2))
    {
      Spot t = spt1; spt1 = spt2; spt2 = t;
    }

  doc->displayRefreshOff ();

  while (wp_spot_order (spt1, spt2))
    if (doc->sanityCheckSpots () ||
	doc->getPlugin ()->deleteChild (view, spt1) == 0)
      break;

  view->hl_on = 0;

  doc->displayRefreshOn ();
  doc->refresh ();
  wp_view_scroll_point_into_view (view);
  return;
}


void
wp_view_delete_char (WPView *view)
{
  WPDocument *doc = doc;
  if (! doc)
    return;

  doc->sanityCheckSpots ();

  if (wp_view_hl_is_on (view))
    {
      wp_view_delete_range (view);
      wp_view_hl_off (view);
      return;
    }

  doc->getPlugin ()->deleteChild (view, view->point);
  wp_view_scroll_point_into_view (view);
}


void
wp_view_insert_char (WPView *view, unsigned long int c)
{
  g_return_if_fail (view);

  WPDocument *doc = wp_view_get_document (view);

  doc->sanityCheckSpots ();

  if (wp_view_hl_is_on (view))
    edit_clear_cmd_callback (NULL, view);

  if (c == '\n') c = GDK_Linefeed;
  if (c == '\r') c = GDK_Return;

  doc->getPlugin ()->insertCharacter (view, view->point, c);

  wp_view_scroll_point_into_view (view);
}


void wp_view_move_to_start_of_line (WPView *view)
{
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  doc->moveToStartOfLine (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
}


void wp_view_move_to_end_of_line (WPView *view)
{
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  doc->moveToEndOfLine (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
}


int wp_view_move_to_previous_character (WPView *view)
{
  g_return_val_if_fail (view, 0);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  int ret = doc->moveToPreviousCharacter (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
  return ret;
}


int wp_view_move_to_next_character (WPView *view)
{
  g_return_val_if_fail (view, 0);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  int ret = doc->moveToNextCharacter (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
  return ret;
}


int wp_view_move_to_next_word (WPView *view)
{
  g_return_val_if_fail (view, 0);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  int ret = doc->moveToNextWord (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
  return ret;
}


int wp_view_move_to_previous_word (WPView *view)
{
  g_return_val_if_fail (view, 0);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  wp_view_clear_user_set_cursor_pos (view);
  int ret = doc->moveToPreviousWord (view->point);
  wp_view_scroll_point_into_view (view);

  doc->getPlugin ()->updateUI (view);
  return ret;
}


void wp_view_move_to_previous_line (WPView *view)
{
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  int x, y;

  if (!wp_view_user_set_cursor_pos_is_set (view))
    wp_view_set_user_set_cursor_pos (view);
  wp_view_get_user_set_cursor_pos (view, &x, &y);

  doc->moveToPreviousLine (view->point, x, y);

  wp_view_scroll_point_into_view (view);
  doc->getPlugin ()->updateUI (view);
}


void wp_view_move_to_next_line (WPView *view)
{
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  int x, y;

  if (!wp_view_user_set_cursor_pos_is_set (view))
    wp_view_set_user_set_cursor_pos (view);
  wp_view_get_user_set_cursor_pos (view, &x, &y);

  doc->moveToNextLine (view->point, x, y);

  wp_view_scroll_point_into_view (view);
  doc->getPlugin ()->updateUI (view);
}


void wp_view_hl_on (WPView *view)
{
  view->hl_on = 1;
  if (!view->have_selection)
    {
      view->have_selection = gtk_selection_owner_set (view->drawing_area,
						      GDK_SELECTION_PRIMARY,
						      GDK_CURRENT_TIME);
      /* printf ("acquired selection = %i\n", view->have_selection); */
    }
}


void wp_view_hl_off (WPView *view)
{
  view->hl_on = 0;
  if (view->have_selection) {
    if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) ==
       view->drawing_area->window) {
      gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY,
                               GDK_CURRENT_TIME);
    }
    view->have_selection = FALSE;
  }
}


int wp_view_hl_is_on (WPView *view)
{
  return view->hl_on;
}


Spot wp_view_get_point (WPView *view)
{
  return view->point;
}
Spot wp_view_get_mark (WPView *view)
{
  return view->mark;
}

void wp_view_set_point (WPView *view, Slab slab, int position)
{
  view->point->setSlab (slab);
  view->point->setPosition (position);
}
void wp_view_set_mark (WPView *view, Slab slab, int position)
{
  view->mark->setSlab (slab);
  view->mark->setPosition (position);
}


void wp_view_set_user_set_cursor_pos (WPView *view)
{
  g_return_if_fail (view);
  WPDocument *doc = wp_view_get_document (view);
  doc->sanityCheckSpots ();

  int x, y, height;

  doc->getPointFromSpot (view->point, view, &x, &y, &height);

  x = doc_xcoord_to_view (view, x);
  y = doc_ycoord_to_view (view, y);
  height = doc_coords_to_view (view, height);

  view->user_set_cursor_pos_x = x;
  view->user_set_cursor_pos_y = y;
  view->user_set_cursor_pos_set = 1;
}


void wp_view_clear_user_set_cursor_pos (WPView *view)
{
  view->user_set_cursor_pos_set = 0;
}


int wp_view_user_set_cursor_pos_is_set (WPView *view)
{
  return view->user_set_cursor_pos_set;
}


void wp_view_get_user_set_cursor_pos (WPView *view, int *x, int *y)
{
  if (x) *x = view->user_set_cursor_pos_x;
  if (y) *y = view->user_set_cursor_pos_y;
}


void wp_view_clear_back_pixmap (WPView *view)
{
  if (! view) return;
  if (! view->gc) return;
  if (! view->bg) return;
  if (! view->back_pixmap) return;

  gdk_gc_set_foreground (view->gc, view->bg);
  gdk_draw_rectangle (view->back_pixmap, view->gc, TRUE,
		      0, 0,
		      GTK_WIDGET (view)->allocation.width,
		      GTK_WIDGET (view)->allocation.height);
}


void wp_view_copy_back_pixmap (WPView *view)
{
  if (! view) return;
  if (! view->gc) return;
  if (! view->drawing_area->window) return;
  if (! view->back_pixmap) return;

  gdk_gc_set_exposures (view->gc, 0);
  gdk_window_copy_area (view->drawing_area->window,
			view->gc,
			0, 0,
			view->back_pixmap,
			0, 0,
			GTK_WIDGET (view)->allocation.width,
			GTK_WIDGET (view)->allocation.height);
}


void wp_view_cursor_on (WPView *view)
{
  view->cursor_on = true;
  wp_view_draw_cursor (view);
}


void wp_view_cursor_off (WPView *view)
{
  view->cursor_on = false;
  wp_view_erase_cursor (view);
}


