/*
 XXX
 gdk_char_width (f, word[ i ]);
 need to become
 gtk_text_width (f, word + 1, 2); if this is a 2 byte font

 * Copyright  1997 The Hungry Programmers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <gdk/gdk.h>
#include "wp_word_slab.h"
#include "wp_cont_slab.h"
#include "wp_utils.h"
#include "wp_character.h"
#include "wp_slab.h"
#include "wp_view.h"
#include "wp_prefs.h"


_WordSlab::_WordSlab (WPStyle *set_style) : _Slab (set_style)
{
  parent = NULL;
  word = wp_resize_maybe (NULL, &word_alloced, sizeof (char), 3);
  /*word[ 0 ] = '\0';*/
  word_length = 0;
  this->style = set_style;
  calculateSize (-1);
}


_WordSlab::_WordSlab (WPStyle *set_style, char *set_word, int set_word_length)
  : _Slab (set_style)
{
  word = wp_resize_maybe (NULL,
			  &word_alloced,
			  sizeof (char),
			  set_word_length+2);

  memcpy (word, set_word, set_word_length);
  word_length = set_word_length;
  style = set_style;

  calculateSize (-1);
}


Slab _WordSlab::newOfThisType ()
{
  WordSlab w = new _WordSlab (style, "", 0);
  w->setPluginData (getPluginData ());
  return w;
}

void _WordSlab::setWord (char *set_word, int set_word_length)
{
  word = wp_resize_maybe (word,
			  &word_alloced,
			  sizeof (char),
			  /*strlen (set_word)+1);*/
			  set_word_length+2);
  /*strcpy (word, set_word);*/
  memcpy (word, set_word, set_word_length);
  word_length = set_word_length;
}


char *_WordSlab::getWord ()
{
  word[ word_length ] = '\0';
  return word;
}


int _WordSlab::getWordLength ()
{
  return word_length;
}


void _WordSlab::calculateSize (int debug)
{
  /*int width;*/
  GdkFont *f1 = style->getFont (1); /* scale is one because we want
				       the size in points, not pixels. */

  if (!f1)
    {
      printf ("word has null font!\n");
      return;
    }

  if (debug >= 0)
    {
      printf ("%sWordSlab->calculateSize this=",
	      (const char *) tabber (debug));
      dd ();
      printf ("(l,u,r,d): %d,%d,%d,%d ---> ", left, up, right, down);
    }

  /*width = gdk_string_width (f, word);*/
  /*width = gdk_text_width (f1, word, word_length);*/

  if (style->getDirection () == DIR_RIGHT)
    {
      left = 0;
      /*right = gdk_string_width (f, word);*/
      right = gdk_text_width (f1, word, word_length);
    }
  else /* assume right to left */
    {
      /*left = gdk_string_width (f, word);*/
      left = gdk_text_width (f1, word, word_length);
      right = 0;
    }
  up = f1->ascent;
  down = f1->descent;

  if (debug >= 0)
    printf ("(%d,%d,%d,%d)\n", left, up, right, down);
}


/* x, y are in points */

void _WordSlab::draw (int x, int y, GList *views, int debug)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;

      GdkFont *f = style->getFont (wp_view_get_scale_factor (g));
      GdkFont *f1 = style->getFont (1);
      GdkColor *fgc = style->getForeground ();
      GdkColor *bgc = style->getBackground ();

      if (debug >= 0)
	{
	  printf ("%sdraw this=", (const char *) tabber (debug));
	  dd ();
	  printf (" xy=%d,%d\n", x, y);
	}

      /* draw background color (if different than parents */
      if (parent &&
	  colorcmp (parent->getStyle ()->getBackground (), bgc) == 0)
	{
	  wp_view_set_color (g, bgc);
	  wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
	}


      /* draw highlighting (if any) */
      if (wp_view_hl_is_on (g) &&
	  wp_slab_is_between (wp_view_get_point (g)->getSlab (),
			      this,
			      wp_view_get_mark (g)->getSlab ()))
	{
	  Spot point = wp_view_get_point (g);
	  Spot mark = wp_view_get_mark (g);
	  Slab ps = point->getSlab ();
	  Slab ms = mark->getSlab ();
	  WPDirection d = style->getDirection ();

	  wp_view_set_color (g, wp_view_get_cursor_color (g));

	  /* vodka can turn *anyone* into a serious programmer: */
	  if (this != ps && this != ms)
	    wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
	  else if (this == ps && this == ms)
	    {
	      int a = gdk_text_width (f1, word, point->getPosition ());
	      int b = gdk_text_width (f1, word, mark->getPosition ());
	      if (d == DIR_RIGHT)
		{
		  if (a<b)
		    wp_view_fill_rect (g, x-left + a, y-up, b - a, up+down);
		  else
		    wp_view_fill_rect (g, x-left + b, y-up, a - b, up+down);
		}
	      else /* assume DIR_LEFT */
		{
		  if (a<b)
		    wp_view_fill_rect (g, x - b, y-up, b - a, up+down);
		  else
		    wp_view_fill_rect (g, x - a, y-up, a - b, up+down);
		}
	    }
	  else if (this == ps)
	    {
	      int a = gdk_text_width (f1, word, point->getPosition ());
	      Slab f = topOfTree ()-> firstLowestContainer ();
	      if (d == DIR_RIGHT)
		{
		  if (wp_slab_is_between (f, ms, ps))
		    wp_view_fill_rect (g, x-left, y-up, a, up+down);
		  else
		    wp_view_fill_rect (g, x-left+a, y-up,
				       left+right-a, up+down);
		}
	      else /* assume DIR_LEFT */
		{
		  if (wp_slab_is_between (f, ms, ps))
		    wp_view_fill_rect (g, x-a, y-up, a, up+down);
		  else
		    wp_view_fill_rect (g, x-left, y-up,
				       left+right-a, up+down);
		}

	    }
	  else /* (this == ms)*/
	    {
	      int a = gdk_text_width (f1, word, mark->getPosition ());
	      Slab f = topOfTree ()-> firstLowestContainer ();
	      if (d == DIR_RIGHT)
		{
		  if (wp_slab_is_between (f, ps, ms))
		    wp_view_fill_rect (g, x-left, y-up, a, up+down);
		  else
		    wp_view_fill_rect (g, x-left + a, y-up,
				       left+right - a, up+down);
		}
	      else /* assume DIR_LEFT */
		{
		  if (wp_slab_is_between (f, ps, ms))
		    wp_view_fill_rect (g, x-a, y-up, a, up+down);
		  else
		    wp_view_fill_rect (g, x-left, y-up, left-a, up+down);
		}
	    }
	}


      wp_view_set_color (g, fgc);
      wp_view_draw_string (g, word, word_length, f,
			   x, y, style->getDirection ());

      if (style->getUnderline ())
	wp_view_draw_line (g, x - left, y, x + left + right, y);
    }
}


void _WordSlab::erase (int x, int y, GList *views)
{
  for (GList *v = views; v != NULL; v = v->next)
    {
      WPView *g = (WPView *) v->data;
      if (! slab_overlaps_view (x, y, g)) continue;
      wp_view_set_color (g, style->getBackground ());
      wp_view_fill_rect (g, x-left, y-up, left+right, up+down);
    }
}


boolean _WordSlab::insertChild (int i, Object what, GList *spots,
				GList *views, int debug)
{
  char what_c = ((Character) what)->charValue ();
  int j;
  boolean ret;

  if (debug >= 0)
    {
      printf ("%sWordSlab->insertChild this=",
	      tabber (debug));
      dd ();
      printf (" i=%d what='%c'", i, what_c);
      printf ("\n");
    }


  if (i<0 || i>(int)word_length)
    {
      printf ("error -- WordSlab->insertChild, i=%d", i);
      printf (" word_length=%d", word_length);
      return false;
    }

  if (what->classID () != CharacterID)
    {
      printf ("error -- WordSlab->insertChild, ");
      printf ("'what' isn't a character\n");
      return false;
    }

  unsigned long int smc = style->getMaxChildren ();
  if (0 < smc && smc <= word_length)
    {
      return false;
    }


  word = wp_resize_maybe (word,
			  &word_alloced,
			  sizeof (char),
			  word_length + 2);

  /* insert 'what' */
  for (j=word_length+1; j>i; j--)
    word[ j ] = word[ j-1 ];
  word[ i ] = what_c;
  word_length ++;


  /* figure out what this is going to do to the size of the word
     I used to just add the size of the character on to the size
     of the word, but that messed up with 2 byte fonts */

  int sleft, sup, sright, sdown; /* save size */
  int dleft, dup, dright, ddown; /* size change */

  sleft = left;
  sup = up;
  sright = right;
  sdown = down;

  calculateSize (debug);

  dleft = left - sleft;
  dup = up - sup;
  dright = right - sright;
  ddown = down - sdown;

  /* undo the changes for a moment... */
  left = sleft;
  up = sup;
  right = sright;
  down = sdown;
  word_length --;

  if (style->getDirection () == DIR_RIGHT)
    ret = parent->childSizeChange (this, spots,
				   dleft, dup, dright, ddown,
				   views, (debug >= 0)?debug+1:-1);
  else /* assume DIR_LEFT */
    ret = parent->childSizeChange (this, spots,
				   dleft, dup, dright, ddown,
				   views, (debug >= 0)?debug+1:-1);


  if (ret == false)
    {
      calculateSize (-1);
      if (preferences->draw_fast)
	{
	  /* erase and redraw 'this' */
	  WpPoint p = parent->getPointOfChild (this);
	  erase (p.x, p.y, views);
	  draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
	}
      return false;
    }


  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () >= i)
	s->setPosition (s->getPosition () + 1);
    }

  /* redo the changes */
  word_length ++;

  if (preferences->draw_fast)
    {
      /* erase and redraw 'this' */
      WpPoint p = parent->getPointOfChild (this);
      erase (p.x, p.y, views);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  return true;
}


Object _WordSlab::deleteChild (int i, GList *spots, GList *views,
			       int debug, boolean /*ignore_lost_spots*/)
{
  GdkFont *f = style->getFont (1); /* scale is 1 to get points. */
  char ret;
  int j;
  /*int len = strlen (word);*/

  if (debug >= 0)
    {
      printf ("%sdeleteChild this=", tabber (debug));
      dd ();
      printf ("i=%d\n", i);
    }

  if (i<0 || i >= (int)word_length)
    {
      printf ("error -- WordSlab->deleteChild, i=%d", i);
      printf (" len=%d", word_length);
      return NULL;
    }

  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () > i)
	s->setPosition (s->getPosition () - 1);
    }

  ret = word[ i ];

  WpPoint p = parent->getPointOfChild (this);
  erase (p.x, p.y, views);

  if (style->getDirection () == DIR_RIGHT)
    parent->childSizeChange (this, spots,
			     0, /* left */
			     0, /* up */
			     - gdk_char_width (f, ret),
			     0, /* down */
			     views, (debug >= 0)?debug+1:-1);
  else /* assume DIR_LEFT */
    parent->childSizeChange (this, spots,
			     - gdk_char_width (f, ret),
			     0, /* up */
			     0, /* right */
			     0, /* down */
			     views, (debug >= 0)?debug+1:-1);


  /* delete a character */

  for (j=i; j<(int)word_length+1; j++)
    word[ j ] = word[ j+1 ];
  word_length --;

  /*calculateSize ((debug >= 0)?debug+1:-1);*/

  if (preferences->draw_fast)
    {
      p = parent->getPointOfChild (this);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  if (word_length == 0 &&
      parent != NULL &&
      (getNextNode () != NULL ||
       getPrevNode () != NULL))
    {
      if (debug >= 0)
	printf ("%sWordSlab->deleteChild -- cleaning up a node->->->",
		tabber (debug));


      parent->deleteChild (parent->getIndexOfChild (this),
			   spots, views, (debug >= 0)?debug+1:-1, false);
    }


  /* offer children to prev */
  if (parent != NULL && parent->getIndexOfChild (this) == 0)
    {
      Slab prev = parent->getPrevNode ();

      /*if (prev != NULL && prev instanceof ContSlab)*/
      if (prev != NULL && prev->classID () == ContSlabID)
	{
	  ContSlab cs = (ContSlab) prev;
	  cs->attemptToStealChildren (spots, views, (debug >= 0)?debug+1:-1);
	}
    }

  return new _Character (ret);
}



/*
  change the contents of word and do visual updates and wrapping
  */

boolean _WordSlab::replaceWord (char *new_word, int new_word_length,
				GList *spots,
				GList *views, int debug)
{
  GdkFont *f1 = style->getFont (1); /* scale is 1 to get points. */
  int width_change;

  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () > new_word_length)
	s->setPosition (new_word_length);
    }

  width_change = gdk_text_width (f1, new_word, new_word_length);
  width_change -= (left + right);


  parent->childSizeChange (this, spots,
			   0, /* left */
			   0, /* up */
			   width_change,
			   0, /* down */
			   views, (debug >= 0)?debug+1:-1);

  word = wp_resize_maybe (word,
			  &word_alloced,
			  sizeof (char),
			  new_word_length + 2); /* +1? XXX */


  /*strcpy (word, new_word);*/
  memcpy (word, new_word, new_word_length);
  word_length = new_word_length;

  /*calculateSize ((debug >= 0)?debug+1:-1);*/

  if (preferences->draw_fast)
    {
      /* erase and redraw 'this' */
      WpPoint p = parent->getPointOfChild (this);
      erase (p.x, p.y, views);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  return true;
}



boolean _WordSlab::joinWithNext (GList *spots, GList *views, int debug)
{
  char *nextword;

  if (parent == NULL)
    {
      fprintf (stderr, "error -- WordSlab->joinWithNext, parent=NULL");
      return false;
    }

  if (debug >= 0)
    {
      printf ("%sWordSlab->joinWithNext this=",
	      tabber (debug));
      dd ();
      printf ("\n");
    }

  Slab nextSlab = getNextNode ();
  if (nextSlab == NULL)
    return true;

  if (nextSlab->classID () !=  WordSlabID)
    return true;

  WordSlab next = (WordSlab) nextSlab;

  if (getNumberOfChildren () == 0)
    {
      parent->deleteChild (parent->getIndexOfChild (this),
			   spots, views, (debug >= 0)?debug+1:-1, false);
      return true;
    }

  if (next->getNumberOfChildren () == 0)
    {
      Slab p = next->getParent ();
      p->deleteChild (p->getIndexOfChild (next),
		      spots, views, (debug >= 0)?debug+1:-1, false);
      return true;
    }


  if (this->getStyle () != next->getStyle ())
    return true;

  int origNumChildren = getNumberOfChildren ();
  int nextWidth = next->getWidth ();

  /*word += next->getWord ();*/
  /*g_string_append (word, next->getWord ()->str);*/

  nextword = next->getWord ();
  int next_word_length = next->getNumberOfChildren ();
  word = wp_resize_maybe (word,
			  &word_alloced,
			  sizeof (char),
			  /*
			  strlen (word) + strlen (next->getWord ()) + 1);
			  */
			  word_length + next_word_length + 2);
  /*strcat (word, nextword);*/
  memcpy (word + word_length, nextword, next_word_length);
  word_length += next_word_length;


  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == next)
	{
	  s->setSlab (this);
	  s->setPosition (s->getPosition () + origNumChildren);
	}
    }


  Slab nextParent = next->getParent ();

  if (nextParent == parent)
    {
      /* cheat */
      if (style->getDirection () == DIR_RIGHT)
	{
	  right += nextWidth;
	  next->setRight (0);
	  next->setLeft (0);
	}
      else /* assume DIR_LEFT */
	{
	  left += nextWidth;
	  next->setRight (0);
	  next->setLeft (0);
	}
      parent->deleteChild (parent->getIndexOfChild (next),
			   spots, views, (debug >= 0)?debug+1:-1, false);
    }
  else
    {
      nextParent->deleteChild (nextParent->getIndexOfChild (next),
			       spots, views, (debug >= 0)?debug+1:-1, false);

      parent->childSizeChange (this, spots, 0, 0, nextWidth, 0, views,
			       (debug >= 0)?debug+1:-1);

      if (preferences->draw_fast)
	{
	  WpPoint p = parent->getPointOfChild (this);
	  erase (p.x, p.y, views);
	  draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
	}
    }

  return true;
}


boolean _WordSlab::childSizeChange (Slab child, GList */*spots*/,
				    int /*dLeft*/, int /*dUp*/,
				    int /*dRight*/, int /*dDown*/,
				    GList */*views*/, int debug)
{
  if (debug >= 0)
    {
      printf ("%schildSizeChange this=", tabber (debug));
      dd ();
      printf (" child=");
      child->dd ();
      printf ("\n");
    }

  fprintf (stderr, "error -- WordSlab::childSizeChange called\n");
  return false;
}


boolean _WordSlab::ensureBorder (int i, GList *spots,
				 GList *views, int debug)
{
  WordSlab newWS;

  if (debug >= 0)
    {
      printf ("%sensureBorder this=", tabber (debug));
      dd ();
      printf (" i=%d\n", i);
    }
  if (i<0 || i>(int)word_length)
    {
      fprintf (stderr, "error -- WordSlab->ensureBorder, i=%d, ", i);
      fprintf (stderr, "word->length=%d\n", word_length);
      return false;
    }
  if (parent == NULL)
    {
      fprintf (stderr, "error -- WordSlab->ensureBorder, parent=NULL");
      return false;
    }

  /*
  if (debug >= 0)
    printf (Utils->tabber (debug) +
			"WordSlab->ensureBorder this=" + dd () +
			" i=" + i);
			*/

  if (i == 0 || i == (int)word_length)
    return true;

  newWS = (WordSlab) newOfThisType ();
  newWS->setWord (word + i, word_length - i);
  /*int orig_word_length = word_length;*/
  /*word[ i ] = '\0';*/
  word_length = i;

  newWS->calculateSize ((debug >= 0)?debug+1:-1);
  calculateSize ((debug >= 0)?debug+1:-1);

  /* fix the spots */
  for (int spotnum=0; (guint)spotnum<g_list_length (spots); spotnum++)
    {
      Spot s = (Spot) g_list_nth (spots, spotnum)->data;
      if (s->getSlab () == this && s->getPosition () >= i)
	{
	  s->setPosition (s->getPosition () - i);
	  s->setSlab (newWS);
	}
    }

  /* do an insert without all the redrawing, etc. */
  parent->insertChildAt (newWS, parent->getIndexOfChild (this) + 1);

  /*
  if (parent->insertChild (parent->getIndexOfChild (this) + 1, newWS,
			   spots, views, (debug >= 0)?debug+1:-1) == false)
    {
      delete newWS;
      word_length = orig_word_length;
      calculateSize ((debug >= 0)?debug+1:-1);
    }
  */

  if (preferences->draw_fast)
    {
      /* redraw both children, because splitting can move them around */
      WpPoint p = getParent ()->getPointOfChild (this);
      erase (p.x, p.y, views);
      draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
      p = newWS->getParent ()->getPointOfChild (newWS);
      newWS->erase (p.x, p.y, views);
      newWS->draw (p.x, p.y, views, (debug >= 0)?debug+1:-1);
    }

  /* offer children to prev */
  if (parent->getIndexOfChild (this) == 0)
    {
      Slab prev = parent->getPrevNode ();

      if (prev != NULL && prev->classID () == ContSlabID)
	{
	  ContSlab cs = (ContSlab) prev;
	  cs->attemptToStealChildren (spots, views, (debug >= 0)?debug+1:-1);
	}
    }

  return true;
}


int _WordSlab::getIndexOfChild (Object /*what*/)
{
  fprintf (stderr, "error -- WordSlab->getIndexOfChild called");
  return -1;
}


Object _WordSlab::getChildAt (int i)
{
  /*if ((int) strlen (word) < i)*/
  if ((int)word_length < i)
    {
      fprintf (stderr, "WordSlab->getChildAt -- error -- %d > %d\n",
	       i, word_length);
      return NULL;
    }

  return new _Character (word[ i ]);
}


void _WordSlab::insertChildAt (Object /*what*/, int /*i*/)
{
  fprintf (stderr, "error -- WordSlab->insertChildAt called");
}


void _WordSlab::dumpToScreen (int tab)
{
  for (int t=0; t<tab; t++) printf ("+-");
  printf ("Word(%d,%d,%d,%d) pp:%p : ", left, up, right, down, plugin_data);

  /*for (int i=0; i< (int) strlen (word); i++)*/
  for (int i=0; i< (int)word_length; i++)
    {
      char c = word[ i ];
      switch (c)
	{
	case '\n': printf ("\\n"); break;
	case '\r': printf ("\\r"); break;
	default: printf ("%c", c); break;
	}
    }

  printf (" <this=%p> <parent=%p>\n", this, parent);
}


void _WordSlab::dd ()
{
  /*
    return
    "["+word+"("+ left + "," +
    + up + "," +
    + right + "," +
    + down + ")]";
    */

  /*printf ("[%s]", word);*/

  printf ("[");
  for (int i=0; i< (int)word_length; i++)
    printf ("%c", word[ i ]);
  printf ("]");
}


Slab _WordSlab::firstLowestContainer () {return this;}
Slab _WordSlab::lastLowestContainer () {return this;}


Slab _WordSlab::getNextNode ()
{
  if (parent == NULL)
    return NULL;
  return parent->getNextNodeUp (this, 1);
}
Slab _WordSlab::getNextNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _WordSlab::getNextNodeDown (int depth)
{
  if (depth == 0)
    return this;
  return NULL;
}


Slab _WordSlab::getPrevNode ()
{
  if (parent == NULL)
    return NULL;
  return parent->getPrevNodeUp (this, 1);
}
Slab _WordSlab::getPrevNodeUp (Slab /*child*/, int /*depth*/) {return NULL;}
Slab _WordSlab::getPrevNodeDown (int depth)
{
  if (depth == 0)
    return this;
  return NULL;
}


int _WordSlab::getNumberOfChildren ()
{
  /*return strlen (word);*/
  return word_length;
}


WpPoint _WordSlab::getPointOfChild (Slab child)
{
  return getPointOfChild (getIndexOfChild (child));
}


/* returns values in points, not pixels */
WpPoint _WordSlab::getPointOfChild (int childNum)
{
  GdkFont *f1 = style->getFont (1); /* scale is 1 to get points. */
  int x=0, y=0;

  /*if (childNum < 0 || childNum > (int) strlen (word))*/
  if (childNum < 0 || childNum > (int)word_length)
    {
      fprintf (stderr, "error -- WordSlab->getPointOfChild: ");
      fprintf (stderr, "word->length=%d,", word_length);
      fprintf (stderr, "childNum=%d\n", childNum);
      return newWpPoint (x, y);
    }

  if (parent != NULL)
    {
      WpPoint my_point = parent->getPointOfChild (this);
      x = my_point.x;
      y = my_point.y;
    }

  if (word_length > 0)
    {
      if (style->getDirection () == DIR_RIGHT)
	x += gdk_text_width (f1, word, childNum);
      else /* assume DIR_LEFT */
	x -= gdk_text_width (f1, word, childNum);

#     if 0 /* I'm *so* sorry to see this go... */
      char sv = word[ childNum ];
      word[ childNum ] = '\0';
      if (style->getDirection () == DIR_RIGHT)
	x += gdk_text_width (f1, word, word_length);
      else /* assume DIR_LEFT */
	x -= gdk_text_width (f1, word, word_length);
      word[ childNum ] = sv;
#     endif /* 0 */
    }

  return newWpPoint (x, y);
}


/* px, py are in points, not pixels. */
Spot _WordSlab::getSpotOfPoint (int px, int /*py*/)
{
  GdkFont *f = style->getFont (1); /* scale is 1 to get points. */

  int i;
  /*int len = strlen (word);*/
  int x;
  WPDirection d = style->getDirection ();

  if (d == DIR_RIGHT)
    {
      x = 0;

      if (px < 0)
	return new _Spot (this, 0);

      for (i=0; i<(int)word_length; i++)
	{
	  x += gdk_char_width (f, word[ i ]);
	  if (x > px)
	    return new _Spot (this, i);
	}

      return new _Spot (this, word_length);
    }
  else /* assume DIR_LEFT */
    {
      x = 0;

      if (px > 0)
	return new _Spot (this, word_length);

      for (i=0; i<(int)word_length; i++)
	{
	  x -= gdk_char_width (f, word[ i ]);
	  if (x < px)
	    return new _Spot (this, i);
	}

      return new _Spot (this, 0);
    }
}


Slab _WordSlab::createFollower (GList */*spots*/,
				GList */*views*/,
				int /*debug*/)
{
  printf ("WordSlab -- createFollower got called\n");
  return NULL;
}
